/************************************************************************
 *
 *  Application.java
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2.1, as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA  02111-1307  USA
 *
 *  Copyright: 2002-2004 by Henrik Just
 *
 *  All Rights Reserved.
 * 
 *  Version 0.3.3a (2003-12-24) 
 *
 */

package writer2latex;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.URL;
import java.util.Enumeration;
import javax.xml.parsers.ParserConfigurationException;
import org.xml.sax.SAXException;

import org.openoffice.xmerge.Convert;
import org.openoffice.xmerge.ConvertData;
import org.openoffice.xmerge.Document;
import org.openoffice.xmerge.ConverterFactory;
import org.openoffice.xmerge.util.registry.ConverterInfoMgr;
import org.openoffice.xmerge.util.registry.ConverterInfoReader;
import org.openoffice.xmerge.util.registry.RegistryException;

import writer2latex.util.Misc;
import writer2latex.office.*;

/**
 * <p>Commandline utility to convert an OpenOffice.org Writer XML file to a LaTeX file<.</p>
 * <p>The code is based on org.openoffice.xmerge.test.Driver</p>
 */
public final class Application {

    /**  Based on command-line parameters. */
    private String sOutputFormat = null;
    private String toMime = null;
    private String sFileName = null;
    private String sOutFileName = null;
    private String sConfigFileName = null;
    private String sJarFileName = null;

    /* Note: For some reason xmerge does not use the official MIME type */
    private final String fromMime = "staroffice/sxw";

    Config config;

    /**
     *  Main method
     *
     *  @param  args  The argument passed on the command line.
     */

    public static final void main (String[] args){

        try {
            Application app = new Application();
            app.parseCommandLine(args);
            System.out.print("Preparing to convert...");
            app.loadConfiguration();
            app.loadPlugin();
            System.out.print("starting conversion...");
            app.doConversion();
            System.out.println("done!");
        } catch (IllegalArgumentException ex) {
            String msg = ex.getMessage();
            if (msg != null) System.out.println("\n" + msg);
            showUsage();
        } catch (Exception ex) {
            String msg = ex.getMessage();
            if (msg != null) System.out.println("\n" + msg);
            ex.printStackTrace();
        }

    }
	
    /** Load the configuration and save it for this file */
    private void loadConfiguration() {
        if (sConfigFileName==null) { return; }
        String sLocalConfigFileName
            = sFileName.substring(0,sFileName.lastIndexOf("."))+"-config.xml";
        Config config = new Config();
        config.read(sConfigFileName);
        config.write(sLocalConfigFileName);
    }

	/**
     *  <p>Load the plugin</p>
     */
    private void loadPlugin() {
        // TODO: Is there a more elegant approach to find the jar file name???
        String s = this.getClass().getResource("symbols.xml").toString();
        sJarFileName = s.substring(s.indexOf("file:"),s.indexOf("!"));
		
        ConverterInfoReader cir = null;
        try {
            cir = new ConverterInfoReader(sJarFileName, false);
        } catch (Exception e) {
            System.out.println("Oops, there was an error loading the converter");
            e.printStackTrace();
        }
		
        Enumeration jarInfoEnumeration = cir.getConverterInfoEnumeration();
        try {
            ConverterInfoMgr.addPlugIn(jarInfoEnumeration);
        } catch (RegistryException e) {
            System.out.println("Oops, there seems to be a configuration problem");
            e.printStackTrace();
        }
    }
    /**
     *  Gets a <code>Convert</code> object using the
     *  <code>ConverterFactory</code> and  does the conversion using
     *  this object.
     *
     *  @throws  IllegalArgumentException  If an argument is invalid.
     */
    private void doConversion() throws IllegalArgumentException {

        ConverterFactory cf = new ConverterFactory();
        Convert myConvert   = cf.getConverter(fromMime, toMime);
        String processFile  = null;

        if (myConvert == null) {
            System.out.println("Oops, there seems to be a configuration problem!");
            throw new IllegalArgumentException();
        }

        try {
            File f = new File(sFileName);
            // Make sure the input file actually exists before using it
            if (!f.exists()) {
                System.out.println("I'm sorry, I can't find "+sFileName);
                System.exit(0);
            }
            // If filename ends with .xml, we suppose that it's flat xml:
            boolean bFlat = sFileName.toLowerCase().endsWith(".xml");
            FileInputStream fis = new FileInputStream(f);
            myConvert.addInputStream(sOutFileName, fis, !bFlat);
            //myConvert.addInputStream(f.getName(), fis, !bFlat);
        }
        catch (IOException e) {
            System.out.println("Oops, there was an error reading "+sFileName);
            e.printStackTrace();
        }
		
        ConvertData dataOut = null;

        try {
            dataOut = myConvert.convert();
        } catch (Exception convertExcept) {
            System.out.println("\nThere was an error in the conversion");
            convertExcept.printStackTrace();
        }

        Enumeration docEnum = dataOut.getDocumentEnumeration();
        while (docEnum.hasMoreElements()) {
            Document docOut      = (Document)docEnum.nextElement();
            String fileName      = docOut.getFileName();
            try {
                FileOutputStream fos = new FileOutputStream(fileName);
                docOut.write(fos);
                fos.flush();
                fos.close();
            } catch (Exception writeExcept) {
                System.out.println("\nThere was an error writing out file <" +
                    fileName + ">");
                writeExcept.printStackTrace();
            }
        }
    }

    /**
     *  Display usage.
     */
    private static void showUsage() {
        System.out.println();
        System.out.println("Usage:");
        System.out.println("   java -jar <path>/writer2latex.jar <options> <document> <output file name>");
        System.out.println("where <options> is one of the following:");
        System.out.println("   [-latex] [-config <configuration file>]");
        System.out.println("   -bibtex");
        System.out.println("   -xhtml [-config <configuration file>]");
        System.out.println("   -xhtml+mathml [-config <configuration file>]");
        System.out.println("   -xhtml+mathml+xsl [-config <configuration file>]");
        //System.out.println("   -text <document>");
    }


    /**
     *  Parse command-line arguments.
     *
     *  @param  args  Array of command line arguments.
     *
     *  @throws  IllegalArgumentException  If an argument is invalid.
     */
    private void parseCommandLine(String args[])
        throws IllegalArgumentException {

        if (args.length == 0) {
            throw new IllegalArgumentException();
        }

        int i = 0;
        String sArg;
		

        sArg = getArg(i,args);
        if ("-xhtml".equals(sArg)) {
            sOutputFormat = "xhtml";
            toMime = MIMETypes.XHTML;
            i++;
        } 
        else if ("-xhtml+mathml".equals(sArg)) {
            sOutputFormat = "xhtml";
            toMime = MIMETypes.XHTML_MATHML;
            i++;
        } 
        else if ("-xhtml+mathml+xsl".equals(sArg)) {
            sOutputFormat = "xhtml";
            toMime = MIMETypes.XHTML_MATHML_XSL;
            i++;
        } 
        else if ("-text".equals(sArg)) {
            sOutputFormat = "Text";
            toMime = MIMETypes.TEXT;
            i++;
        }
        else if ("-bibtex".equals(sArg)) {
            sOutputFormat = "BibTeX";
            toMime = MIMETypes.BIBTEX;
            i++;
        }
        else if ("-latex".equals(sArg)) {
            sOutputFormat = "LaTeX";
            toMime = MIMETypes.LATEX;
            i++;
        }
        else { // assume LaTeX as default
            sOutputFormat = "LaTeX";
            toMime = MIMETypes.LATEX;
        }

        //System.out.println("Converting to "+toMime);
		
        /*sArg = getArg(i,args);
        if ("-plugin".equals(sArg)) {
            i++;
            sJarFileName = getArg(i++,args);
        }*/

        sArg = getArg(i,args);
        if ("-config".equals(sArg)) {
            i++;
            sConfigFileName = getArg(i++,args);
        }

        sFileName = getArg(i++,args);
		
        if (i<args.length) { sOutFileName = getArg(i,args); }
        else { sOutFileName = sFileName; }
        if (sOutFileName.endsWith(".tex")) {
            sOutFileName = sOutFileName.substring(0,sOutFileName.length()-4);
        }		

        System.out.println();
        System.out.println("This is Writer2" + sOutputFormat
            + ", Version " + Misc.VERSION + " (" + Misc.DATE + ")");
        System.out.println();

    }


    /**
     *  Extract the next argument from the array, while checking to see
     *  that the array size is not exceeded.  Throw a friendly error
     *  message in case the arg is missing.
     *
     *  @param  i     Argument index.
     *  @param  args  Array of command line arguments.
     *
     *  @return  The argument with the specified index.
     *
     *  @throws  IllegalArgumentException  If an argument is invalid.
     */
    private String getArg(int i, String args[])
        throws IllegalArgumentException {

        if (i < args.length)
            return args[i];
        else throw new
            IllegalArgumentException("I'm sorry, the commandline ended abnormally");
    }



}
