/************************************************************************
 *
 *  Converter.java
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2.1, as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA  02111-1307  USA
 *
 *  Copyright: 2002-2004 by Henrik Just
 *
 *  All Rights Reserved.
 * 
 *  Version 0.3.3a (2004-02-20)
 *
 */

package writer2latex;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Element;

import java.io.File;
import java.io.OutputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.zip.ZipException;
import java.util.LinkedList;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

import org.openoffice.xmerge.ConvertData;
import org.openoffice.xmerge.converter.xml.EmbeddedObject;
import org.openoffice.xmerge.converter.xml.EmbeddedBinaryObject;
import org.openoffice.xmerge.converter.xml.EmbeddedXMLObject;
import org.openoffice.xmerge.converter.xml.sxw.SxwDocument;

import writer2latex.util.*;
import writer2latex.xmerge.*;
import writer2latex.office.*;
import writer2latex.bibtex.*;

/**
 *  <p>This class converts a Writer XML file to a LaTeX file<.</p>
 */
public final class Converter {
    private Config config;
    private I18n i18n;
    private ConvertData convertData;            
    private LaTeXDocument texDoc;
	private LaTeXDocumentPortion preamble;
	private LaTeXDocumentPortion body;
    private SxwDocument sxwDoc;
	
    private BibTeXDocument bibDoc;
    
    private WriterStyleCollection sc = new WriterStyleCollection();
    private StyleConverter sconv;
	
    private MetaData metaData;
	
    private ExportNameCollection targets = new ExportNameCollection(true);
    private String[] sHeadingStyles = new String[6];
    private ExportNameCollection refnames = new ExportNameCollection(true);
    private ExportNameCollection bookmarknames = new ExportNameCollection(true);
    private ExportNameCollection footnotenames = new ExportNameCollection(true);
    private ExportNameCollection endnotenames = new ExportNameCollection(true);
    private ExportNameCollection seqrefnames = new ExportNameCollection(true);
    //private String sDefaultLanguage = null;
	
    private boolean bInHeading = false;
    private int nFrameLevel = 0;
    private int nMulticolLevel = 0;
    private int nNoFootnotesLevel = 0;
    private boolean bProcessingFootnote = false;

    private boolean bNeedLongtable = false;
    private boolean bNeedMulticol = false;
    private boolean bNeedGraphicx = false;
    private boolean bContainsAlphabeticalIndex = false;
    private boolean bContainsEndnotes = false;
    private boolean bContainsFootnotes = false;
    private boolean bContainsTables = false;
    // Keep track of footnotes (inside minipage etc.), that should be typeset later
    private LinkedList postponedFootnotes = new LinkedList();
    // Keep track of floating frames (images, textboxes...)
    private LinkedList floatingFrames = new LinkedList();
    // Keep track of postponed hyperlink targets (cannot occur in headings)
    private LinkedList hyperlinkTargets = new LinkedList();
    // Name of first used master page, if any
    private String sMainMasterPage = null;
    // Value of attribute text:display of most recent text:chapter field
    // This is used to handle chaptermarks in headings
    private String sChapterField1 = null;
    private String sChapterField2 = null;
	
    private MathmlConverter mathmlconverter;
    private ImageLoader imageLoader;
	
    public Converter(SxwDocument sxwDoc) {
        // The configuration file is supposed to be saved in a special file for
        // this document. If it doesn't exist, create it!
        String sConfigFileName = sxwDoc.getName()+"-config.xml";
        //System.out.println("Config file: "+sConfigFileName);
        config = new Config();
        File f = new File(sConfigFileName);
        if (f.exists()) {
            config.read(sConfigFileName);
        }
        else if (config.createDocumentConfig()) {
            config.write(sConfigFileName);
        }
        i18n = new I18n(config);
        sconv = new StyleConverter(sc,i18n,config);
        metaData = new MetaData(sxwDoc);
        this.sxwDoc = sxwDoc;
        this.texDoc = new LaTeXDocument(sxwDoc.getName());
        texDoc.setEncoding(I18n.writeJavaEncoding(config.getInputencoding()));
        preamble = texDoc.getPreamble();
        body = texDoc.getBody();
        if (config.useBibtex()) {
            bibDoc = new BibTeXDocument(sxwDoc.getName());
        }
        convertData = new ConvertData();
		mathmlconverter = new MathmlConverter(i18n, config.getMathSymbols());
        imageLoader = new ImageLoader(sxwDoc,true);
        if (config.debug()) {
            System.out.println();
            System.out.println("Loaded "+i18n.getCharCount()+" unicode symbols");
        }
    }
    
    
    public ConvertData convert() {      
        Document styles = sxwDoc.getStyleDOM();
        Document doc = sxwDoc.getContentDOM();
		
        // Load style info
        sc.loadStylesFromDOM(styles,doc);
        
        // Traverse the body (in flat HTML there may be several bodies, but
        // the first one is the main body).
        NodeList list;
        list = doc.getElementsByTagName(XMLString.OFFICE_BODY);
        int nLen = list.getLength();
        if (nLen > 0) {
            Node node = list.item(0);
            handleBody(node);
        }

        // Determine which master page and page master to use first
        // If no page master was found when traversing the document, use the first
        // master page in styles.xml
        if (sMainMasterPage==null) {
            sMainMasterPage = sc.getFirstMasterPageName();
            if (sMainMasterPage!=null) { sconv.masterPageNames.addName(sMainMasterPage); }
        }
        String sMainPageMaster = null;
        if (sMainMasterPage!=null) {
            MasterPage style = sc.getMasterPage(sMainMasterPage);
            if (style!=null) {
                sMainPageMaster = style.getProperty(XMLString.STYLE_PAGE_MASTER_NAME);
            }
        }

        // Create options for documentclass
        CSVList docOptions = new CSVList(",");

        StyleWithProperties style = sc.getDefaultParStyle();
        if (style!=null) {
            String s = style.getProperty(XMLString.FO_FONT_SIZE);
            if ("10pt".equals(s)) { docOptions.addValue("10pt"); }
            if ("11pt".equals(s)) { docOptions.addValue("11pt"); }
            if ("12pt".equals(s)) { docOptions.addValue("12pt"); }
        }
		
        // TODO (0.4): twoside option should be optional :-)
        if (config.pageFormatting()!=Config.IGNORE_ALL) { docOptions.addValue("twoside"); }

        // Add preamble
		preamble.append("% This file was converted to LaTeX by Writer2LaTeX ver. "+Misc.VERSION).nl()
                .append("% see http://www.hj-gym.dk/~hj/writer2latex for more info").nl();
        preamble.append("\\documentclass");
        if (!docOptions.isEmpty()) {
            preamble.append("[").append(docOptions.toString()).append("]");
        }
        preamble.append("{").append(config.getDocumentclass()).append("}").nl();

        // Use packages if needed
        // inputenc
        if (config.getInputencoding()==I18n.UTF8){
            preamble.append("\\usepackage[dirty,autogenerated]{ucs}").nl();
        }
        preamble.append("\\usepackage[")
                .append(I18n.writeInputenc(config.getInputencoding()))
                .append("]{inputenc}").nl();

        // fontenc
        CSVList fontencs = new CSVList(",");
        if (i18n.cyrillic()) { fontencs.addValue("T2A"); }
        if (i18n.greek()) { fontencs.addValue("LGR"); }
        if (config.useTipa()) { fontencs.addValue("T3"); }
        fontencs.addValue("T1");
        preamble.append("\\usepackage[").append(fontencs.toString())
                .append("]{fontenc}").nl();

        if (config.useTipa()) { preamble.append("\\usepackage[noenc]{tipa}").nl(); }

        // Bbding: (Has to avoid some nameclashes.)
        if (config.useBbding()) {
            preamble.append("\\usepackage{bbding}").nl()
                    .append("\\let\\bbCross\\Cross\\let\\Cross\\undefined").nl()
                    .append("\\let\\bbSquare\\Square\\let\\Square\\undefined").nl()
                    .append("\\let\\bbTrianbleUp\\TriangleUp\\let\\TriangleUp\\undefined").nl()
                    .append("\\let\\bbTrianlgeDown\\TriangleDown\\let\\TriangleDown\\undefined").nl();
        }
		
        if (config.useIfsym()) { preamble.append("\\usepackage[geometry,weather,misc,clock]{ifsym}").nl(); }

        if (config.usePifont()) { preamble.append("\\usepackage{pifont}").nl(); }

        if (config.useEurosym()) { preamble.append("\\usepackage{eurosym}").nl(); }

        // wasysym must be loaded between amsmath and amsfonts!
        if (config.useWasysym()) { 
            preamble.append("\\usepackage{amsmath,wasysym,amssymb,amsfonts,textcomp,writer}").nl();
        }
        else {
            preamble.append("\\usepackage{amsmath,amssymb,amsfonts,textcomp,writer}").nl();
        }
		
        if (config.useColor()) { preamble.append("\\usepackage{color}").nl(); }

        if (bNeedGraphicx) { 
            preamble.append("\\usepackage");
            if (config.getBackend()==Config.PDFTEX) preamble.append("[pdftex]");
            else if (config.getBackend()==Config.DVIPS) preamble.append("[dvips]");
            preamble.append("{graphicx}").nl();
        }
		
        if (bContainsAlphabeticalIndex) {
            preamble.append("\\usepackage{makeidx}").nl()
                    .append("\\makeindex").nl();
        }
	
        if (bNeedMulticol) { preamble.append("\\usepackage{multicol}").nl(); }
		
        if (bNeedLongtable) { preamble.append("\\usepackage{longtable}").nl(); }
		
        if (bContainsEndnotes) { preamble.append("\\usepackage{endnotes}").nl(); }

        if (sconv.needUlem()) { preamble.append("\\usepackage[normalem]{ulem}").nl(); }
		
        if (config.useHyperref()){
            preamble.append("\\usepackage{hyperref}").nl()
                    .append("\\hypersetup{");
            if (config.getBackend()==Config.PDFTEX) preamble.append("pdftex");
            else if (config.getBackend()==Config.DVIPS) preamble.append("dvips");
            else preamble.append("hypertex");
            preamble.append(", colorlinks=true, linkcolor=blue, filecolor=blue, pagecolor=blue, urlcolor=blue");
            if (config.getBackend()==Config.PDFTEX) {
                createPdfMeta("pdftitle",metaData.getTitle());
                createPdfMeta("pdfauthor",metaData.getCreator());
                createPdfMeta("pdfsubject",metaData.getSubject());
                createPdfMeta("pdfkeywords",metaData.getKeywords());
            }
            preamble.append("}").nl();
        }

        // Use babel.sty
        sconv.convertLanguageUsage(preamble);

        // Set padding for table cells (1mm is default in OOo!)
        // For vertical padding we can only specify a relative size
        if (bContainsTables) {
            preamble.append("\\setlength\\tabcolsep{1mm}").nl();
            preamble.append("\\renewcommand\\arraystretch{1.3}").nl();
        }

        // We typeset with \raggedbottom since OOo doesn't use rubber lengths
        // TODO: Maybe turn vertical spacing from OOo into rubber lengths?
        preamble.append("\\raggedbottom").nl();

        // Until reading settings for formulas is implemented, we need this
        preamble.append("\\everymath{\\displaystyle}").nl();

        // Add style declarations
        StyleWithProperties dpstyle = sc.getDefaultParStyle();
        String sLang = dpstyle!=null ? dpstyle.getProperty(XMLString.FO_LANGUAGE) : "";
        convertMasterPages(sLang);
        sconv.useMasterPage(sMainMasterPage,preamble);
        if (bContainsFootnotes) sconv.convertFootnotesConfiguration(preamble);
        if (bContainsEndnotes) sconv.convertEndnotesConfiguration(preamble);
        sconv.convertHeadingStyles(sHeadingStyles,preamble);
        sconv.convertListStyles(preamble);
        sconv.convertTextStyles(preamble);
        sconv.convertParStyles(preamble);
		
        // Add custom preamble
        LinkedList customPreamble = config.getCustomPreamble();
        int nCPLen = customPreamble.size();
        for (int i=0; i<nCPLen; i++) {
            preamble.append( (String) customPreamble.get(i) ).nl();
        }
		
        // Set \title, \author and \date (for \maketitle)
        createMeta("title",metaData.getTitle());
        //createMeta("author",metaData.getCreator());
        // todo: format date
        //createMeta("date",metaData.getDate());

        // Return the converted data
        convertData.addDocument(texDoc);
        if (config.useBibtex()) { convertData.addDocument(bibDoc); }
        return convertData;
    }
	
    private void createMeta(String sName, String sValue) {
        if (sValue==null) { return; }
        // Meta data is assumed to be in the default language:
        StyleWithProperties dpstyle = sc.getDefaultParStyle();
        String sLang = dpstyle!=null ? dpstyle.getProperty(XMLString.FO_LANGUAGE) : "en";
        preamble.append("\\"+sName+"{"+i18n.convert(sValue,false,sLang)+"}").nl();
    }

    private void createPdfMeta(String sName, String sValue) {
        if (sValue==null) { return; }
        // Replace commas with semicolons (the keyval package doesn't like commas):
        sValue = sValue.replace(',', ';');
        // Meta data is assumed to be in the default language:
        StyleWithProperties dpstyle = sc.getDefaultParStyle();
        String sLang = dpstyle!=null ? dpstyle.getProperty(XMLString.FO_LANGUAGE) : "en";
        preamble.append(", "+sName+"="+i18n.convert(sValue,false,sLang));
    }
  
    /*
     * Process header or footer contents
     */
    private void convertMasterPages(String sLang) {
        if (config.pageFormatting()==Config.IGNORE_ALL) { return; }
        Enumeration styles = sc.getMasterPages().getStylesEnumeration();
        preamble.append("% Pages styles (master pages)").nl()
                .append("\\makeatletter").nl();
        while (styles.hasMoreElements()) {
            MasterPage style = (MasterPage) styles.nextElement();
            String sName = style.getName();
            if (sconv.masterPageNames.containsName(sName)) {
                preamble.append("\\newcommand\\ps@")
                        .append(sconv.masterPageNames.getExportName(sName))
                        .append("{%").nl();
                sChapterField1 = null;
                sChapterField2 = null;
                // Header
                preamble.append("\\renewcommand\\@oddhead{");
                if (style.getHeader()!=null) {
                    traverseHeaderFooter(style.getHeader(),sLang);
                }
                preamble.append("}%").nl();
                preamble.append("\\renewcommand\\@evenhead{");
                if (style.getHeaderLeft()!=null) {
                    traverseHeaderFooter(style.getHeaderLeft(),sLang);
                }
                else if (style.getHeader()!=null) {
                    preamble.append("\\@oddhead");
                }
                preamble.append("}%").nl();
                // Footer
                preamble.append("\\renewcommand\\@oddfoot{");
                if (style.getFooter()!=null) {
                    traverseHeaderFooter(style.getFooter(),sLang);
                }
                preamble.append("}%").nl();
                preamble.append("\\renewcommand\\@evenfoot{");
                if (style.getFooterLeft()!=null) {
                    traverseHeaderFooter(style.getFooterLeft(),sLang);
                }
                else if (style.getFooter()!=null) {
                    preamble.append("\\@oddfoot");
                }
                preamble.append("}%").nl();
                // Sectionmark and subsectionmark
                if (sChapterField1!=null) {
                    preamble.append("\\def\\sectionmark##1{\\markboth{");
                    if ("name".equals(sChapterField1)) { preamble.append("##1"); }
                    else if ("number".equals(sChapterField1) || "plain-number".equals(sChapterField1)) {
                        preamble.append("\\thesection");
                    }
                    else { preamble.append("\\thesection ##1"); }
                    preamble.append("}{}}").nl();
                }
                if (sChapterField2!=null) {
                    if (sChapterField1==null) {
                        preamble.append("\\def\\sectionmark##1{\\markboth{}{}}").nl();
                    }
                    preamble.append("\\def\\subsectionmark##1{\\markright{");
                    if ("name".equals(sChapterField2)) { preamble.append("##1"); }
                    else if ("number".equals(sChapterField2) || "plain-number".equals(sChapterField1)) {
                        preamble.append("\\thesubsection");
                    }
                    else { preamble.append("\\thesubsection ##1"); }
                    preamble.append("}{}}").nl();
                }
                // Convert PageMaster
                String sPageMaster = style.getProperty(XMLString.STYLE_PAGE_MASTER_NAME);
                sconv.convertPageMaster(sPageMaster,preamble);

                preamble.append("}").nl();
            }
        }
        preamble.append("\\makeatother").nl();
    }

    private void traverseHeaderFooter(Node node, String sLang) {
        // get first paragraph; all other content is ignored
        Node par = getChildByTagName(node,XMLString.TEXT_P);
        if (par==null) { return; }
		
        String sStyleName = getAttribute(par, XMLString.TEXT_STYLE_NAME);
        StyleWithProperties style = sc.getParStyle(sStyleName);
        // Temp solution: Ignore hard formatting in header/footer (name clash problem)
        // only in package format. TODO: Find a better solution for 0.4!
        if (sxwDoc.getStyleDOM()!=null && style!=null && style.isAutomatic()) {
            style = null;
        }

        if (par.hasChildNodes()) {
            BeforeAfter ba = new BeforeAfter();
            if (style!=null) { sconv.applyHardCharFormatting(style,ba); }
            preamble.append(ba.getBefore());
            traverseInlineText(par,false,true,preamble,sLang);
            preamble.append(ba.getAfter());
        }
        
    }

    /*
     * Process the office:body tag. This is traversed in logical order.
     * Note: All handling of block level elements should add a newline to the
     * LaTeX document at the end.
     */
    private void handleBody(Node node) {
        body.append("\\begin{document}").nl();
        // Get language from default paragraph style:
        StyleWithProperties style = sc.getDefaultParStyle();
        String sLang = style!=null ? style.getProperty(XMLString.FO_LANGUAGE) : "";
        traverseBlockText(node,false,sLang);
        if (bContainsEndnotes) {
            body.append("\\clearpage").nl()
                .append("\\theendnotes").nl();
        }
        body.append("\\end{document}").nl();
    }

    private void handleSection(Node node, String sLang) {
        String sName = getAttribute(node,XMLString.TEXT_NAME);
        addTarget(node,"%7Cregion",body,sLang);
        String sStyleName = getAttribute(node,XMLString.TEXT_STYLE_NAME);
        SectionStyle style = sc.getSectionStyle(sStyleName);
        if (style!=null && style.getColCount()>1) {
            bNeedMulticol = true;
            nMulticolLevel++;
            body.append("\\begin{multicols}{"+style.getColCount()+"}").nl();
        }
        traverseBlockText(node,false,sLang);
        if (style!=null && style.getColCount()>1) {
            body.append("\\end{multicols}").nl();
            nMulticolLevel--;
        }
    }


    private void addDebugInfo(Node node, String sAttribute) {
        String sValue=Misc.getAttribute(node,sAttribute);
        if (sValue!=null) {
            body.append(" ").append(sAttribute).append("=\"").append(sValue).append("\"");
        }
    }

    private void addDebugInfo(Node node) {	
        if (config.debug()) {
            body.append("% ").append(node.getNodeName());
            addDebugInfo(node,XMLString.TEXT_ID);
            addDebugInfo(node,XMLString.TEXT_NAME);
            addDebugInfo(node,XMLString.TABLE_NAME);
            addDebugInfo(node,XMLString.TEXT_STYLE_NAME);
            body.nl();
        }
    }

    private void traverseBlockText(Node node, boolean bInTable, String sLang) {
        // The current paragraph block:
        StyleMap blockMap = config.getParBlockStyleMap();
        String sBlockName = null;

        if (node.hasChildNodes()) {
            NodeList nList = node.getChildNodes();
            int len = nList.getLength();
            
            for (int i = 0; i < len; i++) {
                Node child = nList.item(i);
				
                if (child.getNodeType() == Node.ELEMENT_NODE) {
                    String nodeName = child.getNodeName();

                    addDebugInfo(child);
                
                    // Start/End a paragraph block
                    if (nodeName.equals(XMLString.TEXT_P)) {
                        String sStyleName = getAttribute(child, XMLString.TEXT_STYLE_NAME);
                        if (sBlockName!=null && !blockMap.isNext(sBlockName,sStyleName)) {
                            // end current block
                            String sAfter = blockMap.getAfter(sBlockName);
                            if (sAfter.length()>0) body.append(sAfter).nl();
                            sBlockName = null;
                        }
                        if (sBlockName==null && blockMap.contains(sStyleName)) {
                            // start a new block
                            sBlockName = sStyleName;
                            String sBefore = blockMap.getBefore(sBlockName);
                            if (sBefore.length()>0) body.append(sBefore).nl();
                        }
                    }
                    else if (sBlockName!=null) {
                        // non-paragraph: end current block
                        String sAfter = blockMap.getAfter(sBlockName);
                        if (sAfter.length()>0) body.append(sAfter).nl();
                        sBlockName = null;
                    }
					
                    if (nodeName.startsWith("draw:")) {
                        handleDrawElement(child,false,body,sLang);
                    }
                    else if (nodeName.equals(XMLString.TEXT_P)) {
                        boolean bLastInCell = bInTable && (i==len-1);
                        handleParagraph(child,bLastInCell,sLang);
                    }

                    else if(nodeName.equals(XMLString.TEXT_H)) {
                        handleHeading(child,sLang);
                    }
                    
                    else if (nodeName.equals(XMLString.TEXT_UNORDERED_LIST)) {
                        handleList(child,null,1,false,false,sLang);
                    }
                    
                    else if (nodeName.equals(XMLString.TEXT_ORDERED_LIST)) {
                        handleList(child,null,1,true,false,sLang);
                    }
                    else if (nodeName.equals(XMLString.TABLE_TABLE)) {
                        handleTable(child,sLang);
                    }

                    else if (nodeName.equals(XMLString.TEXT_SECTION)) {
                        handleSection(child,sLang);
                    }

                    else if (nodeName.equals(XMLString.TEXT_TABLE_OF_CONTENT)) {
                        handleTOC(child,sLang);
                    }

                    else if (nodeName.equals(XMLString.TEXT_ILLUSTRATION_INDEX)) {
                        handleLOF(child,sLang);
                    }

                    else if (nodeName.equals(XMLString.TEXT_TABLE_INDEX)) {
                        handleLOT(child,sLang);
                    }

                    else if (nodeName.equals(XMLString.TEXT_OBJECT_INDEX)) {
                        handleObjectIndex(child,sLang);
                    }

                    else if (nodeName.equals(XMLString.TEXT_USER_INDEX)) {
                        handleUserIndex(child,sLang);
                    }

                    else if (nodeName.equals(XMLString.TEXT_ALPHABETICAL_INDEX)) {
                        handleAlphabeticalIndex(child,sLang);
                    }

                    else if (nodeName.equals(XMLString.TEXT_BIBLIOGRAPHY)) {
                        handleBibliography(child,sLang);
                    }

                    else if (nodeName.equals(XMLString.TABLE_SUB_TABLE)) {
                        handleTable(child,sLang);
                    }

                    else if (nodeName.equals(XMLString.TEXT_SEQUENCE_DECLS)) {
                        //handleSeqeuenceDecls(child);
                    }
                    // other tags are ignored
                }
            }
        }

        if (sBlockName!=null) {
            // end current block
            String sAfter = blockMap.getAfter(sBlockName);
            if (sAfter.length()>0) body.append(sAfter).nl();
            sBlockName = null;
        }

    }
    
    /*
     * Process a table:table tag or a table:sub-table tag
     */
    private void handleTable(Node node,String sLang) {
        TableGridModel table = new TableGridModel(node);
        TableFormatter formatter = sconv.getTableFormatter(table, nMulticolLevel==0);

        bContainsTables = true;
        bNeedLongtable |= formatter.isLongtable();

        if (!formatter.isSubtable()) {
            addTarget(node,"%7Ctable",body,sLang);
        }
		
        // export table declaration
        BeforeAfter baTable = new BeforeAfter();
        formatter.applyTableStyle(baTable);
        body.append(baTable.getBefore()).nl();
		
        // export rows
        if (!formatter.isLongtable()) { enterNoFootnotesArea(); }
        int nRowCount = table.getRowCount();
        int nColCount = table.getColCount();
        boolean bInHeader = false;
        // add interrow material before first row:
        body.append(formatter.getInterrowMaterial(0)).nl();
        for (int nRow=0; nRow<nRowCount; nRow++){
            // longtables may have headers:
            if (formatter.isLongtable()) {
                if (nRow==0 && table.getRow(nRow).isHeader()) {
                    bInHeader = true;
                    enterNoFootnotesArea();
                }
                else if (bInHeader && !table.getRow(nRow).isHeader()) {
                    bInHeader = false;
                    leaveNoFootnotesArea();
                    body.append("\\endhead").nl();
                }
            }
            
            // export columns in this row
            int nCol = 0;
            while (nCol<nColCount) {
                Node cell = table.getCell(nRow,nCol);
                if (XMLString.TABLE_TABLE_CELL.equals(cell.getNodeName())) {
                    BeforeAfter baCell = new BeforeAfter();
                    formatter.applyCellStyle(nRow,nCol,baCell);
                    body.append(baCell.getBefore());
                    traverseBlockText(cell,true,sLang);
                    body.append(baCell.getAfter());
                }
                // Otherwise ignore; the cell is covered by a \multicolumn entry.
                // (table:covered-table-cell)
                int nColSpan = Misc.getPosInteger(getAttribute(cell,
                                   XMLString.TABLE_NUMBER_COLUMNS_SPANNED),1);
                if (nCol+nColSpan<nColCount) { body.append("&").nl(); }
                nCol+=nColSpan;
            }
            body.append("\\\\").append(formatter.getInterrowMaterial(nRow+1)).nl();
        }
        if (!formatter.isLongtable()) { leaveNoFootnotesArea(); }

        // end table
        body.append(baTable.getAfter()).nl();
    }
	
    private void collectMasterPage(ParStyle style) {
        if (style==null || sMainMasterPage!=null) { return; }
        String s = style.getMasterPageName();
        if (s!=null && s.length()!=0) { sMainMasterPage = s; }
    }


    /* Check to see if this node contains character formatting (text:span) */
    private boolean containsSpan(Node node) {
        if (!node.hasChildNodes()) { return false; }
        NodeList list = node.getChildNodes();
        int nLen = list.getLength();
        for (int i = 0; i < nLen; i++) {
            Node child = list.item(i);
            if (child.getNodeType()==Node.ELEMENT_NODE
                && XMLString.TEXT_SPAN.equals(child.getNodeName())) {
                return true;
            }
        }
        return false;
    }
	
    /*
     * Process a text:h tag
     */
    private void handleHeading(Node node, String sLang) {
        // Note: Conditional styles are not supported
        int level = Misc.getPosInteger(getAttribute(node, XMLString.TEXT_LEVEL),1);
        HeadingMap hm = config.getHeadingMap();
        if (level<=hm.getMaxLevel()) {
            // Always push the font used
            String sStyleName = getAttribute(node, XMLString.TEXT_STYLE_NAME);
            i18n.pushSpecialTable(getFontName(sc.getParStyle(sStyleName)));

            BeforeAfter baHardPage = new BeforeAfter();
            BeforeAfter baHardChar = new BeforeAfter();
            ParStyle style = sc.getParStyle(sStyleName);
            if (style!=null) {
                if (style.isAutomatic()) {
                    sconv.applyHardPageBreak(style,baHardPage);
                    sconv.applyHardCharFormatting(style,baHardChar);                    
                    // The real stylename is the underlying soft style:
                    sStyleName = style.getParentName();
                }
                if (sHeadingStyles[level]==null) {
                    sHeadingStyles[level] = sStyleName;
                }
            }
            body.append(baHardPage.getBefore());
            body.append("\\"+hm.getName(level));
            bInHeading = true;
            // If this heading contains formatting, add special item for toc:
            if (baHardChar.getBefore().length()>0 || containsSpan(node)) {
                body.append("[");
                traversePlainInlineText(node,body,sLang);
                body.append("]");
            }
            body.append("{").append(baHardChar.getBefore());
            traverseInlineText(node,false,sLang);
            bInHeading = false;
            body.append(baHardChar.getAfter()).append("}").nl();
            body.append(baHardPage.getAfter());
            int n = hyperlinkTargets.size();
            for (int i=0; i<n; i++) {
                String sName = (String) hyperlinkTargets.get(i);
                body.append("\\hypertarget{")
                    .append(targets.getExportName(sName))
                    .append("}{}");
            }
            hyperlinkTargets.clear();

            if (style!=null && style.isAutomatic()) {
                // Apply page break; other hard formatting is ignored
                // (this really belongs in StyleConverter.java)
                String s = style.getProperty(XMLString.FO_BREAK_AFTER);
                if ("page".equals(s)){ body.append("\\clearpage").nl(); }
            }

            collectMasterPage(sc.getParStyle(sHeadingStyles[level]));	
            flushFloatingFrames();
			
            // pop special table
            i18n.popSpecialTable();
        }
        else { // beyond \subparagraph - export as ordinary paragraph
            handleParagraph(node,false,sLang);
        }
    }

    /*
     * Process a text:p tag
     */
    private void handleParagraph(Node node, boolean bLastInCell, String sLang) {
        // Always push the font used
        String sStyleName = getAttribute(node, XMLString.TEXT_STYLE_NAME);
        i18n.pushSpecialTable(getFontName(sc.getParStyle(sStyleName)));

        // Empty paragraphs are often used to achieve vertical spacing in WYSIWYG
        // word processors. Hence we translate an empty paragraph to \bigskip.
        // This also solves the problem that LaTeX ignores empty paragraphs, Writer doesn't.
        // In a well-structured document, an empty paragraph is probably a mistake,
        // hence the configuration can specify that it should be ignored.
        if (!node.hasChildNodes()) {
            if (!config.ignoreEmptyParagraphs()) {
                body.nl().append("\\bigskip").nl();
                if (!bLastInCell) { body.nl(); }
            }
            return;
        }
        // Otherwise use the style and export the content:
        // Note: Conditional styles are not supported
        BeforeAfter ba = new BeforeAfter();
        sconv.applyParStyle(sStyleName,ba);
        ParStyle style = sc.getParStyle(sStyleName);
        if (style!=null) { 
            collectMasterPage(style);
            String sLang1 = style.getProperty(XMLString.FO_LANGUAGE);
            if (sLang1!=null) { sLang = sLang1; }
        }

        if (ba.getBefore().length()>0) { body.append(ba.getBefore()).nl(); }
        traverseInlineText(node,false,sLang);
        if (ba.getAfter().length()>0) {
            body.nl().append(ba.getAfter()).nl();
            if (bLastInCell) {
                // A \par right before the end of a table cell will add an
                // empty paragraph. We have to compensate for that
                body.append("\\vspace{-\\baselineskip}").nl();
            }
        }
        else if (bLastInCell) { body.nl(); }
        else { body.nl().nl(); }
		
        flushFloatingFrames();

        // pop the font name
        i18n.popSpecialTable();
    }
    
    /* -----------------------------------------------------------------
       handle lists
     ------------------------------------------------------------------- */
	
    /*
     * Check to see, if this list contains headings
     * (in that case we will ignore the list!)  
     */
    private boolean listContainsHeadings (Node node) {
        if (node.hasChildNodes()) {
            NodeList nList = node.getChildNodes();
            int len = nList.getLength();
            for (int i = 0; i < len; i++) {
                Node child = nList.item(i);
                if (child.getNodeType() == Node.ELEMENT_NODE) {
                    String nodeName = child.getNodeName();
                    if (nodeName.equals(XMLString.TEXT_LIST_ITEM)) {
                        if (listItemContainsHeadings(child)) return true;
                    }
                    if (nodeName.equals(XMLString.TEXT_LIST_HEADER)) {
                        if (listItemContainsHeadings(child)) return true;
                    }
                }
            }
        }
        return false;
    }
    
    private boolean listItemContainsHeadings(Node node) {
        if (node.hasChildNodes()) {
            NodeList nList = node.getChildNodes();
            int len = nList.getLength();
            for (int i = 0; i < len; i++) {
                Node child = nList.item(i);
                if (child.getNodeType() == Node.ELEMENT_NODE) {
                    String nodeName = child.getNodeName();
                    if(nodeName.equals(XMLString.TEXT_H)) {
                        return true;
                    }
                    if (nodeName.equals(XMLString.TEXT_ORDERED_LIST)) {
                        if (listContainsHeadings(child)) return true;
                    }
                    if (nodeName.equals(XMLString.TEXT_UNORDERED_LIST)) {
                        if (listContainsHeadings(child)) return true;
                    }
                }
            }
        }
        return false;
    }

    // Proces a list (ordered or unordered)
    private void handleList(Node node, String sStyleName, int nLevel,
                            boolean bOrdered, boolean bIgnore, String sLang) {
        // If the list contains headings, ignore it!        
        if (bIgnore || listContainsHeadings(node)) {
            traverseList(node,sStyleName,nLevel,true,sLang);
            return;
        }

        if (sStyleName==null) {
            sStyleName = Misc.getAttribute(node,XMLString.TEXT_STYLE_NAME);
        }
        BeforeAfter ba = new BeforeAfter();
        sconv.applyListStyle(sStyleName,nLevel,bOrdered,
            "true".equals(getAttribute(node,XMLString.TEXT_CONTINUE_NUMBERING)),
            ba);
        if (ba.getBefore().length()>0) { body.append(ba.getBefore()).nl(); }
        traverseList(node,sStyleName,nLevel,bIgnore,sLang);
        if (ba.getAfter().length()>0) { body.append(ba.getAfter()).nl(); }
    }

    /*
     * Process the contents of a list
     */
    private void traverseList (Node node, String sStyleName, int level,
                               boolean bIgnore,String sLang) {
        
        if (node.hasChildNodes()) {
            NodeList nList = node.getChildNodes();
            int len = nList.getLength();
            
            for (int i = 0; i < len; i++) {
                Node child = nList.item(i);
                
                if (child.getNodeType() == Node.ELEMENT_NODE) {
                    String nodeName = child.getNodeName();
					
                    addDebugInfo(child);
                    
                    if (nodeName.equals(XMLString.TEXT_LIST_ITEM)) {
                        handleListItem(child,sStyleName,level,bIgnore,sLang);
                    }
                    if (nodeName.equals(XMLString.TEXT_LIST_HEADER)) {
                        handleListItem(child,sStyleName,level,bIgnore,sLang);
                    }
                }
            }
        }
    }
    
    private void handleListItem(Node node, String sStyleName, int nLevel,
                                boolean bIgnore, String sLang) {
        // Are we ignoring this list?
        if (bIgnore) {
            traverseListItem(node,sStyleName,nLevel,bIgnore,sLang);
            return;
        }
        
        BeforeAfter ba = new BeforeAfter();
        sconv.applyListItemStyle(sStyleName,nLevel,
            node.getNodeName().equals(XMLString.TEXT_LIST_HEADER),
            "true".equals(Misc.getAttribute(node,XMLString.TEXT_RESTART_NUMBERING)),
            Misc.getPosInteger(getAttribute(node,XMLString.TEXT_START_VALUE),1)-1,
            ba);
        if (ba.getBefore().length()>0) { body.append(ba.getBefore()).nl(); }
        traverseListItem(node,sStyleName,nLevel,bIgnore,sLang);
        if (ba.getAfter().length()>0) { body.append(ba.getAfter()).nl(); }
    }
    
    /*
     * Process the contents of a list item
     * (a list header should only contain paragraphs, but we don't care)
     */
    private void traverseListItem (Node node, String sStyleName, int level, boolean bIgnore, String sLang) {
        if (node.hasChildNodes()) {
            NodeList nList = node.getChildNodes();
            int len = nList.getLength();
            
            for (int i = 0; i < len; i++) {
                Node child = nList.item(i);
                
                if (child.getNodeType() == Node.ELEMENT_NODE) {
                    String nodeName = child.getNodeName();

                    addDebugInfo(child);
                    
                    if (nodeName.equals(XMLString.TEXT_P)) {
                        handleParagraph(child,false,sLang);
                    }
					
                    else if(nodeName.equals(XMLString.TEXT_H)) {
                        handleHeading(child,sLang);
                    }
                    
                    if (nodeName.equals(XMLString.TEXT_ORDERED_LIST)) {
                        handleList(child,sStyleName,level+1,true,bIgnore,sLang);
                    }
                    
                    if (nodeName.equals(XMLString.TEXT_UNORDERED_LIST)) {
                        handleList(child,sStyleName,level+1,false,bIgnore,sLang);
                    }
                }
            }
        }
    }
    
    /* Process table of contents
       TODO (0.5): Apply more formatting by modfification of \l@section etc.
       Something like this:
       \newcommand\l@section[2]{\@dottedtocline{1}{1.5em}{2.3em}{\textbf{#1}}{\textit{#2}}
       Textformatting is trivial; see article.cls for examples of more complicated
       formatting. Note: The section number can't be formatted indivdually.
     */
    private void handleTOC (Node node, String sLang) {
        Node source = getChildByTagName(node,XMLString.TEXT_TABLE_OF_CONTENT_SOURCE);
        if (source!=null) {
            if ("chapter".equals(getAttribute(source,XMLString.TEXT_INDEX_SOURCE))) {
                 body.append("[Warning: Table of content (for this chapter) ignored!]").nl().nl();
            }
            else {
                int nLevel = Misc.getPosInteger(getAttribute(source,XMLString.TEXT_OUTLINE_LEVEL),1);
                body.append("\\setcounter{tocdepth}{"+nLevel+"}").nl();
                Node title = getChildByTagName(source,XMLString.TEXT_INDEX_TITLE_TEMPLATE);
                if (title!=null) {
                    body.append("\\renewcommand\\contentsname{");
                    traversePCDATA(title,body,sLang);
                    body.append("}").nl();
                }
            }
        }
        body.append("\\tableofcontents").nl();
    }

    /*
     * Process list of illustrations
     */
    private void handleLOF (Node node, String sLang) {
        body.append("[Warning: List of illustrations ignored]").nl().nl();
    }

    /*
     * Process list of tables
     */
    private void handleLOT (Node node, String sLang) {
        body.append("[Warning: List of tables ignored]").nl().nl();
    }

    /*
     * Process Object index
     */
    private void handleObjectIndex (Node node, String sLang) {
        body.append("[Warning: Object index ignored]").nl().nl();
    }

    /*
     * Process User index
     */
    private void handleUserIndex (Node node, String sLang) {
        body.append("[Warning: User index ignored]").nl().nl();
    }

    /*
     * Process Alphabetical index
     */
    private void handleAlphabeticalIndex (Node node, String sLang) {
        body.append("\\printindex").nl();
        bContainsAlphabeticalIndex = true;
    }

    /*
     * Process Bibliography
     */
    private void handleBibliography (Node node, String sLang) {
        if (config.useBibtex()) {
            body.append("\\bibliographystyle{")
                .append(config.bibtexStyle())
                .append("}").nl();
            body.append("\\bibliography{")
                .append(bibDoc.getName())
                .append("}").nl();
        }
        else { // typeset current content
            Node body = getChildByTagName(node,XMLString.TEXT_INDEX_BODY);
            if (body!=null) {
                Node title = getChildByTagName(body,XMLString.TEXT_INDEX_TITLE);
                if (title!=null) { traverseBlockText(title,false,sLang); }
                traverseBlockText(body,false,sLang);
            }
        }     
    }

    /*
     * Process inline text
     */
    private void traverseInlineText (Node node, boolean styled, String sLang) {
        traverseInlineText(node,styled,false,body,sLang);
    }
	        
    private void traverseInlineText (Node node,boolean styled,boolean bInHeaderFooter,
        LaTeXDocumentPortion ldp, String sLang) {        
		
        boolean bNoFootnotes = false;
        
        String styleName = getAttribute(node, XMLString.TEXT_STYLE_NAME);
		
        // Always push the font used
        i18n.pushSpecialTable(getFontName(sc.getTextStyle(styleName)));
		
        BeforeAfter ba = new BeforeAfter();
        sconv.applyTextStyle(styleName,ba);
		
        // Temp solution: Ignore hard formatting in header/footer (name clash problem)
        // only in package format.
        StyleWithProperties style = sc.getTextStyle(styleName);

        if (style!=null) {
            String sLang1 = style.getProperty(XMLString.FO_LANGUAGE);
            if (sLang1!=null) { sLang = sLang1; }
            // Cannot have footnotes in sub/superscript
            String s = style.getProperty(XMLString.STYLE_TEXT_POSITION);
            if (s!=null && !bProcessingFootnote) { bNoFootnotes = true; }
        }

        if (sxwDoc.getStyleDOM()!=null && style!=null && style.isAutomatic() && bInHeaderFooter) {
            styled = false;
        }

        if (styled) {    
            if (bNoFootnotes) { enterNoFootnotesArea(); }
            ldp.append(ba.getBefore());
        }
                              
        if (node.hasChildNodes()) {
            NodeList nList = node.getChildNodes();
            int len = nList.getLength();
                       
            for (int i = 0; i < len; i++) {
                
                Node child = nList.item(i);
                short nodeType = child.getNodeType();
               
                switch (nodeType) {
                    case Node.TEXT_NODE:
                        String s = child.getNodeValue();
                        if (s.length() > 0) {
                            ldp.append(i18n.convert(s,false,sLang));
                        }
                        break;
                        
                    case Node.ELEMENT_NODE:
                        String sName = child.getNodeName();
                        if (child.getNodeName().startsWith("draw:")) {
                            handleDrawElement(child,bInHeaderFooter,ldp,sLang);
                        }
                        else if (child.getNodeName().equals(XMLString.TEXT_S)) {
                            int count= Misc.getPosInteger(getAttribute(child,XMLString.TEXT_C),1);
                            for ( ; count > 0; count--) {
                                ldp.append("\\ ");
                            }
                        }
                        else if (sName.equals(XMLString.TEXT_TAB_STOP)) {
                            // tab stops are not supported by the onverter, but the special usage
                            // of tab stops in header and footer can be emulated with \hfill
                            if (bInHeaderFooter) { ldp.append("\\hfill "); }
                            else { ldp.append(" "); }
                        }
                        else if (sName.equals(XMLString.TEXT_LINE_BREAK)) {
                            if (bInHeaderFooter) { ldp.append("\\newline").nl(); }
                            else { ldp.append(" "); } // no line breaks in header and footer!
                        }
                        else if (sName.equals(XMLString.TEXT_SPAN)) {
                            traverseInlineText (child,true,bInHeaderFooter,ldp,sLang);
                        }
                        else if (sName.equals(XMLString.TEXT_A)) {
                            handleAnchor(child,bInHeaderFooter,ldp,sLang);
                        }
                        else if (sName.equals(XMLString.OFFICE_ANNOTATION)) {
                            handleOfficeAnnotation(child,ldp,sLang);
                        }
                        else if (sName.equals(XMLString.TEXT_PAGE_NUMBER)) {
                            handlePageNumber(child,ldp,sLang);
                        }
                        else if (sName.equals(XMLString.TEXT_PAGE_COUNT)) {
                            handlePageCount(child,ldp,sLang);
                        }
                        else if (bInHeaderFooter) {
                            if (sName.equals(XMLString.TEXT_CHAPTER)) {
                                int nLevel = Misc.getPosInteger(Misc.getAttribute(child,XMLString.TEXT_OUTLINE_LEVEL),1);
		                        if (nLevel==1) {
		                            sChapterField1=Misc.getAttribute(child,XMLString.TEXT_DISPLAY);
                                    ldp.append("{\\leftmark}");
                                }
		                        else if (nLevel==2) {
		                            sChapterField2=Misc.getAttribute(child,XMLString.TEXT_DISPLAY);
                                    ldp.append("{\\rightmark}");
                                }
                            }
                            else if (sName.startsWith("text:")) {
                                traverseInlineText(child,false,bInHeaderFooter,ldp,sLang);
                            }
                        }
                        else {
	                        // These tags are ignored in header and footer
                            if (sName.equals(XMLString.TEXT_FOOTNOTE)) {
                                handleFootnote(child,sLang);
                            }
                            else if (sName.equals(XMLString.TEXT_ENDNOTE)) {
                                handleEndnote(child,sLang);
                            }
                            else if (sName.equals(XMLString.TEXT_SEQUENCE)) {
	                            handleSequence(child,sLang);
                            }
                            else if (sName.equals(XMLString.TEXT_SEQUENCE_REF)) {
	                            handleSequenceRef(child,sLang);
                            }
                            else if (sName.equals(XMLString.TEXT_FOOTNOTE_REF)) {
	                            handleFootnoteRef(child,sLang);
                            }
                            else if (sName.equals(XMLString.TEXT_ENDNOTE_REF)) {
                                handleEndnoteRef(child,sLang);
                            }
                            else if (sName.equals(XMLString.TEXT_REFERENCE_MARK)) {
	                            handleReferenceMark(child,sLang);
                            }
                            else if (sName.equals(XMLString.TEXT_REFERENCE_MARK_START)) {
                                handleReferenceMark(child,sLang);
                            }
                            else if (sName.equals(XMLString.TEXT_REFERENCE_REF)) {
	                            handleReferenceRef(child,sLang);
                            }
                            else if (sName.equals(XMLString.TEXT_BOOKMARK)) {
                                handleBookmark(child,sLang);
                            }
                            else if (sName.equals(XMLString.TEXT_BOOKMARK_START)) {
	                            handleBookmark(child,sLang);
                            }
                            else if (sName.equals(XMLString.TEXT_BOOKMARK_REF)) {
                                handleBookmarkRef(child,sLang);
                            }
                            else if (sName.equals(XMLString.TEXT_BIBLIOGRAPHY_MARK)) {
                                handleBibliographyMark(child,ldp,sLang);
                            }
                            else if (sName.equals(XMLString.TEXT_ALPHABETICAL_INDEX_MARK)) {
                                handleAlphabeticalIndexMark(child,ldp,sLang);
                            }
                            else if (sName.equals(XMLString.TEXT_ALPHABETICAL_INDEX_MARK_START)) {
                                handleAlphabeticalIndexMark(child,ldp,sLang);
                            }
                            else if (sName.startsWith("text:")) {
                                traverseInlineText(child,false,bInHeaderFooter,ldp,sLang);
                            }
                        }
                        // lots of stuff missing.....  
                        break;
                    default:
                        // Do nothing
                }
            }
        }
        
        if (styled) {    
            ldp.append(ba.getAfter());
            if (bNoFootnotes) { leaveNoFootnotesArea();}
        }
		
        // finally pop the special table
        i18n.popSpecialTable();
    }
	
    /* traverse inline text, ignoring any draw objects, footnotes, formatting and hyperlinks */
    private void traversePlainInlineText(Node node,LaTeXDocumentPortion ldp, String sLang) {        
        if (node.hasChildNodes()) {
		
            // Always push the font used
            String styleName = getAttribute(node, XMLString.TEXT_STYLE_NAME);
            i18n.pushSpecialTable(getFontName(sc.getTextStyle(styleName)));

            NodeList nList = node.getChildNodes();
            int len = nList.getLength();
                       
            for (int i = 0; i < len; i++) {
                
                Node child = nList.item(i);
                short nodeType = child.getNodeType();
               
                switch (nodeType) {
                    case Node.TEXT_NODE:
                        String s = child.getNodeValue();
                        if (s.length() > 0) {
                            ldp.append(i18n.convert(s,false,sLang));
                        }
                        break;
                        
                    case Node.ELEMENT_NODE:
                        String sName = child.getNodeName();
                        if (child.getNodeName().equals(XMLString.TEXT_S)) {
                            int count= Misc.getPosInteger(getAttribute(child,XMLString.TEXT_C),1);
                            for ( ; count > 0; count--) {
                                ldp.append("\\ ");
                            }
                        }
                        else if (sName.equals(XMLString.TEXT_TAB_STOP)) {
                            // tab stops are not supported by the onverter
                            ldp.append(" ");
                        }
                        else if (sName.equals(XMLString.TEXT_FOOTNOTE)) {
                            // ignore
                        }
                        else if (sName.equals(XMLString.TEXT_ENDNOTE)) {
                            // ignore
                        }
                        else if (sName.startsWith("text:")) {
                            traversePlainInlineText(child,ldp,sLang);
                        }
                        break;
                    default:
                        // Do nothing
                }
            }
            // finally pop the special table
            i18n.popSpecialTable();
        }
    }

    private void traversePCDATA(Node node, LaTeXDocumentPortion ldp, String sLang) {
        if (node.hasChildNodes()) {
            NodeList nl = node.getChildNodes();
            int nLen = nl.getLength();
            for (int i=0; i<nLen; i++) {
                if (nl.item(i).getNodeType()==Node.TEXT_NODE) {
                    ldp.append(i18n.convert(nl.item(i).getNodeValue(),false,sLang));
                }
            }
        }
    }
	
    private void addTarget(Node node, String sSuffix, LaTeXDocumentPortion ldp, String sLang) {
        String sName = Misc.getAttribute(node,XMLString.TEXT_NAME);
        if (sName == null) { sName = Misc.getAttribute(node,XMLString.TABLE_NAME); }
        if (sName==null || !config.useHyperref()) { return; }
        if (bInHeading) {
            hyperlinkTargets.add(targets.getExportName(sName+sSuffix));
        }
        else {
            ldp.append("\\hypertarget{")
               .append(targets.getExportName(sName+sSuffix))
               .append("}{}");
        }
    }
    
    private void handleAnchor(Node node, boolean bInHeaderFooter, LaTeXDocumentPortion ldp, String sLang) {
        String sHref = getAttribute(node,XMLString.XLINK_HREF);
        if (sHref!=null) {
            if (sHref.startsWith("#")) {
                // TODO: hyperlinks to headings (?) and objects
			    if (config.useHyperref()) {
                    ldp.append("\\hyperlink{")
                       .append(targets.getExportName(sHref.substring(1)))
                       .append("}{");
                    // ignore text style (let hyperref.sty handle the decoration):
                    traverseInlineText(node,false,bInHeaderFooter,ldp,sLang);
                    ldp.append("}");
                }
                else { // user don't want to include hyperlinks
                    traverseInlineText(node,true,bInHeaderFooter,ldp,sLang);
                }
            }
			else {
			    if (config.useHyperref()) {
                    ldp.append("\\href{").append(sHref).append("}{");
                    // ignore text style (let hyperref.sty handle the decoration):
                    traverseInlineText(node,false,bInHeaderFooter,ldp,sLang);
                    ldp.append("}");
                }
                else { // user don't want to include hyperlinks
                    traverseInlineText(node,true,bInHeaderFooter,ldp,sLang);
                }
            }
        }
        else {
            traverseInlineText(node,true,bInHeaderFooter,ldp,sLang);
        }
    }

    /* Process a footnote */
    private void handleFootnote(Node node, String sLang) {
        String sId = getAttribute(node, XMLString.TEXT_ID);
        if (!footnotenames.containsName(sId)) { footnotenames.addName(sId); }
        Node fntbody = getChildByTagName(node,XMLString.TEXT_FOOTNOTE_BODY);
        if (fntbody != null) {
            bContainsFootnotes = true;
            if (nNoFootnotesLevel>0) {
                body.append("\\footnotemark{}");
                postponedFootnotes.add(fntbody);
            }
            else {
                body.append("\\footnote");
	    		body.append("{");
		        if (sId != null) {
                    body.append("\\label{fnt:"+footnotenames.getExportName(sId)+"}");
                }
            	traverseNoteBody(fntbody,sLang);
                body.append("}");
            }
        }
	} 
	
    // Enter an area, where footnotes should be avoided
    private void enterNoFootnotesArea() {
        nNoFootnotesLevel++;
    }

    // Leave an area, where footnotes should be avoided
    private void leaveNoFootnotesArea() {
	    // todo: fix language
        if (--nNoFootnotesLevel == 0) {
            // back in business! Type out all postponed footnotes:
            int n = postponedFootnotes.size();
            if (n==1) {
                body.append("\\footnotetext{");
                traverseNoteBody((Node) postponedFootnotes.get(0),"en");
                body.append("}").nl();
                postponedFootnotes.clear();
            }
            else if (n>1) {
                // Several footnotes; have to adjust the footnote counter
                body.append("\\addtocounter{footnote}{-"+n+"}").nl();
                for (int i=0; i<n; i++) {
                    body.append("\\stepcounter{footnote}\\footnotetext{");
                    traverseNoteBody((Node) postponedFootnotes.get(i),"en");
                    body.append("}").nl();
                }
                postponedFootnotes.clear();
            }
        }
    }
	
    /* Process an endnote */
    private void handleEndnote(Node node, String sLang) {
        String sId = getAttribute(node, XMLString.TEXT_ID);
        if (!endnotenames.containsName(sId)) { endnotenames.addName(sId); }
        Node entbody = getChildByTagName(node,XMLString.TEXT_ENDNOTE_BODY);
        if (entbody != null) {
            if (nNoFootnotesLevel>0 && !config.useEndnotes()) {
                body.append("\\footnotemark()");
                postponedFootnotes.add(entbody);
            }
            else {
                if (config.useEndnotes()) {
                    body.append("\\endnote");
                    bContainsEndnotes = true;
                }
                else {
                    body.append("\\footnote");
                    bContainsFootnotes = true;
                }
	            body.append("{");
                if (sId != null) {
			        body.append("\\label{ent:"+endnotenames.getExportName(sId)+"}");
                }
                traverseNoteBody(entbody,sLang);
                body.append("}");
            }
		}
	} 

	/*
     * Process the contents of a footnote or endnote
     */
    private void traverseNoteBody (Node node, String sLang) {
        bProcessingFootnote = true;
        if (node.hasChildNodes()) {
            NodeList nList = node.getChildNodes();
            int len = nList.getLength();
            
            for (int i = 0; i < len; i++) {
                Node child = nList.item(i);
                
                if (child.getNodeType() == Node.ELEMENT_NODE) {
                    String nodeName = child.getNodeName();

                    addDebugInfo(child);
                    
                    if (nodeName.equals(XMLString.TEXT_H)) {
                        handleHeading(child,sLang);
                    }

                    if (nodeName.equals(XMLString.TEXT_P)) {
                        traverseInlineText(child,false,sLang);
                        if (i<len-1) {
                            if (nList.item(i+1).getNodeName().equals(XMLString.TEXT_P)) {
                                body.nl().nl();
                            }
                            else {
                                body.nl();
                            }
                        }
                    }
                    
                    if (nodeName.equals(XMLString.TEXT_ORDERED_LIST)) {
                        handleList(child,null,0,true,false,sLang);
                    }
                    
                    if (nodeName.equals(XMLString.TEXT_UNORDERED_LIST)) {
                        handleList(child,null,0,false,false,sLang);
                    }
                }
            }
        }
        bProcessingFootnote = false;
    }

    private void handlePageNumber(Node node,LaTeXDocumentPortion ldp,String sLang) {
        // TODO: Obey attributes!
        ldp.append("\\thepage{}");
    }
    
    private void handlePageCount(Node node, LaTeXDocumentPortion ldp, String sLang) {
        // TODO: Obey attributes!
        ldp.append("\\thepagecount{}");
    }

    private void handleBibliographyMark(Node node, LaTeXDocumentPortion ldp, String sLang) {
         if (config.useBibtex()) {
             String sIdentifier = Misc.getAttribute(node,XMLString.TEXT_IDENTIFIER);
             if (!bibDoc.containsKey(sIdentifier)) {
                 bibDoc.put(new BibMark(node));
             }
             ldp.append("\\cite{")
                .append(bibDoc.getExportName(sIdentifier))
                .append("}");
         }
         else { // use current value
             traverseInlineText(node,false,false,ldp,sLang);
         }
    }
	
    private void writeIndexText(String sText, LaTeXDocumentPortion ldp, String sLang) {
        String sTextOut = i18n.convert(sText,false,sLang);
        //  need to escape !, @, | and ":
        int nLen = sTextOut.length();
        boolean bBackslash = false;
        for (int i=0; i<nLen; i++) {
            if (bBackslash) {
                ldp.append(sTextOut.substring(i,i+1));
                bBackslash = false;
            }
            else {
                switch (sTextOut.charAt(i)) {
                    case '\\' : bBackslash = true;
                                ldp.append("\\");
                                break;
                    case '!' :
                    case '@' :
                    case '|' :
                    case '"' : ldp.append("\"");
                    default : ldp.append(sTextOut.substring(i,i+1)); 
                }
            }
        }
    }

    private void handleAlphabeticalIndexMark(Node node, LaTeXDocumentPortion ldp, String sLang) {
        String sValue = IndexMark.getIndexValue(node);
        if (sValue!=null) {
            ldp.append("\\index{");
            String sKey1 = IndexMark.getKey1(node);
            if (sKey1!=null) {
                writeIndexText(sKey1,ldp,sLang);
                ldp.append("!");
            }
            String sKey2 = IndexMark.getKey2(node);
            if (sKey2!=null) {
                writeIndexText(sKey2,ldp,sLang);
                ldp.append("!");
            }
            writeIndexText(sValue,ldp,sValue);
            ldp.append("}");
        }
    }
	
    private void handleSequence(Node node, String sLang) {
        // Temp solution: Use current value, but support page references
        traversePCDATA(node,body,sLang);
        String sName = getAttribute(node,XMLString.TEXT_REF_NAME);
        if (sName!=null) {
            body.append("\\label{seq:"+seqrefnames.getExportName(sName)+"}");
        }        
    }
	
    private void handleSequenceRef(Node node, String sLang) {
        String sFormat = getAttribute(node,XMLString.TEXT_REFERENCE_FORMAT);
        String sName = getAttribute(node,XMLString.TEXT_REF_NAME);
        if ("page".equals(sFormat) && sName!=null) {
            body.append("\\pageref{seq:"+seqrefnames.getExportName(sName)+"}");
        }
        else { // use current value
            traversePCDATA(node,body,sLang);
        }
    } 

    private void handleFootnoteRef(Node node, String sLang) {
        String sFormat = getAttribute(node,XMLString.TEXT_REFERENCE_FORMAT);
        String sName = getAttribute(node,XMLString.TEXT_REF_NAME);
        if ("page".equals(sFormat) && sName!=null) {
            body.append("\\pageref{fnt:"+footnotenames.getExportName(sName)+"}");
        }
        else if ("text".equals(sFormat) && sName!=null) {
            body.append("\\ref{fnt:"+footnotenames.getExportName(sName)+"}");
        }
        else { // use current value
            traversePCDATA(node,body,sLang);
        }
    } 
        
    private void handleEndnoteRef(Node node, String sLang) {
        String sFormat = getAttribute(node,XMLString.TEXT_REFERENCE_FORMAT);
        String sName = getAttribute(node,XMLString.TEXT_REF_NAME);
        if ("page".equals(sFormat) && sName!=null) {
            body.append("\\pageref{ent:"+endnotenames.getExportName(sName)+"}");
        }
        else if ("text".equals(sFormat) && sName!=null) {
            body.append("\\ref{ent:"+endnotenames.getExportName(sName)+"}");
        }
        else { // use current value
            traversePCDATA(node,body,sLang);
        }
    } 

    private void handleReferenceMark(Node node, String sLang) {
        String sName = getAttribute(node,XMLString.TEXT_NAME);
        if (sName!=null) {
            body.append("\\label{ref:"+refnames.getExportName(sName)+"}");
        }
    }
	
    private void handleReferenceRef(Node node, String sLang) {
        String sFormat = getAttribute(node,XMLString.TEXT_REFERENCE_FORMAT);
        String sName = getAttribute(node,XMLString.TEXT_REF_NAME);
        if ("page".equals(sFormat) && sName!=null) {
            body.append("\\pageref{ref:"+refnames.getExportName(sName)+"}");
        }
        else { // use current value
            traversePCDATA(node,body,sLang);
        }
    } 

    private void handleBookmark(Node node, String sLang) {
        String sName = getAttribute(node,XMLString.TEXT_NAME);
        if (sName!=null) {
            addTarget(node,"",body,sLang);
            body.append("\\label{bkm:"+bookmarknames.getExportName(sName)+"}");
        }
    }
	
    private void handleBookmarkRef(Node node, String sLang) {
        String sFormat = getAttribute(node,XMLString.TEXT_REFERENCE_FORMAT);
        String sName = getAttribute(node,XMLString.TEXT_REF_NAME);
        if ("page".equals(sFormat) && sName!=null) {
            body.append("\\pageref{bkm:"+bookmarknames.getExportName(sName)+"}");
        }
        else { // use current value
            traversePCDATA(node,body,sLang);
        }
    } 
	
    ///////////////////////////////////////////////////////////////////////////
    // Handling of elements in the draw namespace
	
    private void handleDrawElement(Node node, boolean bInHeaderFooter, 
        LaTeXDocumentPortion ldp, String sLang) {
        // node must be an elment in the draw namespace
        String sName = node.getNodeName();
        if (sName.equals(XMLString.DRAW_OBJECT)) {
            handleDrawObject(node,ldp,sLang);
        }		
        else if ((!bInHeaderFooter) && sName.equals(XMLString.DRAW_IMAGE)) {
            handleDrawImage(node,sLang);
        }		
        else if ((!bInHeaderFooter) && sName.equals(XMLString.DRAW_TEXT_BOX)) {
            handleDrawTextBox(node,sLang);
        }		
        else if (sName.equals(XMLString.DRAW_A)) {
            // we handle this like text:a
            handleAnchor(node,bInHeaderFooter,ldp,sLang);
        }		
        else {
            // Other drawing objects (eg. shapes) are currently not supported
            ldp.append("[Warning: Draw object ignored]");
        }
    }
	
    //-----------------------------------------------------------------
    // handle draw:object elements (OOo objects such as Chart, Math,...)
		
    private void handleDrawObject(Node node, LaTeXDocumentPortion ldp, String sLang) {
        String sHref = Misc.getAttribute(node,XMLString.XLINK_HREF);
        if (sHref!=null) { // embedded or linked object
            EmbeddedObject object = getEmbeddedObject(sHref);
            if (object!=null) {
                if (MIMETypes.MATH.equals(object.getType())) {
                    try {
                        Document settings = ((EmbeddedXMLObject) object).getSettingsDOM();
                        Document formuladoc = ((EmbeddedXMLObject) object).getContentDOM();
                        Node formula = getChildByTagName(formuladoc,XMLString.MATH_MATH);
                        ldp.append(" $")
                           .append(mathmlconverter.convert(settings,formula))
                           .append("$ ");
                    }
                    catch (org.xml.sax.SAXException e) {
                        e.printStackTrace();
                    }
                    catch (java.io.IOException e) {
                        e.printStackTrace();
                    }
                }
            }
        }
        else { // flat xml, object is contained in node
            Node formula = getChildByTagName(node,XMLString.MATH_MATH);
            if (formula!=null) {
                ldp.append(" $")
                   .append(mathmlconverter.convert(null,formula))
                   .append("$ ");
            }
        }
    }
	
    //--------------------------------------------------------------------------
    // Handle draw:image elements
	
    private void handleDrawImage(Node node, String sLang) {
        String sName = getAttribute(node,XMLString.TEXT_NAME);
        addTarget(node,"%7Cgraphics",body,sLang);
        String sAnchor = Misc.getAttribute(node,XMLString.TEXT_ANCHOR_TYPE);
        if (nFrameLevel>0 || "as-char".equals(sAnchor)) {
            handleDrawImageAsChar(node,sLang);
        }
        else {
            floatingFrames.add(node);
        }
    }
	
    private void handleDrawImageAsChar(Node node,String sLang) {
        body.append(" ");
        includeGraphics(node,sLang);
        body.append(" ");
    }

    private void handleDrawImageFloat(Node node,String sLang) {
        body.append("\\begin{center}").nl();
        includeGraphics(node,sLang);
        body.nl().append("\\end{center}").nl();
    }

    private void includeGraphics(Node node, String sLang) {
        String sFileName = null;
        boolean bCommentOut = true;
        String sHref = Misc.getAttribute(node,XMLString.XLINK_HREF);
        if (sHref!=null && !sHref.startsWith("#")) {
            // Linked image is not yet handled by ImageLoader. This is a temp.
            // solution (will go away when ImageLoader is finished)
            sFileName = sHref;
            String sExt = sHref.substring(sHref.lastIndexOf(".")).toLowerCase();
            // Accept only relative filenames and supported filetypes:
            bCommentOut = sFileName.indexOf(":")>-1 || !(
                (config.getBackend()==config.PDFTEX && MIMETypes.JPEG_EXT.equals(sExt)) ||
                (config.getBackend()==config.PDFTEX && MIMETypes.PNG_EXT.equals(sExt)) ||
                (config.getBackend()==config.DVIPS && MIMETypes.EPS_EXT.equals(sExt)));
        }
        else { // embedded or base64 encoded image
            BinaryGraphicsDocument bgd = imageLoader.getImage(node);
            if (bgd!=null) {
                convertData.addDocument(bgd);
                sFileName = bgd.getFileName();
                String sMIME = bgd.getDocumentMIMEType();
                bCommentOut = !(
                    (config.getBackend()==config.PDFTEX && MIMETypes.JPEG.equals(sMIME)) ||
                    (config.getBackend()==config.PDFTEX && MIMETypes.PNG.equals(sMIME)) ||
                    (config.getBackend()==config.DVIPS && MIMETypes.EPS.equals(sMIME)));
            }
        }
		
        if (sFileName==null) {
            body.append("[Warning: Image not found]");
            return;
        }
		
        // Now for the actual inclusion:
        bNeedGraphicx = true;
        /* TODO (0.4): handle cropping and mirror:
           style:mirror kan vre none, vertical (lodret), horizontal (vandret),
           horizontal-on-odd, or
           horizontal-on-even (vandret p ulige hhv. lige side).
   		   mirror is handled with scalebox, eg:
        		%\scalebox{-1}[1]{...}
		   can check for even/odd page first!!
	
          fo:clip="rect(t,r,b,l) svarer til trim
          vrdien kan osse vre auto - dvs. ingen clip!
		  cropping is handled with clip and trim:
		  \includegraphics[clip,trim=l b r t]{...}
		  note the different order from xsl-fo!
         */

        String sWidth = Misc.getAttribute(node,XMLString.SVG_WIDTH);
        String sHeight = Misc.getAttribute(node,XMLString.SVG_HEIGHT);
        if (bCommentOut) {
            body.append(" [Warning: Image ignored] ");
            body.append("% Unhandled or unsupported graphics:").nl().append("%");
        }
        body.append("\\includegraphics");

        CSVList options = new CSVList(",");
        if (sWidth!=null) { options.addValue("width="+sWidth); }
        if (sHeight!=null) { options.addValue("height="+sHeight); }
        if (!options.isEmpty()) {
            body.append("[").append(options.toString()).append("]");
        }

        body.append("{").append(sFileName).append("}");
        if (bCommentOut) { body.nl(); }
    }

    //--------------------------------------------------------------------------
    // handle draw:text-box element
	
    private void handleDrawTextBox(Node node, String sLang) {
        String sName = getAttribute(node,XMLString.TEXT_NAME);
        addTarget(node,"%7Cframe",body,sLang);
        String sAnchor = Misc.getAttribute(node,XMLString.TEXT_ANCHOR_TYPE);
        if (!bInHeading && (nFrameLevel>0 || "as-char".equals(sAnchor))) {
            makeDrawTextBox(node, sLang);
        }
        else {
            floatingFrames.add(node);
        }
    }
	
    private void handleDrawTextBoxFloat(Node node, String sLang) {
        body.append("\\begin{center}").nl();
        makeDrawTextBox(node, sLang);
        body.append("\\end{center}").nl();
    }

    private void makeDrawTextBox(Node node, String sLang) {
        String sWidth = Misc.getAttribute(node,XMLString.SVG_WIDTH);
        enterNoFootnotesArea();
        nFrameLevel++;
        body.append("\\begin{minipage}{").append(sWidth).append("}").nl();
        traverseBlockText(node,false,sLang);
        body.append("\\end{minipage}");
        nFrameLevel--;
        leaveNoFootnotesArea();
    }
	
    //-------------------------------------------------------------------------
    //handle any pending floating frames
    
    private void flushFloatingFrames() {
	    // todo: fix language
        if (nFrameLevel>0) { return; }
        int n = floatingFrames.size();
        if (n==0) { return; }
        for (int i=0; i<n; i++) {
            Node node = (Node) floatingFrames.get(i);
            String sName = node.getNodeName();
            if (sName.equals(XMLString.DRAW_IMAGE)) {
                handleDrawImageFloat(node,"en");
            }
            else if (sName.equals(XMLString.DRAW_TEXT_BOX)) {
                handleDrawTextBoxFloat(node,"en");
            }
        }
        floatingFrames.clear();
    }
	
    ////////////////////////////////////////////////////////////////////
    // Annotations	
	
    private void handleOfficeAnnotation(Node node, LaTeXDocumentPortion ldp, String sLang) {
        ldp.append("%").nl().append("%");
        Node paragraph = getChildByTagName(node,XMLString.TEXT_P);
        if (paragraph!=null) { traversePCDATA(paragraph,ldp,sLang); }
        ldp.nl();
    }
	
    // utility method to get the font name from a StyleWithProperties
    private String getFontName(StyleWithProperties style) {
        if (style==null) { return null; }
        String sName = style.getProperty(XMLString.STYLE_FONT_NAME);
        if (sName==null) { return null; }
        FontDeclaration fd = sc.getFontDeclaration(sName);
        if (fd==null) { return null; }
        return fd.getProperty(XMLString.FO_FONT_FAMILY);
    }

    /* Utility method to retrieve an embedded object from a xlink:href
     */
    private EmbeddedObject getEmbeddedObject(String sHref) {
        if (sHref.startsWith("#")) {
            // Embedded object in package
            sHref = sHref.substring(1);
            if (sHref.startsWith("./")) { sHref = sHref.substring(2); } 
            return sxwDoc.getEmbeddedObject(sHref);
        }
        else {
            // TODO (0.5): Linked object
            return null;
        }
    }

    /*
     * Utility method to retrieve a Node attribute.
     */
    private String getAttribute (Node node, String attribute) {
        NamedNodeMap attrNodes = node.getAttributes();
        
        if (attrNodes != null) {
            Node attr = attrNodes.getNamedItem(attribute);
            if (attr != null) {
                return attr.getNodeValue();
            }
        }
        
        return null;
    }
	
    /* utility method to get the first child with a given tagname */
    private Node getChildByTagName(Node node, String sTagName){
        if (node.hasChildNodes()){
            NodeList nl=node.getChildNodes();
            int nLen=nl.getLength();
            for (int i=0; i<nLen; i++){
                Node child = nl.item(i);
                if (child.getNodeType() == Node.ELEMENT_NODE &&
                    child.getNodeName().equals(sTagName)){
                    return child;
                }
            }
        }
        return null;
    }

}