/************************************************************************
 *
 *  StyleConverter.java
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2.1, as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA  02111-1307  USA
 *
 *  Copyright: 2002-2004 by Henrik Just
 *
 *  All Rights Reserved.
 * 
 *  Version 0.3.3a (2004-02-19)
 *
 */

package writer2latex;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

import writer2latex.office.*;
import writer2latex.util.*;

/* Utility class to hold LaTeX code to put before/after other LaTeX code
 */
class BeforeAfter {
    private String sBefore="";
    private String sAfter="";

    void add(String sBefore1, String sAfter1) {
        sBefore+=sBefore1; sAfter=sAfter1+sAfter;
    }
	
    String getBefore() { return sBefore; }
    String getAfter() { return sAfter; }
	
    boolean isEmpty() { return sBefore.length()==0 && sAfter.length()==0; }
} 

/* This class tries to convert OOo Writer styles to LaTeX.
 */
class StyleConverter {
    private WriterStyleCollection wsc;
    private CharFormatHelper cfh;
    private I18n i18n;
    private Config config;
	
    // Text style data
    private ExportNameCollection textStyleNames = new ExportNameCollection(false);
    private StyleMap textStyleMap = new StyleMap();
    private LaTeXDocumentPortion textStyleDeclarations = new LaTeXDocumentPortion();

    // Paragraph style data
    private ExportNameCollection parStyleNames = new ExportNameCollection(false);
    private StyleMap parStyleMap = new StyleMap();
    private LaTeXDocumentPortion parStyleDeclarations = new LaTeXDocumentPortion();
    
    // List style data
    boolean bNeedSaveEnumCounter = false;
    private ExportNameCollection listStyleNames = new ExportNameCollection(false);
    private Hashtable listStyleLevelNames = new Hashtable();
    private LaTeXDocumentPortion listStyleDeclarations = new LaTeXDocumentPortion();
	
    // Master page data
    ExportNameCollection masterPageNames = new ExportNameCollection(false);
	
    /** <p>Constructs a new <code>StyleConverter</code>.</p>
     *  @param <code>wsc</code> the style collection to use
     */
    StyleConverter(WriterStyleCollection wsc, I18n i18n, Config config) {
        this.wsc = wsc;
        this.i18n = i18n;
        this.config = config;
        cfh = new CharFormatHelper(wsc,config);
    }
	
    public boolean needUlem() { return cfh.needUlem(); }

    ///////////////////////////////////////////////////////////////////////
    // Text styles
	
    /** <p>Convert a text style to LaTeX. </p> 
     *  <p>A soft style is declared in <code>textStyleDeclaration</code> as eg.
     *  <code>\newcommand\textstyleName[1]{\textbf{\textit{#1}}}</code></p>
     *  <p>A hard style is used by applying LaTeX code directly, eg.
     *  \textbf{\textit{...}}</p>
     *  @param <code>sName</code> the OOo name of the style 
    */
    void createTextStyle(String sName) {
        // The style may already be declared in the configuration:
        StyleMap sm = config.getTextStyleMap();
        if (sm.contains(sName)) {
            textStyleMap.put(sName,sm.getBefore(sName),sm.getAfter(sName));
            return;
        }
        // Otherwise do conversion:  
        StyleWithProperties style = wsc.getTextStyle(sName);
        if (style==null) {
            textStyleMap.put(sName,"","");
            return;
        }
        BeforeAfter ba = new BeforeAfter();
        cfh.applyLanguage(style,false,true,ba);
        cfh.applyFont(style,false,true,ba);
        cfh.applyFontEffects(style,true,ba);
        if (style.isAutomatic()) {
            textStyleMap.put(sName,ba.getBefore(),ba.getAfter());
        }
        else { // declare the text style (\newcommand)
            String sTeXName = textStyleNames.getExportName(sName);
            textStyleMap.put(sName,"\\textstyle"+sTeXName+"{","}");
            textStyleDeclarations.append("\\newcommand\\textstyle")
                .append(sTeXName).append("[1]{")
                .append(ba.getBefore()).append("#1").append(ba.getAfter())
                .append("}").nl();
        }
    }
	
    /** <p>Use a text style in LaTeX.</p>
     *  @param <code>sName</code> the name of the text style
     *  @param <code>ba</code> a <code>BeforeAfter</code> to put code into
     */
    void applyTextStyle(String sName, BeforeAfter ba) {
        if (sName==null) { return; }
        if (!textStyleMap.contains(sName)) { createTextStyle(sName); }
        ba.add(textStyleMap.getBefore(sName),textStyleMap.getAfter(sName));
    }
	
    /** <p>Convert text style declarations to LaTeX.</p>
     *  <p>(Temporary solution until new LaTeX output classes are in place.)</p>
     *  @param <code>ldp</code> the <code>LaTeXDocumentPortion</code> to add code to
     */
    void convertTextStyles(LaTeXDocumentPortion ldp) {
        if (!textStyleNames.isEmpty()) {
            ldp.append("% Text styles").nl();
            ldp.append(textStyleDeclarations.getContents());
        }
    }

    ///////////////////////////////////////////////////////////////////////
    // Paragraph formatting (convert paragraph properties to LaTeX)

    /** <p>Apply page break properties from a style.</p>
     *  @param <code>style</code> the paragraph style to use
     *  @param <code>bInherit</code> true if inheritance from parent style should be used
     *  @param <code>ba</code> a <code>BeforeAfter</code> to put code into
     */
    private void applyPageBreak(ParStyle style, boolean bInherit, BeforeAfter ba) {
        if (style==null) { return; }
        // A page break can be a simple page break before or after...
        String s = style.getProperty(XMLString.FO_BREAK_BEFORE,bInherit);
        if ("page".equals(s)) { ba.add("\\clearpage",""); }
        s = style.getProperty(XMLString.FO_BREAK_AFTER,bInherit);
        if ("page".equals(s)) { ba.add("","\\clearpage"); }
        // ...or it can be a new master page
        String sMasterPage = style.getMasterPageName();
        if (sMasterPage==null || sMasterPage.length()==0) { return; }
        ba.add("\\clearpage","");
        if (config.pageFormatting()==Config.IGNORE_ALL) return;
        MasterPage masterPage = wsc.getMasterPage(sMasterPage);
        if (masterPage==null) { return; }
        String sNext = masterPage.getProperty(XMLString.STYLE_NEXT_STYLE_NAME);
        MasterPage next = wsc.getMasterPage(sNext);
        if (masterPage==next || next==null) {
            ba.add("\\pagestyle{"+masterPageNames.getExportName(sMasterPage)+"}\n","");
        }
        else {
            ba.add("\\pagestyle{"+masterPageNames.getExportName(sNext)+"}\n","");
            ba.add("\\thispagestyle{"+masterPageNames.getExportName(sMasterPage)+"}\n","");
        }
        // todo: should warn the user if next master also contains a next-style-name;
        // LaTeX's page style mechanism cannot handle that
    }

    /** <p>Apply line spacing from a style.</p>
     *  @param <code>style</code> the paragraph style to use
     *  @param <code>ba</code> a <code>BeforeAfter</code> to put code into
     */
    private void applyLineSpacing(ParStyle style, BeforeAfter ba) {
        if (style==null) { return; }
        String sLineHeight = style.getProperty(XMLString.FO_LINE_HEIGHT);
        if (sLineHeight==null || !sLineHeight.endsWith("%")) { return; }
        float fPercent=Misc.getFloat(sLineHeight.substring(0,sLineHeight.length()-1),1);
        ba.add("\\renewcommand\\baselinestretch{"+fPercent/120+"}","");
    }

    /** <p>Helper: Get a length property that
     * defaults to 0cm. 
     */
    private String getLength(ParStyle style,String sProperty) {
        String s = style.getAbsoluteProperty(sProperty);
        if (s==null) { return "0cm"; }
        else { return s; }
    }	

    /** <p>Helper: Create a horizontal border.</p>
     */
    private String createBorder(String sLeft, String sRight, String sTop,
                                String sHeight, String sColor) {
        BeforeAfter baColor = new BeforeAfter();
        cfh.applyThisColor(sColor,false,baColor);
        return "{\\setlength\\parindent{0pt}\\setlength\\leftskip{" + sLeft + "}"
               + "\\setlength\\baselineskip{0pt}\\setlength\\parskip{" + sHeight + "}"
               + baColor.getBefore()
               + "\\rule{\\textwidth-" + sLeft + "-" + sRight + "}{" + sHeight + "}"
               + baColor.getAfter()
               + "\\par}";
    }	

    /** <p>Apply margin+alignment properties from a style.</p>
     *  @param <code>style</code> the paragraph style to use
     *  @param <code>ba</code> a <code>BeforeAfter</code> to put code into
     */
    private void applyMargins(ParStyle style, BeforeAfter ba) {
        // Read padding/margin/indentation properties:
        String sPaddingTop = getLength(style,XMLString.FO_PADDING_TOP);
        String sPaddingBottom = getLength(style,XMLString.FO_PADDING_BOTTOM);
        String sPaddingLeft = getLength(style,XMLString.FO_PADDING_LEFT);
        String sPaddingRight = getLength(style,XMLString.FO_PADDING_RIGHT);
        String sMarginTop = getLength(style,XMLString.FO_MARGIN_TOP);
        String sMarginBottom = getLength(style,XMLString.FO_MARGIN_BOTTOM);
        String sMarginLeft = getLength(style,XMLString.FO_MARGIN_LEFT);
        String sMarginRight = getLength(style,XMLString.FO_MARGIN_RIGHT);
        String sTextIndent;
        if ("true".equals(style.getProperty(XMLString.STYLE_AUTO_TEXT_INDENT))) {
            sTextIndent = "2em";
        }
        else {
            sTextIndent = getLength(style,XMLString.FO_TEXT_INDENT);
        }
        // Read alignment properties:
        boolean bRaggedLeft = false; // add 1fil to \leftskip
        boolean bRaggedRight = false; // add 1fil to \rightskip
        boolean bParFill = false; // add 1fil to \parfillskip
        String sTextAlign = style.getProperty(XMLString.FO_TEXT_ALIGN);
        if ("center".equals(sTextAlign)) {
            bRaggedLeft = true; bRaggedRight = true; // centered paragraph
        }
        else if ("start".equals(sTextAlign)) {
            bRaggedRight = true; bParFill = true; // left aligned paragraph
        }
        else if ("end".equals(sTextAlign)) {
            bRaggedLeft = true; // right aligned paragraph
        }
        else if (!"justify".equals(style.getProperty(XMLString.FO_TEXT_ALIGN_LAST))) {
            bParFill = true; // justified paragraph with ragged last line
        }
        // Create formatting:
        ba.add("\\setlength\\leftskip{"+sMarginLeft+(bRaggedLeft?" plus 1fil":"")+"}","");
        ba.add("\\setlength\\rightskip{"+sMarginRight+(bRaggedRight?" plus 1fil":"")+"}","");
        ba.add("\\setlength\\parindent{"+sTextIndent+"}","");
        ba.add("\\setlength\\parfillskip{"+(bParFill?"0pt plus 1fil":"0pt")+"}","");
        ba.add("\\setlength\\parskip{"+sMarginTop+"}","\\unskip\\vspace{"+sMarginBottom+"}");
    }
	
    private void applyAlignment(ParStyle style, boolean bIsSimple, boolean bInherit, BeforeAfter ba) {
        if (bIsSimple || style==null) { return; }
        String sTextAlign = style.getProperty(XMLString.FO_TEXT_ALIGN,bInherit);
        if ("center".equals(sTextAlign)) { ba.add("\\centering",""); }
        else if ("start".equals(sTextAlign)) { ba.add("\\raggedright",""); }
        else if ("end".equals(sTextAlign)) { ba.add("\\raggedleft",""); }
    }


    /** <p>Apply all paragraph properties.</p>
     *  @param <code>style</code> the paragraph style to use
     *  @param <code>ba</code> a <code>BeforeAfter</code> to put code into
     */
    private void applyParProperties(ParStyle style, BeforeAfter ba) {
        applyPageBreak(style,true,ba);
        ba.add("","\\par");
        applyLineSpacing(style,ba);
        applyMargins(style,ba);
    }

    ////////////////////////////////////////////////////////////////////////
    // Paragraph styles

    /** <p>Convert a paragraph style to LaTeX. </p> 
     *  <p>A soft style is declared in <code>parStyleDeclarations</code> as
     *  <code>\newenvironment...</code></p>
     *  <p>A hard style is used by applying LaTeX code directly</p>
     *  @param <code>sName</code> the OOo name of the style
     */
    void createParStyle(String sName) {
        boolean bConvertLists = config.listFormatting()==Config.CONVERT_ALL;
        ParStyle style = wsc.getParStyle(sName);
        if (style==null) {
            parStyleMap.put(sName,"","");
            return;
        }
        // Style exists, do conversion:
        BeforeAfter ba = new BeforeAfter();
        StyleMap sm = config.getParStyleMap();
        if (style.isAutomatic()) {
            String sParentName = style.getParentName();
            if (config.paragraphFormatting()!=Config.CONVERT_ALL ||
                sm.contains(sParentName)) {
                // Hard paragraph formatting from this style should be ignored.

                // Apply hard page break:
                applyPageBreak(style,false,ba);

                // Apply paragraph formatting from parent:
                BeforeAfter baPar = new BeforeAfter();
                applyParStyle(sParentName,baPar);

                // Apply hard character formatting:
                BeforeAfter baText = new BeforeAfter();
                cfh.applyLanguage(style,true,false,baText);
                cfh.applyFont(style,true,false,baText);

                // Assemble the bits. If there is any hard character formatting
                // we must group the contents.
                if (baPar.isEmpty() && !baText.isEmpty()) { ba.add("{","\\par}"); }
                else { ba.add(baPar.getBefore(),baPar.getAfter()); }
                ba.add(baText.getBefore(),baText.getAfter());

                parStyleMap.put(sName,ba.getBefore(),ba.getAfter());
                return;
            }
        }
        // Soft style (or hard style treated as soft):
        // The style may already be declared in the configuration:
        if (sm.contains(sName)) {
            parStyleMap.put(sName,sm.getBefore(sName),sm.getAfter(sName));
            return;
        }
        // Ignore styles?
        if (config.paragraphFormatting()==Config.IGNORE_ALL) {
            parStyleMap.put(sName,"","");
            return;
        }
        // Otherwise do conversion
        applyParProperties(style,ba);
        if (bConvertLists) { ba.add("\\writerlistparindent\\writerlistleftskip",""); }
        cfh.applyLanguage(style,true,true,ba);
        ba.add("\\leavevmode","");
        cfh.applyNormalFont(ba);
        cfh.applyFont(style,true,true,ba);
        if (bConvertLists) { ba.add("\\writerlistlabel\\ignorespaces",""); }
        // Declare the paragraph style (\newenvironment)
        String sTeXName = "style" + parStyleNames.getExportName(sName);
        parStyleMap.put(sName,"\\begin{"+sTeXName+"}","\\end{"+sTeXName+"}");
        parStyleDeclarations.append("\\newenvironment{").append(sTeXName)
            .append("}{").append(ba.getBefore()).append("}{")
            .append(ba.getAfter()).append("}").nl();
    }

    /** <p>Use a paragraph style in LaTeX.</p>
     *  @param <code>sName</code> the name of the text style
     *  @param <code>ba</code> a <code>BeforeAfter</code> to put code into
     */
    void applyParStyle(String sName, BeforeAfter ba) {
        if (sName==null) { return; }
        if (!parStyleMap.contains(sName)) { createParStyle(sName); }
        ba.add(parStyleMap.getBefore(sName),parStyleMap.getAfter(sName));
    }
	
    /** <p>Apply hard page break (no inheritance).</p>
     *  <p>This is used in sections</p>
     *  @param <code>style</code> the style to use
     *  @param <code>ba</code> the <code>BeforeAfter</code> to add LaTeX code to
     */
    void applyHardPageBreak(ParStyle style, BeforeAfter ba) {
        applyPageBreak(style,false,ba);
    }

    /** <p>Apply hard character formatting (no inheritance).</p>
     *  <p>This is used in sections and {foot|end}notes</p>
     *  @param <code>style</code> the style to use
     *  @param <code>ba</code> the <code>BeforeAfter</code> to add LaTeX code to
     */
    void applyHardCharFormatting(StyleWithProperties style, BeforeAfter ba) {
        cfh.applyLanguage(style,true,false,ba);
        cfh.applyFont(style,true,false,ba);
        if (!ba.isEmpty()) { ba.add(" ",""); }
    }

    /** <p>Convert paragraph style declarations to LaTeX.</p>
     *  <p>Note: This method should not be called before the end, since it only
     *  declares used paragraph styles.</p>
     *  <p>Temp. until new LaTeX output class is ready</p>
     *  @param <code>ldp</code> the <code>LaTeXDocumentPortion</code> to add code to
     */
    void convertParStyles(LaTeXDocumentPortion ldp) {
        if (config.paragraphFormatting()!=Config.IGNORE_ALL) {
            ldp.append("% Paragraph styles").nl();
            // First default paragraph style
            cfh.applyDefaultFont(wsc.getDefaultParStyle(),ldp);
            ldp.append(parStyleDeclarations.getContents());
        }         
    }
	
    ////////////////////////////////////////////////////////////////////////
    // Numbering

    static final String numFormat(String sFormat){
        if ("1".equals(sFormat)) { return "\\arabic"; }
        else if ("i".equals(sFormat)) { return "\\roman"; }
        else if ("I".equals(sFormat)) { return "\\Roman"; }
        else if ("a".equals(sFormat)) { return "\\alph"; }
        else if ("A".equals(sFormat)) { return "\\Alph"; }
        else { return null; }
    }    

    /** <p>Create LaTeX list labels from an OOo list style. Examples:</p>
     *  <p>Bullets:</p>
     *  <pre>\newcommand\labelliststylei{\textbullet}
     *  \newcommand\labelliststyleii{*}
     *  \newcommand\labelliststyleiii{\textstylebullet{>}}</pre>
     *  <p>Numbering:</p>
     *  <pre>\newcounter{liststylei}
     *  \newcounter{liststyleii}[liststylei]
     *  \newcounter{liststyleiii}[liststyleii]
     *  \renewcommand\theliststylei{\Roman{liststylei}}
     *  \renewcommand\theliststyleii{\Roman{liststylei}.\arabic{liststyleii}}
     *  \renewcommand\theliststyleiii{\alph{liststyleiii}}
     *  \newcommand\labelliststylei{\textstylelabel{\theliststylei .}}
     *  \newcommand\labelliststyleii{\textstylelabel{\theliststyleii .}}
     *  \newcommand\labelliststyleiii{\textstylelabel{\theliststyleiii )}}</pre>
     *
     *  @param <code>style</code> the OOo list style to use
     *  @param <code>sName</code> an array of label basenames to use
     *  @param <code>nMaxLevel</code> the highest level in this numbering
     *  @param <code>bDeclareCounters</code> true if counters should be declared (they may
     *  exist already, eg. "section", "subsection"... or "enumi", "enumii"... 
     *  @param <code>bRenewLabels</code> true if labels should be defined with \renewcommand
     *  @param <code>bUseTextStyle</code> true if labels should be formatted with the associated text style
     *  (rather than \newcommand).
     *  @param <code>ldp</code> the <code>LaTeXDocumentPortion</code> to add LaTeX code to.
     */
    void createLabels(ListStyle style, String[] sName, int nMaxLevel,
                      boolean bDeclareCounters, boolean bRenewLabels,
                      boolean bUseTextStyle, LaTeXDocumentPortion ldp) {
        // Declare counters if required (eg. "\newcounter{countername1}[countername2]")
        if (bDeclareCounters) {
            int j = 0;
            for (int i=1; i<=nMaxLevel; i++) {
                if (style.isNumber(i)) {
                    ldp.append("\\newcounter{").append(sName[i]).append("}");
                    if (j>0) { ldp.append("[").append(sName[j]).append("]"); }
                    ldp.nl();
                    j = i;
                }
            }
        }
        // Create numbering for each level (eg. "\arabic{countername}")
        String[] sNumFormat = new String[nMaxLevel+1];
        for (int i=1; i<=nMaxLevel; i++) {
            String s = numFormat(style.getLevelProperty(i,XMLString.STYLE_NUM_FORMAT));
            if (s==null) { sNumFormat[i]=""; }
            else { sNumFormat[i] = s + "{" + sName[i] + "}"; }
        }
        // Create numberings (ie. define "\thecountername"):
        for (int i=1; i<=nMaxLevel; i++) {
            if (style.isNumber(i)) {
                ldp.append("\\renewcommand\\the").append(sName[i]).append("{");
                int nLevels = Misc.getPosInteger(style.getLevelProperty(i,XMLString.TEXT_DISPLAY_LEVELS),1);
                for (int j=i-nLevels+1; j<i; j++) {
                    if (style.isNumber(j)) {
                        ldp.append(sNumFormat[j]).append(".");
                    }
                } 
                ldp.append(sNumFormat[i]);
                ldp.append("}").nl();
            }
        }
        // Create labels (ie. define "\labelcountername"):
        for (int i=1; i<=nMaxLevel; i++) {
            ldp.append(bRenewLabels ? "\\renewcommand" : "\\newcommand")
               .append("\\label").append(sName[i]).append("{");
            // Apply text style if required
            BeforeAfter baText = new BeforeAfter();
            if (bUseTextStyle) {
                String sStyleName = style.getLevelProperty(i,XMLString.TEXT_STYLE_NAME);
                applyTextStyle(sStyleName,baText);
            }

            // Create label content
            if (style.isNumber(i)) {
                String sPrefix = style.getLevelProperty(i,XMLString.STYLE_NUM_PREFIX);
                String sSuffix = style.getLevelProperty(i,XMLString.STYLE_NUM_SUFFIX);
                // Apply style
                ldp.append(baText.getBefore());
                if (sPrefix!=null) { ldp.append(sPrefix); }
                ldp.append("\\the").append(sName[i]);
                if (sSuffix!=null) { ldp.append(sSuffix); }
                ldp.append(baText.getAfter());
            }
            else if (style.isBullet(i)) {
                String sBullet = style.getLevelProperty(i,XMLString.TEXT_BULLET_CHAR);
                // Apply style
                ldp.append(baText.getBefore());
                if (sBullet!=null) {
                    // Bullets are usually symbols, so this should be OK:
                    ldp.append(i18n.convert(sBullet,false,"en"));
                }
                ldp.append(baText.getAfter());
            }
            else {
                // TODO: Support images!
                ldp.append("\\textbullet");
            }

            ldp.append("}").nl();
        }
    }

    ////////////////////////////////////////////////////////////////////////
    // Heading styles and outline numbering

    static final String sectionName(int nLevel){
        switch (nLevel) {
            case 1: return "section";
            case 2: return "subsection";
            case 3: return "subsubsection";
            case 4: return "paragraph";
            case 5: return "subparagraph";
            default: return null;
        }
    }

    /** <p>Convert heading styles and outline numbering to LaTeX.</p>
     *  <p>An array of stylenames to use is required: The OOo writer file format
     *  allows different paragraph styles to be applied to individual headings,
     *  so this is not included in the styles.
     *  LaTeX (and OOo Writer!) usually uses the same format for all headings.</p>
     *  @param sHeadingStyles    array of stylesnames to use for headings.
     *  @param ldp    the <code>LaTeXDocumentPortion</code> to add definitions to.
     */
    void convertHeadingStyles(String[] sHeadingStyles, LaTeXDocumentPortion ldp) {
        // The user may not want to convert the formatting of headings
        if (config.headingFormatting()==Config.IGNORE_ALL) { return; }

        HeadingMap hm = config.getHeadingMap();

        // OK, we are going to convert. First find the max level for headings
        int nMaxLevel = 0;
        for (int i=1; i<=5; i++) { if (sHeadingStyles[i]!=null) { nMaxLevel=i; } }
        if (nMaxLevel==0) { return; } // no headings, nothing to do!
        if (nMaxLevel>hm.getMaxLevel()) { nMaxLevel = hm.getMaxLevel(); }

        boolean bOnlyNum = config.headingFormatting()==Config.CONVERT_NUMBERING;
        if (bOnlyNum) {
            ldp.append("% Outline numbering").nl();
        }
        else {
            ldp.append("% Headings and outline numbering").nl()
               .append("\\makeatletter").nl();
        }

        // Paragraph style for headings:
        if (!bOnlyNum) {
            for (int i=1; i<=nMaxLevel; i++) {
                if (sHeadingStyles[i]!=null) {
                    ParStyle style = wsc.getParStyle(sHeadingStyles[i]);
                    if (style!=null) {
                        BeforeAfter decl = new BeforeAfter();
                        BeforeAfter comm = new BeforeAfter();
					
                        applyPageBreak(style,true,decl);

                        cfh.applyNormalFont(decl);
                        cfh.applyFont(style,true,true,decl);
                        applyAlignment(style,false,true,decl);
		    			
                        cfh.applyLanguage(style,false,true,comm);
                        cfh.applyFontEffects(style,true,comm);
										
                        String sMarginTop = getLength(style,XMLString.FO_MARGIN_TOP);
                        String sMarginBottom = getLength(style,XMLString.FO_MARGIN_BOTTOM);
                        String sMarginLeft = getLength(style,XMLString.FO_MARGIN_LEFT);
    
                        String sSecName = hm.getName(i);
                        if (!comm.isEmpty()) { // have to create a cs for this heading
                            ldp.append("\\newcommand\\cs").append(sSecName).append("[1]{")
                               .append(comm.getBefore()).append("#1").append(comm.getAfter())
                               .append("}").nl();
                        }
                        ldp.append("\\renewcommand\\").append(sSecName)
                           .append("{\\@startsection{").append(sSecName).append("}{"+i)
                           .append("}{"+sMarginLeft+"}{");
                        // Suppress indentation after heading?
                        if (config.paragraphFormatting()==Config.IGNORE_ALL) {
                            ldp.append("-");
                        }
                        ldp.append(sMarginTop)
                           .append("}{").append(sMarginBottom).append("}{");
                        // Note: decl.getAfter() may include a page break after, which we ignore
	                    ldp.append(decl.getBefore());
                        if (!comm.isEmpty()) {
                            ldp.append("\\cs").append(sSecName);
                        }
                        ldp.append("}}").nl();
                    }
                }
            }
        }

        // redefine formatting of section counters
        // simplified if the user wants to ignore formatting
        if (!bOnlyNum) {
            ldp.append("\\renewcommand\\@seccntformat[1]{")
               .append("\\csname @textstyle#1\\endcsname{\\csname the#1\\endcsname}")
               .append("\\csname @distance#1\\endcsname}").nl();
        }

        // Collect numbering styles and set secnumdepth
        int nSecnumdepth = nMaxLevel;
        ListStyle outline = wsc.getOutlineStyle();
        String[] sNumFormat = new String[6];
        for (int i=nMaxLevel; i>=1; i--) {
            sNumFormat[i] = numFormat(outline.getLevelProperty(i,
                               XMLString.STYLE_NUM_FORMAT));
            if (sNumFormat[i]==null || "".equals(sNumFormat[i])) {
                nSecnumdepth = i-1;
            }
        }
        ldp.append("\\setcounter{secnumdepth}{"+nSecnumdepth+"}").nl();

        for (int i=1; i<=nMaxLevel; i++) {
            if (sNumFormat[i]==null || "".equals(sNumFormat[i])) {
                // no numbering at this level
                if (!bOnlyNum) {
                    ldp.append("\\newcommand\\@distance")
                       .append(hm.getName(i)).append("{}").nl()
                       .append("\\newcommand\\@textstyle")
                       .append(hm.getName(i)).append("[1]{#1}").nl();
                }
            }
            else {
                if (!bOnlyNum) {
                    // Distance between label and text:
                    String sDistance = outline.getLevelProperty(i,XMLString.TEXT_MIN_LABEL_DISTANCE);
                    ldp.append("\\newcommand\\@distance")
                       .append(hm.getName(i)).append("{");
                    if (sDistance!=null) { 
                        ldp.append("\\hspace{").append(sDistance).append("{");
                    }
                    ldp.append("}").nl();
                    // Textstyle to use for label:
                    String sStyleName = outline.getLevelProperty(i,XMLString.TEXT_STYLE_NAME);
                    BeforeAfter baText = new BeforeAfter();
                    if (!bOnlyNum) {applyTextStyle(sStyleName,baText); }
                    ldp.append("\\newcommand\\@textstyle")
                       .append(hm.getName(i)).append("[1]{")
			           .append(baText.getBefore())
                       .append("#1")
                       .append(baText.getAfter())
                       .append("}").nl();
                }

                // The label:
                String sPrefix = outline.getLevelProperty(i,XMLString.STYLE_NUM_PREFIX);
                String sSuffix = outline.getLevelProperty(i,XMLString.STYLE_NUM_SUFFIX);
                int nLevels = Misc.getPosInteger(outline.getLevelProperty(i,
                                      XMLString.TEXT_DISPLAY_LEVELS),1);
                ldp.append("\\renewcommand\\the")
                   .append(hm.getName(i))
                   .append("{");
                StringBuffer labelbuf = new StringBuffer();
                if (sPrefix!=null) { labelbuf.append(sPrefix); }
                for (int j=i-nLevels+1; j<i; j++) {
                    labelbuf.append(sNumFormat[j])
                            .append("{").append(sectionName(j)).append("}")
                            .append(".");
                }
                labelbuf.append(sNumFormat[i])
                        .append("{").append(hm.getName(i)).append("}");
                if (sSuffix!=null) { labelbuf.append(sSuffix); }
                if (bOnlyNum) {
                    ldp.append(labelbuf.toString().trim());
                }
                else {
                    ldp.append(labelbuf.toString());
                }
                ldp.append("}").nl();
            }
            
        }

        if (!bOnlyNum) {
            ldp.append("\\makeatother").nl();
        }

    }
	
    ////////////////////////////////////////////////////////////////////////
    // List styles
	
    /** <p>Helper: Get a length property that
     * defaults to 0cm. 
     */
    private String getLength(ListStyle style,int nLevel,String sProperty) {
        String s = style.getLevelStyleProperty(nLevel,sProperty);
        if (s==null) { return "0cm"; }
        else { return s; }
    }	

    /** <p>Apply a list style to an ordered or unordered list.</p> */
    void applyListStyle(String sStyleName, int nLevel, boolean bOrdered,
                        boolean bContinue, BeforeAfter ba) {
        // Step 1. We may have a style map, this always takes precedence
        if (config.getListStyleMap().contains(sStyleName)) {
            ba.add(config.getListStyleMap().getBefore(sStyleName),
                   config.getListStyleMap().getAfter(sStyleName)); 
            return;
        }
        // Step 2: The list style may not exist, or the user wants to ignore it.
        // In this case we create default lists
        ListStyle style = wsc.getListStyle(sStyleName);
        if (style==null || config.listFormatting()==Config.IGNORE_ALL) {
            if (nLevel<=4) {
                if (bOrdered) {
                    ba.add("\\begin{enumerate}","\\end{enumerate}");
                }
                else {
                    ba.add("\\begin{itemize}","\\end{itemize}");
                }
            }
            return;
        }
        // Step 3: Export as default lists, but redefine labels
        if (config.listFormatting()==Config.CONVERT_NUMBERING) {
            if (nLevel==1) {
                if (!listStyleNames.containsName(sStyleName)) {
                    createListStyleLabels(sStyleName);
                }
                ba.add("\\liststyle"+listStyleNames.getExportName(sStyleName)+"\n","");
            }
            if (nLevel<=4) {
                String sCounterName = ((String[]) listStyleLevelNames.get(sStyleName))[nLevel]; 
                if (bContinue && style.isNumber(nLevel)) {
                    bNeedSaveEnumCounter = true;
                    ba.add("\\setcounter{saveenum}{\\value{"+sCounterName+"}}\n","");
                }
                if (bOrdered) {
                    ba.add("\\begin{enumerate}","\\end{enumerate}");
                }
                else {
                    ba.add("\\begin{itemize}","\\end{itemize}");
                }
                if (bContinue && style.isNumber(nLevel)) {
                    ba.add("\n\\setcounter{"+sCounterName+"}{\\value{saveenum}}","");
                }
            }
            return;			
        }
        // Step 4: Export with formatting, as "Writer style" custom lists
        if (nLevel<=4) { // TODO: Max level should not be fixed
            if (!listStyleNames.containsName(sStyleName)) {
                createListStyle(sStyleName);
            }
            String sTeXName="list"+listStyleNames.getExportName(sStyleName)
                           +"level"+Misc.int2roman(nLevel);
            if (!bContinue && style.isNumber(nLevel)) {
                ba.add("\\setcounter{"+sTeXName+"}{0}\n","");
            }
            ba.add("\\begin{"+sTeXName+"}","\\end{"+sTeXName+"}");
        }
    }
	
    /** <p>Apply a list style to a list item.</p> */
    void applyListItemStyle(String sStyleName, int nLevel, boolean bHeader,
                            boolean bRestart, int nStartValue, BeforeAfter ba) {
        // Step 1. We may have a style map, this always takes precedence
        if (config.getListItemStyleMap().contains(sStyleName)) {
            ba.add(config.getListItemStyleMap().getBefore(sStyleName),
                   config.getListItemStyleMap().getAfter(sStyleName)); 
            return;
        }
        // Step 2: The list style may not exist, or the user wants to ignore it.
        // In this case we create default lists
        ListStyle style = wsc.getListStyle(sStyleName);
        if (style==null || config.listFormatting()==Config.IGNORE_ALL) {
            if (nLevel<=4) {
                if (bHeader) { ba.add("\\item[] ",""); }
                else { ba.add("\\item ",""); }
            }
            return;
        }
        // Step 3: Export as default lists (with redefined labels)
        if (config.listFormatting()==Config.CONVERT_NUMBERING) {
            if (nLevel<=4) {
                if (bHeader) { 
                    ba.add("\\item[] ","");
                }
                else if (bRestart && style.isNumber(nLevel)) {
                    ba.add("\n\\setcounter{enum"+Misc.int2roman(nLevel)
                           +"}{"+(nStartValue-1)+"}\n\\item ","");
                }
                else {
                    ba.add("\\item ","");
                }
            }
            return;			
        }
        // Step 4: Export with formatting, as "Writer style" custom lists
        if (nLevel<=4 && !bHeader) { // TODO: Max level should not be fixed
            String sTeXName="list"+listStyleNames.getExportName(sStyleName)
                           +"level"+Misc.int2roman(nLevel);
            if (bRestart && style.isNumber(nLevel)) {
                ba.add("\\setcounter{"+sTeXName+"}{"+(nStartValue-1)+"}\n","");
            }
            ba.add("\\item ","");
        }
    }


    /** <p>Create labels for default lists based on a List Style
     */
    private void createListStyleLabels(String sStyleName) {
        String sTeXName = listStyleNames.getExportName(sStyleName);
        listStyleDeclarations.append("\\newcommand\\liststyle")
                             .append(sTeXName).append("{%").nl();
        ListStyle style = wsc.getListStyle(sStyleName);
        int nEnum = 0;
        int nItem = 0;
        String sName[] = new String[5];
        for (int i=1; i<=4; i++) {
            if (style.isNumber(i)) { sName[i]="enum"+Misc.int2roman(++nEnum); }
            else { sName[i]="item"+Misc.int2roman(++nItem); }
        }
        listStyleLevelNames.put(sStyleName,sName);
        createLabels(style, sName, 4, false, true, false, listStyleDeclarations);
        listStyleDeclarations.append("}").nl();
    }

    /** <p>Create "Writer style" lists based on a List Style.
        <p>A list in writer is really a sequence of numbered paragraphs, so
           this is also how we implement it in LaTeX.
           The enivronment + redefined \item defines three hooks:
           \writerlistleftskip, \writerlistparindent, \writerlistlabel
           which are used by exported paragraph styles to apply numbering.
     */
    private void createListStyle(String sStyleName) {
        ListStyle style = wsc.getListStyle(sStyleName);

        // Create labels
        String sTeXName = listStyleNames.getExportName(sStyleName);
        String[] sLevelName = new String[5];
        for (int i=1; i<=4; i++) {
            sLevelName[i]="list"+sTeXName+"level"+Misc.int2roman(i);
        }
        createLabels(style,sLevelName,4,true,false,true,listStyleDeclarations);

        // Create environments
        for (int i=1; i<=4; i++) {
            String sSpaceBefore = getLength(style,i,XMLString.TEXT_SPACE_BEFORE);
            String sLabelWidth = getLength(style,i,XMLString.TEXT_MIN_LABEL_WIDTH);
            String sLabelDistance = getLength(style,i,XMLString.TEXT_MIN_LABEL_DISTANCE);
            listStyleDeclarations
                .append("\\newenvironment{")
                .append(sLevelName[i]).append("}{")
                .append("\\def\\writerlistleftskip{\\addtolength\\leftskip{")
                .append(sSpaceBefore).append("+").append(sLabelWidth).append("}}")
                .append("\\def\\writerlistparindent{}")
                .append("\\def\\writerlistlabel{}");
            // Redefine \item
            listStyleDeclarations
                .append("\\def\\item{")
                .append("\\def\\writerlistparindent{\\setlength\\parindent{")
                .append("-").append(sLabelWidth).append("}}")
                .append("\\def\\writerlistlabel{");
            if (style.isNumber(i)) {
                listStyleDeclarations.append("\\stepcounter{")
                                     .append(sLevelName[i]).append("}");
            }
            listStyleDeclarations
                .append("\\label").append(sLevelName[i])
                .append("\\hspace{").append(sLabelDistance).append("}")
                .append("\\writerlistremovelabel}}}{}").nl();
        }
    }
	
    /** <p>Convert list style declarations to LaTeX.</p>
     *  <p>(Temporary solution until new LaTeX output classes are in place.)</p>
     *  <p>Note: This method should not be called before the end, since it only
     *  declares used list styles.</p>
     *  @param <code>ldp</code> the <code>LaTeXDocumentPortion</code> to add code to
     */
    void convertListStyles(LaTeXDocumentPortion ldp) {
        if (!listStyleNames.isEmpty()) {
            ldp.append("% List styles").nl();
            // May need an extra counter to handle continued numbering in lists
            if (bNeedSaveEnumCounter) {
                ldp.append("\\newcounter{saveenum}").nl();
            }
            // If we export formatting, we need some hooks from lists to paragraphs:
            if (config.listFormatting()==Config.CONVERT_ALL) {
                ldp.append("\\newcommand\\writerlistleftskip{}").nl()
                   .append("\\newcommand\\writerlistparindent{}").nl()
                   .append("\\newcommand\\writerlistlabel{}").nl()
                   .append("\\newcommand\\writerlistremovelabel{")
                   .append("\\aftergroup\\let\\aftergroup\\writerlistparindent\\aftergroup\\relax")
                   .append("\\aftergroup\\let\\aftergroup\\writerlistlabel\\aftergroup\\relax}").nl();
            }
            ldp.append(listStyleDeclarations.getContents());
        }
    }
	
    ////////////////////////////////////////////////////////////////////////
    // Table styles
	
    /** <p> Return a TableFormatter for this style </p>*/
    TableFormatter getTableFormatter(TableGridModel table,
        boolean bAllowLongtable) {
        return new TableFormatter(wsc,table,bAllowLongtable);
    }

    ////////////////////////////////////////////////////////////////////////
    // Page styles


    /** <p>Use a Master Page (pagestyle in LaTeX)</p>
     *  @param sName    name of the master page to use
     *  @param ldp      the <code>LaTeXDocumentPortion</code> to add code to.
     */
    void useMasterPage(String sName, LaTeXDocumentPortion ldp) {
        if (config.pageFormatting()==Config.IGNORE_ALL) return;
        MasterPage style = wsc.getMasterPage(sName);
        if (style==null) { return; }
        String sNextName = style.getProperty(XMLString.STYLE_NEXT_STYLE_NAME);
        MasterPage nextStyle = wsc.getMasterPage(sNextName);
        if (style==nextStyle || nextStyle==null) {
            ldp.append("\\pagestyle{")
               .append(masterPageNames.getExportName(sName))
               .append("}").nl();
        }
        else {
            ldp.append("\\pagestyle{")
               .append(masterPageNames.getExportName(sNextName))
               .append("}").nl()
               .append("\\thispagestyle{")
               .append(masterPageNames.getExportName(sName))
               .append("}").nl();
        }
    }
	 
    /** <p>Convert a page master (pagesize, margins etc.)
     *  @param sName   name of the page master to use 
     *  @param ldp     the <code>LaTeXDocumentPortion</code> to add code to.
     */
    void convertPageMaster(String sName, LaTeXDocumentPortion ldp) {
        if (config.pageFormatting()!=Config.CONVERT_ALL) { return; }
        PageMaster style = wsc.getPageMaster(sName);
        if (style==null) { return; }
		
        String sHeadHeight = "0cm";
        String sHeadSep = "0cm";
        if (style.hasHeaderStyle()) {
            sHeadHeight = "12pt";
            sHeadSep = style.getHeaderProperty(XMLString.FO_MARGIN_BOTTOM);
            if (sHeadSep==null) { sHeadSep = "0cm"; }
        }
		
        String sFootHeight = "0cm";
        String sFootSep = "0cm";
        if (style.hasFooterStyle()) {
            sFootHeight = "12pt";
            sFootSep = style.getFooterProperty(XMLString.FO_MARGIN_TOP);
            if (sFootSep==null) { sFootSep = "0cm"; }
        }
		
        // Page dimensions
        ldp.append("\\setlength\\paperwidth{")
           .append(style.getAbsoluteProperty(XMLString.FO_PAGE_WIDTH))
           .append("}")
           .append("\\setlength\\paperheight{")
           .append(style.getAbsoluteProperty(XMLString.FO_PAGE_HEIGHT))
           .append("}");
        // PDF page dimensions, only if hyperref.sty is not loaded
        if (config.getBackend()==Config.PDFTEX && !config.useHyperref()) {
            ldp.append("\\setlength\\pdfpagewidth{")
               .append(style.getAbsoluteProperty(XMLString.FO_PAGE_WIDTH))
               .append("}")
               .append("\\setlength\\pdfpageheight{")
               .append(style.getAbsoluteProperty(XMLString.FO_PAGE_HEIGHT))
               .append("}");
        }
        // Page starts in upper left corner of paper!!
        ldp.append("\\setlength\\voffset{-1in}");
        ldp.append("\\setlength\\hoffset{-1in}");
        // Top margin
        ldp.append("\\setlength\\topmargin{")
           .append(style.getProperty(XMLString.FO_MARGIN_TOP))
           .append("}");
        // Header and footer
        ldp.append("\\setlength\\headheight{").append(sHeadHeight).append("}");
        ldp.append("\\setlength\\headsep{").append(sHeadSep).append("}");
        ldp.append("\\setlength\\footskip{")
           .append(sFootHeight).append("+").append(sFootSep).append("}");
        // Bottom margin is set indirectly via text height
        ldp.append("\\setlength\\textheight{")
           .append(style.getProperty(XMLString.FO_PAGE_HEIGHT))
           .append("-").append(style.getProperty(XMLString.FO_MARGIN_TOP))
           .append("-").append(style.getProperty(XMLString.FO_MARGIN_BOTTOM))
           .append("-").append(sHeadSep)
           .append("-").append(sHeadHeight)
           .append("-").append(sFootSep)
           .append("-").append(sFootHeight).append("}");
        // Left margin
        ldp.append("\\setlength\\oddsidemargin{")
           .append(style.getProperty(XMLString.FO_MARGIN_LEFT))
           .append("}");
        // Left margin for even (left) pages; only for mirrored page master
        if ("mirrored".equals(style.getPageUsage())) {
            ldp.append("\\setlength\\evensidemargin{")
               .append(style.getProperty(XMLString.FO_MARGIN_RIGHT))
               .append("}");
        }
        // Right margin is set indirectly via text width
        ldp.append("\\setlength\\textwidth{")
           .append(style.getProperty(XMLString.FO_PAGE_WIDTH))
           .append("-")
           .append(style.getProperty(XMLString.FO_MARGIN_LEFT))
           .append("-")
           .append(style.getProperty(XMLString.FO_MARGIN_RIGHT))
           .append("}").nl();
        // Set additional lengths
        //ldp.append("\\@colht\\textheight\\@colroom\\textheight\\vsize\\textheight").nl();

        // Define columns
        //if (style.getColCount()>1) { ldp.append("\\twocolumn").nl(); }
        //else { ldp.append("\\onecolumn").nl(); }

        // Page number
        String sNumFormat = style.getProperty(XMLString.STYLE_NUM_FORMAT);
        if (sNumFormat!=null) {
            ldp.append("\\renewcommand\\thepage{")
               .append(numFormat(sNumFormat))
               .append("{page}}").nl();
        }
        String sPageNumber = style.getProperty(XMLString.STYLE_FIRST_PAGE_NUMBER);
        if (sPageNumber!=null && !sPageNumber.equals("continue")) {
           ldp.append("\\setcounter{page}{")
              .append(Integer.toString(Misc.getPosInteger(sPageNumber,0)))
              .append("}").nl();
        }
        // Footnote rule
        // TODO: Support alignment.
        String sBefore = style.getFootnoteProperty(XMLString.STYLE_DISTANCE_BEFORE_SEP);
        if (sBefore==null) { sBefore = "1mm"; }
        String sAfter = style.getFootnoteProperty(XMLString.STYLE_DISTANCE_AFTER_SEP);
        if (sAfter==null) { sAfter = "1mm"; }
        String sHeight = style.getFootnoteProperty(XMLString.STYLE_WIDTH);
        if (sHeight==null) { sHeight = "0.2mm"; }
        String sWidth = style.getFootnoteProperty(XMLString.STYLE_REL_WIDTH);
        if (sWidth==null) { sWidth = "25%"; }
        sWidth=Float.toString(Misc.getFloat(sWidth.substring(0,sWidth.length()-1),1)/100);
        BeforeAfter baColor = new BeforeAfter();
        cfh.applyThisColor(style.getFootnoteProperty(XMLString.STYLE_COLOR),false,baColor);
 
        ldp.append("\\setlength{\\skip\\footins}{").append(sBefore).append("}");
        ldp.append("\\renewcommand\\footnoterule{\\vspace*{-").append(sHeight)
           .append("}\\noindent")
           .append(baColor.getBefore()).append("\\rule{").append(sWidth)
           .append("\\columnwidth}{").append(sHeight).append("}")
           .append(baColor.getAfter())
           .append("\\vspace*{").append(sAfter).append("}}").nl();
    }
	
    ////////////////////////////////////////////////////////////////////////
    // {Foot|end}note configuration


    /** <p>Convert {foot|end}notes configuration.</p>
     *  <p>Note: All {foot|end}notes are formatted with the default style for {foot|end}footnotes.
     *  (This doesn't conform with the file format specification, but in LaTeX
     *  all {foot|end}notes are usually formatted in a fixed style.)</p>
     *
     *  @param ldp    the <code>LaTeXDocumentPortion</code> to add definitions to.
     */
    void convertFootEndnotesConfiguration(PropertySet notes, String sType, LaTeXDocumentPortion ldp) {
        if (config.ignoreFootnotesConfiguration()) { return; }
        String sTypeShort;
        if (sType.equals("foot")) { sTypeShort="fn"; } else { sTypeShort="en"; }
        if (notes==null) { return; }
        ldp.append("% ").append(sType).append("notes configuration").nl()
           .append("\\makeatletter").nl();

        // The numbering style is controlled by \the{foot|end}note
		String sFormat = notes.getProperty(XMLString.STYLE_NUM_FORMAT);
        if (sFormat!=null) {
            ldp.append("\\renewcommand\\the").append(sType).append("note{")
               .append(numFormat(sFormat))
               .append("{").append(sType).append("note}}").nl();
        }
        
        // Number {foot|end}notes by sections
        if ("chapter".equals(notes.getProperty(XMLString.TEXT_START_NUMBERING_AT))) {
            ldp.append("\\@addtoreset{").append(sType).append("note}{section}").nl();
        }
		
        // Set start value offset (default 0)
        int nStartValue = Misc.getPosInteger(notes.getProperty(XMLString.TEXT_START_VALUE),0);
        if (nStartValue!=0) {
            ldp.append("\\setcounter{").append(sType).append("note}{"+nStartValue+"}").nl();
        }
        
        // The formatting of the {foot|end}note citation is controlled by \@make{fn|en}mark
        String sCitBodyStyle = notes.getProperty(XMLString.TEXT_CITATION_BODY_STYLE_NAME);
        if (sCitBodyStyle!=null && wsc.getTextStyle(sCitBodyStyle)!=null) {
            BeforeAfter baText = new BeforeAfter();
            applyTextStyle(sCitBodyStyle,baText);
            ldp.append("\\renewcommand\\@make").append(sTypeShort).append("mark{\\mbox{")
               .append(baText.getBefore())
               .append("\\@the").append(sTypeShort).append("mark")
               .append(baText.getAfter())
               .append("}}").nl();
        }
		
        // The layout and formatting of the {foot|end}note is controlled by \@make{fn|en}text
        String sCitStyle = notes.getProperty(XMLString.TEXT_CITATION_STYLE_NAME);
        String sStyleName = notes.getProperty(XMLString.TEXT_DEFAULT_STYLE_NAME);
        if (sStyleName!=null) {
            BeforeAfter baText = new BeforeAfter();
            applyTextStyle(sCitStyle,baText);
            ParStyle style = wsc.getParStyle(sStyleName);
            if (style!=null) {
                BeforeAfter baPar = new BeforeAfter();
                applyHardCharFormatting(style,baPar);
                ldp.append("\\renewcommand\\@make").append(sTypeShort)
                   .append("text[1]{\\noindent")
                   .append(baText.getBefore())
                   .append("\\@the").append(sTypeShort).append("mark\\ ")
                   .append(baText.getAfter())
                   .append(baPar.getBefore())
                   .append("#1")
                   .append(baPar.getAfter());
                ldp.append("}").nl();
            }	 
        }
        
        ldp.append("\\makeatother").nl();		
    }

    /** <p>Convert footnotes configuration.</p>
     *  <p>Note: All footnotes are formatted with the default style for footnotes.
     *  (This doesn't conform with the file format specification, but in LaTeX
     *  all footnotes are usually formatted in a fixed style.)</p>
     *
     *  @param ldp    the <code>LaTeXDocumentPortion</code> to add definitions to.
     */
    void convertFootnotesConfiguration(LaTeXDocumentPortion ldp) {
        // Note: Continuation notices are not supported in LaTeX
        // TODO: Support text:footnotes-postion="document" (footnotes as endnotes)
        // TODO: Support text:start-numbering-at="page" (footnpag.sty)
        convertFootEndnotesConfiguration(wsc.getFootnotesConfiguration(),"foot",ldp);
    }

    /** <p>Convert endnotes configuration.</p>
     *  <p>Note: This method automatically includes the package "endnotes.sty".</p>
     *  <p>Note: All endnotes are formatted with the default style for endnotes.
     *  (This doesn't conform with the file format specification, but in LaTeX
     *  all endnotes are usually formatted in a fixed style.)</p>
     *
     *  @param ldp    the <code>LaTeXDocumentPortion</code> to add definitions to.
     */
    void convertEndnotesConfiguration(LaTeXDocumentPortion ldp) {
        // Note: Continuation notices are not supported in LaTeX
        convertFootEndnotesConfiguration(wsc.getEndnotesConfiguration(),"end",ldp);
    }
	
    ////////////////////////////////////////////////////////////////////////
    // Other

    void convertLanguageUsage(LaTeXDocumentPortion ldp) {
        Vector languages = new Vector();
        // Collect languages from text styles
        Enumeration enum = wsc.getTextStyles().getStylesEnumeration();
        while (enum.hasMoreElements()) {
            StyleWithProperties style = (StyleWithProperties) enum.nextElement();
            String sLang = I18n.getBabelLanguage(style.getProperty(XMLString.FO_LANGUAGE));
            if (sLang!=null && !languages.contains(sLang)) { languages.add(sLang); }
        }
        // Collect languages from paragraph styles
        enum = wsc.getParStyles().getStylesEnumeration();
        while (enum.hasMoreElements()) {
            StyleWithProperties style = (StyleWithProperties) enum.nextElement();
            String sLang = I18n.getBabelLanguage(style.getProperty(XMLString.FO_LANGUAGE));
            if (sLang!=null && !languages.contains(sLang)) { languages.add(sLang); }
        }
        // Read the default language from the default paragraph style
        String sDefaultLanguage = null;
        StyleWithProperties style = wsc.getDefaultParStyle();
        if (style!=null) { 
            sDefaultLanguage = I18n.getBabelLanguage(style.getProperty(XMLString.FO_LANGUAGE));
        }
        // If the document contains "anonymous" greek letters we need greek anyway:
        if (i18n.greek() && !languages.contains("greek")) languages.add("greek"); 
        // Load babel with the used languages
        CSVList babelopt = new CSVList(",");		
        Enumeration langenum = languages.elements();
        while (langenum.hasMoreElements()) {
            String sLang = (String) langenum.nextElement();
            if (!sLang.equals(sDefaultLanguage)) {
                if ("greek".equals(sLang) && i18n.polytonicGreek()) {
                    sLang = "polutonikogreek";
                }
                babelopt.addValue(sLang);
            }
        }
        // The default language must be the last one
        if (sDefaultLanguage!=null) {
            if ("greek".equals(sDefaultLanguage) && i18n.polytonicGreek()) {
                babelopt.addValue("polutonikogreek");
            }
            else {
                babelopt.addValue(sDefaultLanguage);
            }
        }
        if (!babelopt.isEmpty()) {
            ldp.append("\\usepackage[")
               .append(babelopt.toString())
               .append("]{babel}").nl();
        }
    }
    	
}