/************************************************************************
 *
 *  StyleConverter.java
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2.1, as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA  02111-1307  USA
 *
 *  Copyright: 2002-2003 by Henrik Just
 *
 *  All Rights Reserved.
 * 
 *  Version 0.3.1 (2003-10-08)
 *
 */

package writer2latex;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

import writer2latex.office.*;
import writer2latex.util.*;

/* Utility class to hold LaTeX code to put before/after other LaTeX code
 */
class BeforeAfter {
    private String sBefore="";
    private String sAfter="";

    void add(String sBefore1, String sAfter1) {
        sBefore+=sBefore1; sAfter=sAfter1+sAfter;
    }
	
    String getBefore() { return sBefore; }
    String getAfter() { return sAfter; }
} 

/* This class tries to convert OOo Writer styles to LaTeX.
 */
class StyleConverter {
    private WriterStyleCollection wsc;
    private I18n i18n;
    private Config config;
	
    // Cache of converted font declarations
    private Hashtable fontDeclsComm = new Hashtable();
    private Hashtable fontDeclsDecl = new Hashtable();
	
    // Cache of converted text styles
    private ExportNameCollection textStyleNames = new ExportNameCollection(false);
    private StyleMap textStyleMap = new StyleMap();

    // Cache of converted paragraph styles
    private ExportNameCollection parStyleNames = new ExportNameCollection(false);
    private StyleMap parStyleMap = new StyleMap();
    private StyleMap parStyleMapSimple = new StyleMap();
    
    // Cache of converted list styles
    private ExportNameCollection listStyleNames = new ExportNameCollection(false);
	
    // Cache of converted master pages
    ExportNameCollection masterPageNames = new ExportNameCollection(false);
	
    // Do we need ulem.sty?
    private boolean bNeedUlem = false;

    /** <p>Constructs a new <code>StyleConverter</code>.</p>
     *  @param <code>wsc</code> the style collection to use
     */
    StyleConverter(WriterStyleCollection wsc, I18n i18n, Config config) {
        this.wsc = wsc;
        this.i18n = i18n;
        this.config = config;
    }
	
    public boolean needUlem() { return bNeedUlem; }

    /** <p>Convert font declarations to LaTeX.</p>
     *  <p>It returns a generic LaTeX font family. Either as a declaration
     *  like <code>\sffamily</code> or as a command like <code>\textsf</code>.</p>
     *  <p>It returns null if the font declaration doesn't exist.</p>
     *  @return <code>String</code> with a LaTeX generic fontfamily
     *  @param  <code>sName</code> the name of the font declaration
     *  @param  <code>bDeclaration</code> true if the declaration form is required
     */
    String convertFontDeclaration(String sName, boolean bDeclaration) {
        FontDeclaration fd = wsc.getFontDeclaration(sName);
        if (fd==null) { return null; }
        if (bDeclaration) {
            if (!fontDeclsDecl.containsKey(sName)) {
                String sFontFamily = fd.getProperty(XMLString.STYLE_FONT_FAMILY);
                String sFontPitch = fd.getProperty(XMLString.STYLE_FONT_PITCH);
                String sFontFamilyGeneric = fd.getProperty(XMLString.STYLE_FONT_FAMILY_GENERIC);
                fontDeclsDecl.put(sName,nfssFamily(sFontFamily,sFontFamilyGeneric,sFontPitch,true));
            }
            return (String) fontDeclsDecl.get(sName);
        }
        else {
            if (!fontDeclsComm.containsKey(sName)) {
                String sFontFamily = fd.getProperty(XMLString.STYLE_FONT_FAMILY);
                String sFontPitch = fd.getProperty(XMLString.STYLE_FONT_PITCH);
                String sFontFamilyGeneric = fd.getProperty(XMLString.STYLE_FONT_FAMILY_GENERIC);
                fontDeclsComm.put(sName,nfssFamily(sFontFamily,sFontFamilyGeneric,sFontPitch,false));
            }
            return (String) fontDeclsComm.get(sName);
        }
    }
	
	////////////////////////////////////////////////////////////////////////
    // Text styles
	
    /** <p>Use a text style in LaTeX: Puts code <i>before</i> the contents.</p>
     *  <p>Does nothing if the supplied name is null or if the text style doesn't exist.</p>
     *  @param <code>sName</code> the name of the text style
     *  @param <code>ldp</code> the <code>LaTeXDocumentPortion</code> to add code to
     */
    void useTextStyleBegin(String sName, LaTeXDocumentPortion ldp) {
        // Is the style specified in the configuration?
        if (config.getTextStyleMap().contains(sName)) {
            ldp.append( config.getTextStyleMap().getBefore(sName) );
            return;
        }

        // Else convert the style:
        StyleWithProperties style = wsc.getTextStyle(sName);
        if (style==null) { return; }
        if (!textStyleMap.contains(sName)) {
            BeforeAfter ba = new BeforeAfter();

            applyLanguage(style,false,ba);
            applyNfssFamily(style,false,ba);
            applyNfssSeries(style,false,ba);
            applyNfssShape(style,false,ba);
            applyNfssSize(style,false,ba);
            applyColor(style,false,ba);
            applyUnderline(style,ba);
            applyCrossout(style,ba);
            applyTextPosition(style,ba);
            applyChangeCase(style,ba);
        
            textStyleMap.put(sName,ba.getBefore(),ba.getAfter());
            if (!style.isAutomatic()) { textStyleNames.addName(sName); }
        }
     
        if (style.isAutomatic()){
            ldp.append( (String) textStyleMap.getBefore(sName) );
        }
        else {
            ldp.append("\\textstyle").append(textStyleNames.getExportName(sName))
               .append("{");
        }
    }
	
    /** <p>Use a text style in LaTeX: Puts code <i>after</i> the contents.</p>
     *  <p>Does nothing if the supplied name is null or if the text style doesn't exist.</p>
     *  @param <code>sName</code> the name of the text style
     *  @param <code>ldp</code> the <code>LaTeXDocumentPortion</code> to add code to
     */
    void useTextStyleEnd(String sName, LaTeXDocumentPortion ldp) {
        // Is the style specified in the configuration?
        if (config.getTextStyleMap().contains(sName)) {
            ldp.append( config.getTextStyleMap().getAfter(sName) );
            return;
        }

        // Else convert the style:
        StyleWithProperties style = wsc.getTextStyle(sName);
        if (style==null) { return; }
        if (style.isAutomatic()){
            ldp.append( textStyleMap.getAfter(sName) );
        }
        else {
            ldp.append("}");
        }
    }
	
    /** <p>Convert text style declarations to LaTeX.</p>
     *  <p>Note: This method should not be called before the end, since it only
     *  declares used text styles.</p>
     *  @param <code>ldp</code> the <code>LaTeXDocumentPortion</code> to add code to
     */
    void convertTextStyles(LaTeXDocumentPortion ldp) {
        if (!textStyleNames.isEmpty()) {
            ldp.append("% Text styles").nl();
        }
        Enumeration styles = wsc.getTextStyles().getStylesEnumeration();
        while (styles.hasMoreElements()) {
            StyleWithProperties style = (StyleWithProperties) styles.nextElement();
            String sName = style.getName();
            if (textStyleNames.containsName(sName) && textStyleMap.contains(sName)) {
                ldp.append("\\newcommand\\textstyle")
                   .append(textStyleNames.getExportName(sName))
                   .append("[1]{")
                   .append( textStyleMap.getBefore(sName) )
                   .append("#1")
                   .append( textStyleMap.getAfter(sName) )
                   .append("}").nl();
            }
        }
    }

	////////////////////////////////////////////////////////////////////////
    // Paragraph styles

    public String getParStyleNfss(StyleWithProperties style) {
        BeforeAfter ba = new BeforeAfter();
        applyNfssFamily(style,true,ba);
        applyNfssSeries(style,true,ba);
        applyNfssShape(style,true,ba);
        applyNfssSize(style,true,ba);
        applyColor(style,true,ba);
        return ba.getBefore();
    }	

    private void parStyleNfss(StyleWithProperties style, BeforeAfter ba) {
        ba.add("\\normalfont\\normalsize\\normalcolor","");
        applyNfssFamily(style,true,ba);
        applyNfssSeries(style,true,ba);
        applyNfssShape(style,true,ba);
        applyNfssSize(style,true,ba);
        applyColor(style,true,ba);
    }	

    /** <p>Use a paragraph style in LaTeX: Puts code <i>before</i> the contents.</p>
     *  <p>Does nothing if the supplied name is null or if the text style doesn't exist.</p>
     *  <p>This method is for normal text flow, <code>useParStyleBeginSimple</code>
     *  is for restricted contexts like table cells, footnotes.</p>
     *  @param <code>sName</code> the name of the paragraph style
     *  @param <code>ldp</code> the <code>LaTeXDocumentPortion</code> to add code to
     */
    void useParStyleBegin(String sName, LaTeXDocumentPortion ldp) {
        useParStyleBeginCommon(sName,ldp,false);
    }

    /** <p>Use a paragraph style in LaTeX: Puts code <i>before</i> the contents.</p>
     *  <p>Does nothing if the supplied name is null or if the text style doesn't exist.</p>
     *  <p>This method is for restricted contexts like table cells, footnotes.</p>
     *  @param <code>sName</code> the name of the paragraph style
     *  @param <code>ldp</code> the <code>LaTeXDocumentPortion</code> to add code to
     */
    void useParStyleBeginSimple(String sName, LaTeXDocumentPortion ldp) {
        useParStyleBeginCommon(sName,ldp,true);
    }
	
    private void useParStyleBeginCommon(String sName, LaTeXDocumentPortion ldp, boolean bIsSimple) {
        // Style doesn't exist? Give up!
        ParStyle style = wsc.getParStyle(sName);
        if (style==null) return;
        // The configuration may specify what to do with this style
        // Step 1: Ignore hard paragraph formatting and use underlying style?
        if (config.hardParagraphFormatting()==Config.IGNORE) {
            if (style.isAutomatic()) {
                // Do not ignore master page!
                String s = style.getMasterPageName();
                if (s!=null && s.length()>0 && !bIsSimple) {
                    ldp.append("\\clearpage").nl();
                    useMasterPage(s,ldp);
                }
                // Continue with underlying soft style:
                sName = style.getParentName();
                style = wsc.getParStyle(sName);
                if (style==null) return; // give up!
            }
        }
        // Step 2: Is this the main paragraph style?
        if (sName.equals(config.mainParagraphStyle())) {
            return; // note: formatting is specified in the preamble
        }
        // Step 3: Is there a map for this style in the configuration?
        if (config.getParStyleMap().contains(sName)) {
            String sBefore = config.getParStyleMap().getBefore(sName); 
            if (sBefore.length()>0) { ldp.append(sBefore).nl(); }
            return;
        }
        // Step 4: Ignore unknow paragraph styles?
        if (config.ignoreUnknownParagraphStyles()) {
            // Do not ignore master page!
            String s = style.getMasterPageName();
            if (s!=null && s.length()>0 && !bIsSimple) {
                ldp.append("\\clearpage").nl();
                useMasterPage(s,ldp);
            }
            return;
        }

        // OK, convert the style!
        StyleMap sm;
        if (bIsSimple) { sm = parStyleMapSimple; }
        else { sm = parStyleMap; }

        if (!sm.contains(sName)) {
            BeforeAfter ba = new BeforeAfter();
            String s;
            if (!bIsSimple) {
                // TODO: Change page master if "next master page" is different
                s = style.getProperty(XMLString.FO_BREAK_BEFORE);
                if ("page".equals(s)){ ba.add("\\clearpage",""); }
                s = style.getProperty(XMLString.FO_BREAK_AFTER);
                if ("page".equals(s)){ ba.add("","\\clearpage"); }

                s = style.getAbsoluteProperty(XMLString.FO_MARGIN_BOTTOM);
                if (s==null) { s="0cm"; }
                ba.add("","\\vspace{"+s+"}");

                //	TODO: \renewcommand\baselinestretch{???}%
			
                ba.add("\\begin{list}{}{\\topsep0cm\\partopsep0cm\\itemsep0cm","\\end{list}");

                s = style.getAbsoluteProperty(XMLString.FO_MARGIN_LEFT);
                if (s==null) { s="0cm"; }
                ba.add("\\leftmargin"+s,"");
			
                s = style.getAbsoluteProperty(XMLString.FO_MARGIN_RIGHT);
                if (s==null) { s="0cm"; }
                ba.add("\\rightmargin"+s,"");

                s = style.getAbsoluteProperty(XMLString.FO_MARGIN_TOP);
                if (s==null) { s="0cm"; }
                ba.add("\\parskip"+s,"");

                s = style.getAbsoluteProperty(XMLString.FO_TEXT_INDENT);
                if (s==null) {
                    if ("true".equals(style.getProperty(XMLString.STYLE_AUTO_TEXT_INDENT))) { s="2em"; }
                    else { s="0cm"; }
                }
                ba.add("\\itemindent"+s,"");

                ba.add("}\\item[]","");
				
                s = alignment(style.getProperty(XMLString.FO_TEXT_ALIGN));
                if (s!=null) { ba.add(s,""); }
            }
			
            applyLanguage(style,true,ba);                

            parStyleNfss(style,ba);

            applyUnderline(style,ba);        
            applyCrossout(style,ba);        
            applyChangeCase(style,ba);

            sm.put(sName,ba.getBefore(),ba.getAfter());

            if (!style.isAutomatic() && !parStyleNames.containsName(sName)) {
                parStyleNames.addName(sName);
            }
        }
     
        if (bIsSimple) {
            if (style.isAutomatic() && !(config.hardParagraphFormatting()==Config.SOFT)){
                ldp.append("{").append( sm.getBefore(sName) ).append(" ");
            }
            else {
                ldp.append("\\begin{style").append(parStyleNames.getExportName(sName))
                   .append("*}");
            }
        }
        else {
            if (style.isAutomatic() && !(config.hardParagraphFormatting()==Config.SOFT)){
                String s = style.getMasterPageName();
                if (s!=null && s.length()>0) {
                    ldp.append("\\clearpage").nl();
                    useMasterPage(s,ldp);
                }
                ldp.append("{")
                   .append( sm.getBefore(sName) ).nl();
            }
            else {
                ldp.append("\\begin{style").append(parStyleNames.getExportName(sName))
                   .append("}").nl();
            }
        }
    }

    /** <p>Use a paragraph style in LaTeX: Puts code <i>after</i> the contents.</p>
     *  <p>Does nothing if the supplied name is null or if the par style doesn't exist.</p>
     *  @param <code>sName</code> the name of the par style
     *  @param <code>ldp</code> the <code>LaTeXDocumentPortion</code> to add code to
     */
    void useParStyleEnd(String sName, LaTeXDocumentPortion ldp) {
        useParStyleEndCommon(sName,ldp,false);
    }
	
    /** <p>Use a paragraph style to LaTeX: Puts code <i>after</i> the contents.</p>
     *  <p>Does nothing if the supplied name is null or if the par style doesn't exist.</p>
     *  @param <code>sName</code> the name of the par style
     *  @param <code>ldp</code> the <code>LaTeXDocumentPortion</code> to add code to
     */
    void useParStyleEndSimple(String sName, LaTeXDocumentPortion ldp) {
        useParStyleEndCommon(sName,ldp,true);
    }
	
    private void useParStyleEndCommon(String sName, LaTeXDocumentPortion ldp, boolean bIsSimple) {
        // Style doesn't exist? Give up!
        ParStyle style = wsc.getParStyle(sName);
        if (style==null) return;
        // The configuration may specify what to do with this style
        // Step 1: Ignore hard paragraph formatting and use underlying style?
        if (config.hardParagraphFormatting()==Config.IGNORE) {
            if (style.isAutomatic()) {
                sName = style.getParentName();
                style = wsc.getParStyle(sName);
                if (style==null) return; // give up!
            }
        }
        // Step 2: Is this the main paragraph style?
        if (sName.equals(config.mainParagraphStyle())) {
            ldp.nl(); // new paragraph
            return; // note: formatting is specified in the preamble
        }
        // Step 3: Is there a map for this style in the configuration?
        if (config.getParStyleMap().contains(sName)) {
            String sAfter = config.getParStyleMap().getAfter(sName); 
            ldp.append(sAfter).nl();
            return;
        }
        // Step 4: Ignore unknown paragraph styles?
        if (config.ignoreUnknownParagraphStyles()) {
            ldp.nl(); // new paragraph
            return;
        }

        // OK, convert the style!
        StyleMap sm;
        if (bIsSimple) { sm = parStyleMapSimple; }
        else { sm = parStyleMap; }

        if (style.isAutomatic() && !(config.hardParagraphFormatting()==Config.SOFT)){
            ldp.append( sm.getAfter(sName) ).append("}").nl();
        }
        else {
            ldp.append("\\end{style").append(parStyleNames.getExportName(sName));
            if (bIsSimple) ldp.append("*");
            ldp.append("}").nl();
        }
    }
	
    /** <p>Convert paragraph style declarations to LaTeX.</p>
     *  <p>Note: This method should not be called before the end, since it only
     *  declares used paragraph styles.</p>
     *  @param <code>ldp</code> the <code>LaTeXDocumentPortion</code> to add code to
     */
    void convertParStyles(LaTeXDocumentPortion ldp) {
        ldp.append("% Paragraph styles").nl();
        // Default paragraph style
        ParStyle style = wsc.getDefaultParStyle();
        if (style!=null) {
            String sFontName=style.getProperty(XMLString.STYLE_FONT_NAME);
            if (sFontName!=null){
                String sLaTeXfontFamily = convertFontDeclaration(sFontName,true);
                if (sLaTeXfontFamily!=null){
                    ldp.append("\\renewcommand\\familydefault{")
                       .append(sLaTeXfontFamily.substring(0,3))
                       .append("default}").nl();
                }
            } // TODO: Else read props directly from the style

            String s = nfssSeries(style.getProperty(XMLString.FO_FONT_WEIGHT),true);
            if (s!=null) {
                ldp.append("\\renewcommand\\seriesdefault{")
                   .append(s.substring(0,3))
                   .append("default}").nl();
            }

            s=nfssShape(style.getProperty(XMLString.FO_FONT_VARIANT),
                                 style.getProperty(XMLString.FO_FONT_STYLE),true);
            if (s!=null) {
                ldp.append("\\renewcommand\\shapedefault{")
                   .append(s.substring(0,3))
                   .append("default}").nl();
            }
           
            s=color(style.getProperty(XMLString.FO_COLOR),true);
            if (s!=null) {
                ldp.append("\\renewcommand\\normalcolor{")
                   .append(s).append("}").nl();
            }
        }

        // Main paragraph style, if specified in configuration
        ParStyle mainParStyle = wsc.getParStyle(config.mainParagraphStyle());
        if (mainParStyle!=null) {
            BeforeAfter ba = new BeforeAfter();
            parStyleNfss(mainParStyle,ba);
            ldp.append(ba.getBefore()).nl();
        }
         

        // Ordinary paragraph styles
        Enumeration styles = wsc.getParStyles().getStylesEnumeration();
        while (styles.hasMoreElements()) {
            style = (ParStyle) styles.nextElement();
            String sName = style.getName();
            if (parStyleNames.containsName(sName)) {
                if (parStyleMap.contains(sName)) {
                    ldp.append("\\newenvironment{style")
                       .append(parStyleNames.getExportName(sName))
                       .append("}{");
                    String s = style.getMasterPageName();
                    if (s!=null && s.length()>0) {
                        ldp.append("\\clearpage").nl();
                        useMasterPage(s,ldp);
                    }
                    ldp.append( parStyleMap.getBefore(sName) )
                       .append("}{")
                       .append( parStyleMap.getAfter(sName) )
                       .append("}").nl();
                }
                if (parStyleMapSimple.contains(sName)) {
                    ldp.append("\\newenvironment{style")
                       .append( parStyleNames.getExportName(sName) )
                       .append("*}{")
                       .append( parStyleMapSimple.getBefore(sName) )
                       .append("}{")
                       .append( parStyleMapSimple.getAfter(sName) )
                       .append("}").nl();
                }
            }
        }
    }
	
	////////////////////////////////////////////////////////////////////////
    // Heading styles and outline numbering


    /** <p>Convert heading styles and outline numbering to LaTeX.</p>
     *  <p>An array of stylenames to use is required: The OOo writer file format
     *  allows different paragraph styles to be applied to individual headings,
     *  so this is not included in the styles.
     *  LaTeX (and OOo Writer!) usually uses the same format for all headings.</p>
     *  @param sHeadingStyles    array of stylesnames to use for headings.
     *  @param ldp    the <code>LaTeXDocumentPortion</code> to add definitions to.
     */
    void convertHeadingStyles(String[] sHeadingStyles, LaTeXDocumentPortion ldp) {
        // Find the max level for headings
        int nMaxLevel = 0;
        for (int i=1; i<=5; i++) { if (sHeadingStyles[i]!=null) { nMaxLevel=i; } }
        if (nMaxLevel==0) { return; } // no headings, nothing to do!
		       
        if (config.ignoreHeadingStyles()) {
            ldp.append("% Outline numbering").nl();
        }
        else {
            ldp.append("% Headings and outline numbering").nl()
               .append("\\makeatletter").nl();
        }

        // Paragraph style for headings:
        if (!config.ignoreHeadingStyles()) {
            for (int i=1; i<=nMaxLevel; i++) {
                if (sHeadingStyles[i]!=null) {
                    ParStyle style = wsc.getParStyle(sHeadingStyles[i]);
                    if (style!=null) {
                        String sDecl="", sCsBegin="", sCsEnd="";

                        if ("page".equals(style.getProperty(XMLString.FO_BREAK_BEFORE))) {
                            sDecl = "\\clearpage";
                        }
    
	// temp:
	                    BeforeAfter ba = new BeforeAfter();
                        parStyleNfss(style,ba);
						sDecl+=ba.getBefore();
		    			
                        String s=alignment(style.getProperty(XMLString.FO_TEXT_ALIGN));
                        if (s!=null) { sDecl+=s; }
		
                        s = babelLanguage(style.getProperty(XMLString.FO_LANGUAGE),false);
                        if (s!=null) {
                            sCsBegin = s+"{"; sCsEnd = "}";
                        }
		
                        s = changeCase(style.getProperty(XMLString.FO_TEXT_TRANSFORM));
                        if (s!=null) {
                            sCsBegin += s+"{"; sCsEnd = "}"+sCsEnd;;
                        }
	    	
                        if (config.useUlem()) {
                            s = underline(style.getProperty(XMLString.STYLE_TEXT_UNDERLINE));
                            if (s!=null) { bNeedUlem=true; sCsBegin+=s+"{"; sCsEnd="}"+sCsEnd; }
                            s = crossout(style.getProperty(XMLString.STYLE_TEXT_CROSSING_OUT));
                            if (s!=null) { bNeedUlem=true; sCsBegin+=s+"{"; sCsEnd="}"+sCsEnd; }
                        }
				
                        String sMarginTop = style.getAbsoluteProperty(XMLString.FO_MARGIN_TOP);
                        if (sMarginTop==null) { sMarginTop = "0mm"; }
                        String sMarginBottom = style.getAbsoluteProperty(XMLString.FO_MARGIN_BOTTOM);
                        if (sMarginBottom==null) { sMarginBottom = "0mm"; }
                        String sMarginLeft = style.getAbsoluteProperty(XMLString.FO_MARGIN_LEFT);
                        if (sMarginLeft==null) { sMarginLeft = "0mm"; }
    
                        String sSecName = sectionName(i);
                        ldp.append("\\newcommand\\cs").append(sSecName).append("[1]{")
                           .append(sCsBegin).append("#1").append(sCsEnd).append("}").nl()
                           .append("\\renewcommand\\").append(sSecName)
                           .append("{\\@startsection{").append(sSecName).append("}{"+i)
                           .append("}{"+sMarginLeft+"}{-").append(sMarginTop)
                           .append("}{").append(sMarginBottom).append("}{");
                        s = style.getMasterPageName();
                        if (s!=null && s.length()>0) {
                            ldp.append("\\clearpage").nl();
                            useMasterPage(s,ldp);
                        }
	                    ldp.append(sDecl)
                           .append("\\cs").append(sSecName).append("}}").nl();
                    }
                }
            }
        }

        // redefine formatting of section counters
        // simplified if the user wants to ignore formatting
        if (!config.ignoreHeadingStyles()) {
            ldp.append("\\renewcommand\\@seccntformat[1]{")
               .append("\\csname @textstyle#1\\endcsname{\\csname the#1\\endcsname}")
               .append("\\csname @distance#1\\endcsname}").nl();
        }

        // Collect numbering styles and set secnumdepth
        int nSecnumdepth = 5;
        ListStyle outline = wsc.getOutlineStyle();
        String[] sNumFormat = new String[6];
        for (int i=nMaxLevel; i>=1; i--) {
            sNumFormat[i] = numFormat(outline.getLevelProperty(i,
                               XMLString.STYLE_NUM_FORMAT));
            if (sNumFormat[i]==null || "".equals(sNumFormat[i])) {
                nSecnumdepth = i-1;
            }
        }
        ldp.append("\\setcounter{secnumdepth}{"+nSecnumdepth+"}").nl();

        for (int i=1; i<=nMaxLevel; i++) {
            if (sNumFormat[i]==null || "".equals(sNumFormat[i])) {
                // no numbering at this level
                if (!config.ignoreHeadingStyles()) {
                    ldp.append("\\newcommand\\@distance")
                       .append(sectionName(i)).append("{}").nl()
                       .append("\\newcommand\\@textstyle")
                       .append(sectionName(i)).append("[1]{#1}").nl();
                }
                //ldp.append("\\renewcommand\\the")
                //   .append(sectionName(i))
                //   .append("{}").nl();
            }
            else {
                if (!config.ignoreHeadingStyles()) {
                    // Distance between label and text:
                    String sDistance = outline.getLevelProperty(i,XMLString.TEXT_MIN_LABEL_DISTANCE);
                    ldp.append("\\newcommand\\@distance")
                       .append(sectionName(i)).append("{");
                    if (sDistance!=null) { 
                        ldp.append("\\hspace{").append(sDistance).append("{");
                    }
                    ldp.append("}").nl();
                    // Textstyle to use for label:
                    String sStyleName = outline.getLevelProperty(i,XMLString.TEXT_STYLE_NAME);
                    ldp.append("\\newcommand\\@textstyle")
                       .append(sectionName(i)).append("[1]{");
                    useTextStyleBegin(sStyleName,ldp);
                    ldp.append("#1");
                    useTextStyleEnd(sStyleName,ldp);
                    ldp.append("}").nl();
                }

                // The label:
                String sPrefix = outline.getLevelProperty(i,XMLString.STYLE_NUM_PREFIX);
                String sSuffix = outline.getLevelProperty(i,XMLString.STYLE_NUM_SUFFIX);
                int nLevels = Misc.getPosInteger(outline.getLevelProperty(i,
                                      XMLString.TEXT_DISPLAY_LEVELS),1);
                ldp.append("\\renewcommand\\the")
                   .append(sectionName(i))
                   .append("{");
                StringBuffer labelbuf = new StringBuffer();
                if (sPrefix!=null) { labelbuf.append(sPrefix); }
                for (int j=i-nLevels+1; j<i; j++) {
                    labelbuf.append(sNumFormat[j])
                            .append("{").append(sectionName(j)).append("}")
                            .append(".");
                }
                labelbuf.append(sNumFormat[i])
                        .append("{").append(sectionName(i)).append("}");
                if (sSuffix!=null) { labelbuf.append(sSuffix); }
                if (config.ignoreHeadingStyles()) {
                    ldp.append(labelbuf.toString().trim());
                }
                else {
                    ldp.append(labelbuf.toString());
                }
                ldp.append("}").nl();
            }
            
        }

        if (!config.ignoreHeadingStyles()) {
            ldp.append("\\makeatother").nl();
        }

    }
	
	////////////////////////////////////////////////////////////////////////
    // List styles

	
    /** <p>Use a list style to LaTeX: Puts code <i>before</i> the list.</p>
     *  <p>Does nothing if the supplied name is null or if the list style doesn't exist.</p>
     *  @param <code>sName</code> the name of the list style
     *  @param <code>ldp</code> the <code>LaTeXDocumentPortion</code> to add code to
     */
	 
    private void defineLabels(ListStyle style, LaTeXDocumentPortion ldp, int nMaxLevel) {
        // Determine the number format for each required level
        String[] sNumFormat = new String[6];
        for (int i=1; i<=nMaxLevel; i++) {
            sNumFormat[i] = numFormat(style.getLevelProperty(i,
                               XMLString.STYLE_NUM_FORMAT));
        }
        // Export labels for all required levels
        for (int i=1; i<=nMaxLevel; i++) {
            String sLevelType = style.getLevelType(i);
            if (XMLString.TEXT_LIST_LEVEL_STYLE_NUMBER.equals(sLevelType)) {
                ldp.append("\\renewcommand\\labelenum").append(Misc.int2roman(i)).append("{");
                // Get properties
                String sStyleName = style.getLevelProperty(i,XMLString.TEXT_STYLE_NAME);
                String sPrefix = style.getLevelProperty(i,XMLString.STYLE_NUM_PREFIX);
                String sSuffix = style.getLevelProperty(i,XMLString.STYLE_NUM_SUFFIX);
                int nLevels = Misc.getPosInteger(style.getLevelProperty(i,XMLString.TEXT_DISPLAY_LEVELS),1);
                // Apply style
                if (!config.ignoreListLabelStyles()) { useTextStyleBegin(sStyleName,ldp); }
                StringBuffer labelbuf = new StringBuffer();
                if (sPrefix!=null) { labelbuf.append(sPrefix); }
                for (int j=i-nLevels+1; j<i; j++) {
                    if (XMLString.TEXT_LIST_LEVEL_STYLE_NUMBER.equals(style.getLevelType(j))) {
                        labelbuf.append(sNumFormat[j])
                                .append("{enum").append(Misc.int2roman(j)).append("}.");
                    }
                } 
                labelbuf.append(sNumFormat[i])
                        .append("{enum").append(Misc.int2roman(i)).append("}");
                if (sSuffix!=null) { labelbuf.append(sSuffix); }
                ldp.append(labelbuf.toString().trim());
                if (!config.ignoreListLabelStyles()) { useTextStyleEnd(sStyleName,ldp); }
                ldp.append("}").nl();
            }
            else if (XMLString.TEXT_LIST_LEVEL_STYLE_BULLET.equals(sLevelType)) {
                ldp.append("\\renewcommand\\labelitem").append(Misc.int2roman(i)).append("{");
                // Get properties			
                String sStyleName = style.getLevelProperty(i,XMLString.TEXT_STYLE_NAME);
                String sBullet = style.getLevelProperty(i,XMLString.TEXT_BULLET_CHAR);
                // Apply style
                if (!config.ignoreListLabelStyles()) { useTextStyleBegin(sStyleName,ldp); }
                if (sBullet!=null) {
                    // Bullets are usually symbols, so this should be OK:
                    ldp.append(i18n.convert(sBullet,false,"en"));
                }
                if (!config.ignoreListLabelStyles()) { useTextStyleEnd(sStyleName,ldp); }
                ldp.append("}").nl();
            }
            else if (XMLString.TEXT_LIST_LEVEL_STYLE_IMAGE.equals(sLevelType)) {
                // TODO: Support images!
                ldp.append("\\renewcommand\\labelitem").append(Misc.int2roman(i))
                   .append("{\\textbullet}").nl();
            }
        }
    }

    void useListStyle(String sName, LaTeXDocumentPortion ldp, int nMaxLevel) {
	
	
        ListStyle style = wsc.getListStyle(sName);
        if (style==null) { return; }
        if (style.isAutomatic()){
            defineLabels(style,ldp,nMaxLevel);
        }
        else {
            if (!listStyleNames.containsName(sName)) { listStyleNames.addName(sName); }
            ldp.append("\\liststyle").append(listStyleNames.getExportName(sName)).nl();
        }
    }
	
    /** <p>Convert list style declarations to LaTeX.</p>
     *  <p>Note: This method should not be called before the end, since it only
     *  declares used list styles.</p>
     *  @param <code>ldp</code> the <code>LaTeXDocumentPortion</code> to add code to
     */
    void convertListStyles(LaTeXDocumentPortion ldp) {
        if (!listStyleNames.isEmpty()) {
            ldp.append("% List styles").nl();
        }
        Enumeration styles = wsc.getListStyles().getStylesEnumeration();
        while (styles.hasMoreElements()) {
            ListStyle style = (ListStyle) styles.nextElement();
            String sName = style.getName();
            if (listStyleNames.containsName(sName)) {
                ldp.append("\\newcommand\\liststyle")
                   .append(listStyleNames.getExportName(sName))
                   .append("{");
                defineLabels(style,ldp,4); // don't know the max. level with soft formatting
                ldp.append("}").nl();
            }
        }
    }
	
	////////////////////////////////////////////////////////////////////////
    // Table styles
	
    /** <p> Return a TableFormatter for this style */
    TableFormatter getTableFormatter(TableGridModel table,
        boolean bAllowLongtable) {
        return new TableFormatter(wsc,table,bAllowLongtable);
    }

	////////////////////////////////////////////////////////////////////////
    // Page styles


    /** <p>Use a Master Page (pagestyle in LaTeX)</p>
     *  @param sName    name of the master page to use
     *  @param ldp      the <code>LaTeXDocumentPortion</code> to add code to.
     */
    void useMasterPage(String sName, LaTeXDocumentPortion ldp) {
        if (config.ignorePageFormatting()) return;
        MasterPage style = wsc.getMasterPage(sName);
        if (style==null) { return; }
        String sNextName = style.getProperty(XMLString.STYLE_NEXT_STYLE_NAME);
        MasterPage nextStyle = wsc.getMasterPage(sNextName);
        if (style==nextStyle || nextStyle==null) {
            ldp.append("\\pagestyle{")
               .append(masterPageNames.getExportName(sName))
               .append("}").nl();
        }
        else {
            ldp.append("\\pagestyle{")
               .append(masterPageNames.getExportName(sNextName))
               .append("}").nl()
               .append("\\thispagestyle{")
               .append(masterPageNames.getExportName(sName))
               .append("}").nl();
        }
    }
	 
    /** <p>Convert a page master (pagesize, margins etc.)
     *  @param sName   name of the page master to use 
     *  @param ldp     the <code>LaTeXDocumentPortion</code> to add code to.
     */
    void convertPageMaster(String sName, LaTeXDocumentPortion ldp) {
        if (config.ignorePageFormatting()) { return; }
        PageMaster style = wsc.getPageMaster(sName);
        if (style==null) { return; }
		
        String sHeadHeight = "0cm";
        String sHeadSep = "0cm";
        if (style.hasHeaderStyle()) {
            sHeadHeight = "12pt";
            sHeadSep = style.getHeaderProperty(XMLString.FO_MARGIN_BOTTOM);
            if (sHeadSep==null) { sHeadSep = "0cm"; }
        }
		
        String sFootHeight = "0cm";
        String sFootSep = "0cm";
        if (style.hasFooterStyle()) {
            sFootHeight = "12pt";
            sFootSep = style.getFooterProperty(XMLString.FO_MARGIN_TOP);
            if (sFootSep==null) { sFootSep = "0cm"; }
        }
		
        // Page dimensions
        ldp.append("\\setlength\\paperwidth{")
           .append(style.getAbsoluteProperty(XMLString.FO_PAGE_WIDTH))
           .append("}")
           .append("\\setlength\\paperheight{")
           .append(style.getAbsoluteProperty(XMLString.FO_PAGE_HEIGHT))
           .append("}");
        // PDF page dimensions, only if hyperref.sty is not loaded
        if (config.getBackend()==Config.PDFTEX && !config.useHyperref()) {
            ldp.append("\\setlength\\pdfpagewidth{")
               .append(style.getAbsoluteProperty(XMLString.FO_PAGE_WIDTH))
               .append("}")
               .append("\\setlength\\pdfpageheight{")
               .append(style.getAbsoluteProperty(XMLString.FO_PAGE_HEIGHT))
               .append("}");
        }
        // Page starts in upper left corner of paper!!
        ldp.append("\\setlength\\voffset{-1in}");
        ldp.append("\\setlength\\hoffset{-1in}");
        // Top margin
        ldp.append("\\setlength\\topmargin{")
           .append(style.getProperty(XMLString.FO_MARGIN_TOP))
           .append("}");
        // Header and footer
        ldp.append("\\setlength\\headheight{").append(sHeadHeight).append("}");
        ldp.append("\\setlength\\headsep{").append(sHeadSep).append("}");
        ldp.append("\\setlength\\footskip{")
           .append(sFootHeight).append("+").append(sFootSep).append("}");
        // Bottom margin is set indirectly via text height
        ldp.append("\\setlength\\textheight{")
           .append(style.getProperty(XMLString.FO_PAGE_HEIGHT))
           .append("-").append(style.getProperty(XMLString.FO_MARGIN_TOP))
           .append("-").append(style.getProperty(XMLString.FO_MARGIN_BOTTOM))
           .append("-").append(sHeadSep)
           .append("-").append(sHeadHeight)
           .append("-").append(sFootSep)
           .append("-").append(sFootHeight).append("}");
        // Left margin
        ldp.append("\\setlength\\oddsidemargin{")
           .append(style.getProperty(XMLString.FO_MARGIN_LEFT))
           .append("}");
        // Left margin for even (left) pages; only for mirrored page master
        if ("mirrored".equals(style.getPageUsage())) {
            ldp.append("\\setlength\\evensidemargin{")
               .append(style.getProperty(XMLString.FO_MARGIN_RIGHT))
               .append("}");
        }
        // Right margin is set indirectly via text width
        ldp.append("\\setlength\\textwidth{")
           .append(style.getProperty(XMLString.FO_PAGE_WIDTH))
           .append("-")
           .append(style.getProperty(XMLString.FO_MARGIN_LEFT))
           .append("-")
           .append(style.getProperty(XMLString.FO_MARGIN_RIGHT))
           .append("}").nl();
        // Set additional lengths
        //ldp.append("\\@colht\\textheight\\@colroom\\textheight\\vsize\\textheight").nl();

        // Define columns
        //if (style.getColCount()>1) { ldp.append("\\twocolumn").nl(); }
        //else { ldp.append("\\onecolumn").nl(); }

        // Page number
        String sNumFormat = style.getProperty(XMLString.STYLE_NUM_FORMAT);
        if (sNumFormat!=null) {
            ldp.append("\\renewcommand\\thepage{")
               .append(numFormat(sNumFormat))
               .append("{page}}").nl();
        }
        String sPageNumber = style.getProperty(XMLString.STYLE_FIRST_PAGE_NUMBER);
        if (sPageNumber!=null && !sPageNumber.equals("continue")) {
           ldp.append("\\setcounter{page}{")
              .append(Integer.toString(Misc.getPosInteger(sPageNumber,0)))
              .append("}").nl();
        }
        // Footnote rule
        // TODO: Support alignment.
        String sBefore = style.getFootnoteProperty(XMLString.STYLE_DISTANCE_BEFORE_SEP);
        if (sBefore==null) { sBefore = "1mm"; }
        String sAfter = style.getFootnoteProperty(XMLString.STYLE_DISTANCE_AFTER_SEP);
        if (sAfter==null) { sAfter = "1mm"; }
        String sHeight = style.getFootnoteProperty(XMLString.STYLE_WIDTH);
        if (sHeight==null) { sHeight = "0.2mm"; }
        String sWidth = style.getFootnoteProperty(XMLString.STYLE_REL_WIDTH);
        if (sWidth==null) { sWidth = "25%"; }
        sWidth=Float.toString(Misc.getFloat(sWidth.substring(0,sWidth.length()-1),1)/100);
        String sColor = color(style.getFootnoteProperty(XMLString.STYLE_COLOR),false);
 
        ldp.append("\\setlength{\\skip\\footins}{").append(sBefore).append("}");
        ldp.append("\\renewcommand\\footnoterule{\\vspace*{-").append(sHeight)
           .append("}\\noindent").append(sColor).append("{\\rule{")
           .append(sWidth).append("\\columnwidth}{").append(sHeight)
           .append("}}\\vspace*{").append(sAfter)
           .append("}}").nl();
    }
	
	////////////////////////////////////////////////////////////////////////
    // {Foot|end}note configuration


    /** <p>Convert {foot|end}notes configuration.</p>
     *  <p>Note: All {foot|end}notes are formatted with the default style for {foot|end}footnotes.
     *  (This doesn't conform with the file format specification, but in LaTeX
     *  all {foot|end}notes are usually formatted in a fixed style.)</p>
     *
     *  @param ldp    the <code>LaTeXDocumentPortion</code> to add definitions to.
     */
    void convertFootEndnotesConfiguration(PropertySet notes, String sType, LaTeXDocumentPortion ldp) {
        if (config.ignoreFootnotesConfiguration()) { return; }
        String sTypeShort;
        if (sType.equals("foot")) { sTypeShort="fn"; } else { sTypeShort="en"; }
        if (notes==null) { return; }
        ldp.append("% ").append(sType).append("notes configuration").nl()
           .append("\\makeatletter").nl();

        // The numbering style is controlled by \the{foot|end}note
		String sFormat = notes.getProperty(XMLString.STYLE_NUM_FORMAT);
        if (sFormat!=null) {
            ldp.append("\\renewcommand\\the").append(sType).append("note{")
               .append(numFormat(sFormat))
               .append("{").append(sType).append("note}}").nl();
        }
        
        // Number {foot|end}notes by sections
        if ("chapter".equals(notes.getProperty(XMLString.TEXT_START_NUMBERING_AT))) {
            ldp.append("\\@addtoreset{").append(sType).append("note}{section}").nl();
        }
		
        // Set start value offset (default 0)
        int nStartValue = Misc.getPosInteger(notes.getProperty(XMLString.TEXT_START_VALUE),0);
        if (nStartValue!=0) {
            ldp.append("\\setcounter{").append(sType).append("note}{"+nStartValue+"}").nl();
        }
        
        // The formatting of the {foot|end}note citation is controlled by \@make{fn|en}mark
        String sCitBodyStyle = notes.getProperty(XMLString.TEXT_CITATION_BODY_STYLE_NAME);
        if (sCitBodyStyle!=null && wsc.getTextStyle(sCitBodyStyle)!=null) {
            ldp.append("\\renewcommand\\@make").append(sTypeShort).append("mark{\\mbox{");
            useTextStyleBegin(sCitBodyStyle,ldp);
            ldp.append("\\@the").append(sTypeShort).append("mark");
            useTextStyleEnd(sCitBodyStyle,ldp);
            ldp.append("}}").nl();
        }
		
        // The layout and formatting of the {foot|end}note is controlled by \@make{fn|en}text
        String sCitStyle = notes.getProperty(XMLString.TEXT_CITATION_STYLE_NAME);
        String sStyleName = notes.getProperty(XMLString.TEXT_DEFAULT_STYLE_NAME);
        if (sStyleName!=null) {
            StyleWithProperties style = wsc.getParStyle(sStyleName);
            if (style!=null) {
                ldp.append("\\renewcommand\\@make").append(sTypeShort).append("text[1]{\\noindent");
                useTextStyleBegin(sCitStyle,ldp);
                ldp.append("\\@the").append(sTypeShort).append("mark\\ ");
                useTextStyleEnd(sCitStyle,ldp);
                useParStyleBeginSimple(sStyleName,ldp);
                ldp.append("#1");
                useParStyleEndSimple(sStyleName,ldp);
                ldp.append("}").nl();
            }	 
        }
        
        ldp.append("\\makeatother").nl();		
    }

    /** <p>Convert footnotes configuration.</p>
     *  <p>Note: All footnotes are formatted with the default style for footnotes.
     *  (This doesn't conform with the file format specification, but in LaTeX
     *  all footnotes are usually formatted in a fixed style.)</p>
     *
     *  @param ldp    the <code>LaTeXDocumentPortion</code> to add definitions to.
     */
    void convertFootnotesConfiguration(LaTeXDocumentPortion ldp) {
        // Note: Continuation notices are not supported in LaTeX
        // TODO: Support text:footnotes-postion="document" (footnotes as endnotes)
        // TODO: Support text:start-numbering-at="page" (footnpag.sty)
        convertFootEndnotesConfiguration(wsc.getFootnotesConfiguration(),"foot",ldp);
    }

    /** <p>Convert endnotes configuration.</p>
     *  <p>Note: This method automatically includes the package "endnotes.sty".</p>
     *  <p>Note: All endnotes are formatted with the default style for endnotes.
     *  (This doesn't conform with the file format specification, but in LaTeX
     *  all endnotes are usually formatted in a fixed style.)</p>
     *
     *  @param ldp    the <code>LaTeXDocumentPortion</code> to add definitions to.
     */
    void convertEndnotesConfiguration(LaTeXDocumentPortion ldp) {
        // Note: Continuation notices are not supported in LaTeX
        convertFootEndnotesConfiguration(wsc.getEndnotesConfiguration(),"end",ldp);
    }
	
	////////////////////////////////////////////////////////////////////////
    // Other

    void convertLanguageUsage(LaTeXDocumentPortion ldp) {
        Vector languages = new Vector();
        // Collect languages from text styles
        Enumeration enum = wsc.getTextStyles().getStylesEnumeration();
        while (enum.hasMoreElements()) {
            StyleWithProperties style = (StyleWithProperties) enum.nextElement();
            String sLang = babelLanguage(style.getProperty(XMLString.FO_LANGUAGE),true);
            if (sLang!=null && !languages.contains(sLang)) { languages.add(sLang); }
        }
        // Collect languages from paragraph styles
        enum = wsc.getParStyles().getStylesEnumeration();
        while (enum.hasMoreElements()) {
            StyleWithProperties style = (StyleWithProperties) enum.nextElement();
            String sLang = babelLanguage(style.getProperty(XMLString.FO_LANGUAGE),true);
            if (sLang!=null && !languages.contains(sLang)) { languages.add(sLang); }
        }
        // Read the default language from the default paragraph style
        String sDefaultLanguage = null;
        StyleWithProperties style = wsc.getDefaultParStyle();
        if (style!=null) { 
            sDefaultLanguage = babelLanguage(style.getProperty(XMLString.FO_LANGUAGE),true);
        }
        // If the document contains "anonymous" greek letters we need greek anyway:
        if (i18n.greek() && !languages.contains("greek")) languages.add("greek"); 
        // Load babel with the used languages
        CSVList babelopt = new CSVList(",");		
        Enumeration langenum = languages.elements();
        while (langenum.hasMoreElements()) {
            String sLang = (String) langenum.nextElement();
            if (!sLang.equals(sDefaultLanguage)) {
                if ("greek".equals(sLang) && i18n.polytonicGreek()) {
                    sLang = "polutonikogreek";
                }
                babelopt.addValue(sLang);
            }
        }
        // The default language must be the last one
        if (sDefaultLanguage!=null) {
            if ("greek".equals(sDefaultLanguage) && i18n.polytonicGreek()) {
                babelopt.addValue("polutonikogreek");
            }
            else {
                babelopt.addValue(sDefaultLanguage);
            }
        }
        if (!babelopt.isEmpty()) {
            ldp.append("\\usepackage[")
               .append(babelopt.toString())
               .append("]{babel}").nl();
        }
    }
	
    // Utility methods to apply individual style properties
    void applyCS(String s, boolean bDeclaration, BeforeAfter ba) {
        if (s!=null) { 
            if (bDeclaration) { ba.add(s,""); }
            else { ba.add(s+"{","}"); }
        }
    }
	
    void applyNfssFamily(StyleWithProperties style, boolean bDeclaration, BeforeAfter ba) {
        String sFontName=style.getProperty(XMLString.STYLE_FONT_NAME);
        if (sFontName!=null){
            String sLaTeXfontFamily = convertFontDeclaration(sFontName,bDeclaration);
            applyCS(sLaTeXfontFamily,bDeclaration,ba);
        } // TODO: Else read props directly from the style
    }

    void applyNfssSeries(StyleWithProperties style, boolean bDeclaration, BeforeAfter ba) {
        String s = nfssSeries(style.getProperty(XMLString.FO_FONT_WEIGHT),bDeclaration);
        applyCS(s,bDeclaration,ba);
    }

    void applyNfssShape(StyleWithProperties style, boolean bDeclaration, BeforeAfter ba) {
        String s = nfssShape(style.getProperty(XMLString.FO_FONT_VARIANT),
                             style.getProperty(XMLString.FO_FONT_STYLE),bDeclaration);
        applyCS(s,bDeclaration,ba);
    }
        
    void applyColor(StyleWithProperties style, boolean bDeclaration, BeforeAfter ba) {
        String s = color(style.getProperty(XMLString.FO_COLOR),bDeclaration);
        applyCS(s,bDeclaration,ba);
    }

    void applyLanguage(StyleWithProperties style, boolean bEnvironment, BeforeAfter ba) {
        String s = babelLanguage(style.getProperty(XMLString.FO_LANGUAGE),bEnvironment);
        if (s!=null) {
            if (bEnvironment) { ba.add("\\begin{otherlanguage}{"+s+"}","\\end{otherlanguage}"); }
            else { ba.add(s+"{","}"); }
        }
    }
	
    void applyNfssSize(StyleWithProperties style, boolean bDeclaration, BeforeAfter ba) {
        String s = nfssSize(style.getAbsoluteProperty(XMLString.FO_FONT_SIZE));
        if (s!=null) { 
            if (bDeclaration) { ba.add(s,""); }
            else { ba.add("{"+s+" ","}"); }
        }
    }

    void applyTextPosition(StyleWithProperties style, BeforeAfter ba) {
        String s = textPosition(style.getProperty(XMLString.STYLE_TEXT_POSITION));
        applyCS(s,false,ba);
    }
	
    void applyChangeCase(StyleWithProperties style, BeforeAfter ba) {
        String s = changeCase(style.getProperty(XMLString.FO_TEXT_TRANSFORM));
        applyCS(s,false,ba);
    }

    void applyUnderline(StyleWithProperties style, BeforeAfter ba) {
        if (!config.useUlem()) { return; }
        String s = underline(style.getProperty(XMLString.STYLE_TEXT_UNDERLINE));
        if (s!=null) bNeedUlem = true;
        applyCS(s,false,ba);
    }

    void applyCrossout(StyleWithProperties style, BeforeAfter ba) {
        if (!config.useUlem()) { return; }
        String s = crossout(style.getProperty(XMLString.STYLE_TEXT_CROSSING_OUT));
        if (s!=null) bNeedUlem = true;
        applyCS(s,false,ba);
    }

    // The remaining methods are static helpers to convert single style properties

    // Font change. These methods return the declaration form if the paramater
    // bDeclaration is true, and otherwise the command form
    
    static final String nfssFamily(String sFontFamily, String sFontFamilyGeneric,
                                   String sFontPitch, boolean bDeclaration){
        // Note: Defaults to rm
        // TODO: What about decorative, script, system?
        String sFamily;
        if ("fixed".equals(sFontPitch)) sFamily="tt";
        else if ("modern".equals(sFontFamilyGeneric)) sFamily="tt";
        else if ("swiss".equals(sFontFamilyGeneric)) sFamily="sf";
        else sFamily="rm";
        if (bDeclaration) return "\\"+sFamily+"family";
        else return "\\text"+sFamily;
    }
    
    static final String nfssSeries(String sFontWeight, boolean bDeclaration){
        String sSeries;
        if (sFontWeight==null) return null;
        if ("bold".equals(sFontWeight)) sSeries="bf";
        else sSeries="md";
        if (bDeclaration) return "\\"+sSeries+"series";
        else return "\\text"+sSeries;
    }
    
    static final String nfssShape(String sFontVariant, String sFontStyle, boolean bDeclaration){
        String sShape;
        if (sFontVariant==null && sFontStyle==null) return null;
        if ("small-caps".equals(sFontVariant)) sShape="sc";
        else if ("italic".equals(sFontStyle)) sShape="it";
        else if ("oblique".equals(sFontStyle)) sShape="sl";
        else sShape="up";
        if (bDeclaration) return "\\"+sShape+"shape";
        else return "\\text"+sShape;
    }
    
    static final String nfssSize(String sFontSize){
        if (sFontSize==null) return null;
        return "\\fontsize{"+sFontSize+"}{"+Misc.multiply("120%",sFontSize)+"}\\selectfont";
    }
    
    static final String color(String sColor, boolean bDeclaration){
        String sTeXcolor;
        if ("#000000".equalsIgnoreCase(sColor)) { sTeXcolor="{black}"; }
        else if ("#ff0000".equalsIgnoreCase(sColor)) { sTeXcolor="{red}"; }
        else if ("#00ff00".equalsIgnoreCase(sColor)) { sTeXcolor="{green}"; }
        else if ("#0000ff".equalsIgnoreCase(sColor)) { sTeXcolor="{blue}"; }
        else if ("#ffff00".equalsIgnoreCase(sColor)) { sTeXcolor="{yellow}"; }
        else if ("#ff00ff".equalsIgnoreCase(sColor)) { sTeXcolor="{magenta}"; }
        else if ("#00ffff".equalsIgnoreCase(sColor)) { sTeXcolor="{cyan}"; }
        //else if ("#ffffff".equalsIgnoreCase(sColor)) { sTeXcolor="{white}"; }
        else {
            float sRed, sGreen, sBlue;
            if (sColor==null || sColor.length()!=7) return null;
            sRed=(float)Misc.getIntegerFromHex(sColor.substring(1,3),0)/255;
            sGreen=(float)Misc.getIntegerFromHex(sColor.substring(3,5),0)/255;
            sBlue=(float)Misc.getIntegerFromHex(sColor.substring(5,7),0)/255;
            // avoid very bright colors (since we don't have background colors):
            if (sRed+sGreen+sBlue>2.7) { sTeXcolor="{black}"; }
            else { sTeXcolor="[rgb]{"+sRed+","+sGreen+","+sBlue+"}"; }
        }
        if (bDeclaration) return "\\color"+sTeXcolor;
        else return "\\textcolor"+sTeXcolor;
    }
    
    // other character formatting
        
    static final String textPosition(String sTextPosition){
        if (sTextPosition==null) return null;
        if (sTextPosition.startsWith("super")) return "\\textsuperscript";
        if (sTextPosition.startsWith("sub")) return "\\textsubscript";
        if (sTextPosition.startsWith("-")) return "\\textsubscript";
        if (sTextPosition.startsWith("0%")) return null;
        return "\\textsuperscript";
    }
    
    static final String changeCase(String sTextTransform){
        if ("lowercase".equals(sTextTransform)) return "\\MakeLowercase";
        if ("uppercase".equals(sTextTransform)) return "\\MakeUppercase";
        return null;
    }
    
    static final String babelLanguage(String sLanguage, boolean bLanguageOnly){
        if (sLanguage==null) return null;
        String sBabelLanguage = I18n.getBabelLanguage(sLanguage);
        if (sBabelLanguage==null) return null;
        if (bLanguageOnly) return sBabelLanguage;
        else return "\\foreignlanguage{"+sBabelLanguage+"}";
    }
	
    static final String underline(String sUnderline) {
        if (sUnderline==null) { return null; }
        if (sUnderline.equals("none")) { return null; }
        if (sUnderline.indexOf("wave")>=0) { return "\\uwave"; }
        return "\\uline";
    }
	
    static final String crossout(String sCrossout) {
        if (sCrossout==null) { return null; }
        if (sCrossout.equals("X")) { return "\\xout"; }
        if (sCrossout.equals("slash")) { return "\\xout"; }
        return "\\sout";
    }
    
    // paragraph formatting
    
    static final String alignment(String sTextAlign){
        if (sTextAlign==null) return null;
        if ("center".equals(sTextAlign)) return "\\centering";
        if ("start".equals(sTextAlign)) return "\\raggedright";
        if ("end".equals(sTextAlign)) return "\\raggedleft";
        return null;
    }
   
    static final String sectionName(int nLevel){
        switch (nLevel) {
            case 1: return "section";
            case 2: return "subsection";
            case 3: return "subsubsection";
            case 4: return "paragraph";
            case 5: return "subparagraph";
            default: return null;
        }
    }

    static final String numFormat(String sFormat){
        if ("1".equals(sFormat)) { return "\\arabic"; }
        else if ("i".equals(sFormat)) { return "\\roman"; }
        else if ("I".equals(sFormat)) { return "\\Roman"; }
        else if ("a".equals(sFormat)) { return "\\alph"; }
        else if ("A".equals(sFormat)) { return "\\Alph"; }
        else { return null; }
    }    

	
}