/************************************************************************
 *
 *  DocumentSerializerImpl.java
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2.1, as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA  02111-1307  USA
 *
 *  Copyright: 2002-2003 by Henrik Just
 *
 *  All Rights Reserved.
 * 
 *  Version 0.3.1 (2003-12-14)
 *
 */

package writer2latex.text;

import java.util.Enumeration;
import java.util.LinkedList;
import java.util.Hashtable;

import java.io.IOException;

//import javax.xml.parsers.DocumentBuilder;
//import javax.xml.parsers.DocumentBuilderFactory;

//import org.xml.sax.SAXException;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
//import org.w3c.dom.Element;
import org.w3c.dom.*;

//import org.openoffice.xmerge.converter.dom.DOMDocument;
import org.openoffice.xmerge.converter.xml.sxw.SxwDocument;
import org.openoffice.xmerge.converter.xml.EmbeddedObject;
//import org.openoffice.xmerge.converter.xml.EmbeddedBinaryObject;
import org.openoffice.xmerge.converter.xml.EmbeddedXMLObject;
import org.openoffice.xmerge.ConvertData;
import org.openoffice.xmerge.ConvertException;
import org.openoffice.xmerge.DocumentSerializer;

import writer2latex.util.*;
import writer2latex.office.*;

/**
 * <p>Formatted Text implementation of <code>DocumentDeserializer</code> 
 * for use by {@link
 * writer2latex.text.PluginFactoryImpl
 * PluginFactoryImpl}.</p>
 *
 * <p>This converts an OpenOffice Writer XML file to a Formatted text file.</p>
 *
 */
public final class DocumentSerializerImpl implements DocumentSerializer {
                        
    private FormattedTextDocument textDoc;
    private SxwDocument sxwDoc;
	
    // additional data used in the conversion:
    private Document styleDOM;
    private Document contentDOM;
    private WriterStyleCollection wsc;
    private ListCounter outlineNumbering;
    private Hashtable numberings = new Hashtable();
    private VBox mainBox;
    private ConvertData cd;
    private boolean bInHeading = false;
	
    // The collection of all headings (used for toc)
    NodeList headings = null;
	
    // Boxes for the footnotes and endnotes:
    private VBox footnotes = new VBox(true);
    private VBox endnotes = new VBox(true);
    // Sometimes we have to create inline text in a block context (footnote symbol)
    // We put it here and insert it in the first paragraph/heading to come:
    private String sAsap = "";
	
    /**
     *  <p>Initialises a new <code>DocumentSerializerImpl</code> using the.<br>
     *     supplied <code>Document</code></p>
     *
     * <p>The supplied document should be an {@link
     *    org.openoffice.xmerge.converter.xml.sxw.SxwDocument SxwDocument}
     *    object.</p>
     *
     *  @param  document  The <code>Document</code> to convert.    
     */
    public DocumentSerializerImpl(org.openoffice.xmerge.Document doc) {
        sxwDoc = (SxwDocument) doc;
        // load root xml streams. Note: When used as an export filter,
        // contentDOM contains *everything* (flat XML).
        styleDOM = sxwDoc.getStyleDOM();
        contentDOM = sxwDoc.getContentDOM();
        wsc = new WriterStyleCollection();
        textDoc = new FormattedTextDocument(sxwDoc.getName());
    }
    
    /**
     *  <p>Convert the data passed into the <code>DocumentSerializerImpl</code>
     *  constructor into Formatted format.</p>
     *
     *  <p>This method may or may not be thread-safe.  It is expected
     *  that the user code does not call this method in more than one
     *  thread.  And for most cases, this method is only done once.</p>
     *
     *  @return  <code>ConvertData</code> object to pass back the
     *           converted data.
     *
     *  @throws  ConvertException  If any conversion error occurs.
     *  @throws  IOException       If any I/O error occurs.
     */
    public ConvertData serialize() throws IOException, ConvertException {      
        mainBox = new VBox(false);
        cd = new ConvertData();
		
        // Load style info first
        wsc.loadStylesFromDOM(styleDOM,contentDOM);
        outlineNumbering = new ListCounter(wsc.getOutlineStyle());
		
        // Traverse the body
        NodeList list = contentDOM.getElementsByTagName(XMLString.OFFICE_BODY);
        int nLen = list.getLength();
        if (nLen > 0) {
            Node body = list.item(0);
            traverseBlockText(body, mainBox);
        }
		
        // Add footnotes and endnotes
        mainBox.addBox(footnotes);
        mainBox.addBox(endnotes);

        textDoc.setContent(mainBox);
        cd.addDocument(textDoc);
        return cd;
    }

    ////////////////////////////////////////////////////////////////////
    // Methods to handle style

    private int lengthToNumber(String s) { 
        if (s==null) { return 0; }
        String sFactor = "0%";
        if (s.endsWith("cm")) sFactor = "600%";
        if (s.endsWith("mm")) sFactor = "60%";
        if (s.endsWith("in")) sFactor = "240%";
        String s1 = Misc.multiply(sFactor,s);
        int nEnd = s1.indexOf(".");
        if (nEnd<0) { nEnd = s1.length(); }
        return Integer.valueOf(s1.substring(0,nEnd)).intValue();
    }
	
    // Create a PBox formatted with a paragraph style   
    private PBox getStyledPBox(String sStyleName) {
        StyleWithProperties style = wsc.getParStyle(sStyleName);
        int nMarginLeft = 0;
        int nMarginRight = 0;
        int nIndent = 0;
        int nAlign = PBox.LEFT;

        if (style!=null) {
            nMarginLeft = lengthToNumber(style.getAbsoluteProperty(XMLString.FO_MARGIN_LEFT));
            nMarginRight = lengthToNumber(style.getAbsoluteProperty(XMLString.FO_MARGIN_RIGHT));
            nIndent = lengthToNumber(style.getAbsoluteProperty(XMLString.FO_TEXT_INDENT));
            String sTextAlign = style.getProperty(XMLString.FO_TEXT_ALIGN);			
            if ("center".equals(sTextAlign)) nAlign = PBox.CENTER;
            if ("start".equals(sTextAlign)) nAlign = PBox.LEFT; 
            if ("end".equals(sTextAlign)) nAlign = PBox.RIGHT;

        }

        return new PBox(nAlign,nMarginLeft,nMarginRight,nIndent);
    } 

    // Create a PBox formatted with a paragraph style + a list style
    private PBox getStyledPBox(String sStyleName, String sListStyleName, int nLevel) {
        // TODO
        return new PBox(PBox.LEFT,0,0,0);
    }
	
    // Create a FBox formatted with a frame style
    private FBox getStyledFBox(String sStyleName) {
        // TODO
        return new FBox(false,true,true,true,true);
    }

    // Methods to traverse the content

    private void handleSection(Node onode, VBox box) {
	    // section style (columns) are ignored
        // TODO: change this??
        traverseBlockText(onode, box);
    }

    private void traverseBlockText(Node onode, VBox box) {
        if (onode.hasChildNodes()) {
            NodeList nList = onode.getChildNodes();
            int len = nList.getLength();
            
            for (int i = 0; i < len; i++) {
                Node child = nList.item(i);
                
                if (child.getNodeType() == Node.ELEMENT_NODE) {
                    String nodeName = child.getNodeName();
                    
                    if (nodeName.equals(XMLString.TEXT_P)) {
                        handleParagraph(child,box);
                    }

                    else if(nodeName.equals(XMLString.TEXT_H)) {
                        handleHeading(child,box);
                    }
                    
                    else if (nodeName.equals(XMLString.TEXT_UNORDERED_LIST)) {
                        handleUL(child,1,null,box);
                    }
                    
                    else if (nodeName.equals(XMLString.TEXT_ORDERED_LIST)) {
                        handleOL(child,1,null,box);
                    }
                    else if (nodeName.equals(XMLString.TABLE_TABLE)) {
                        handleTable(child,box);
                    }

                    else if (nodeName.equals(XMLString.TEXT_SECTION)) {
                        handleSection(child,box);
                    }

                    else if (nodeName.equals(XMLString.TEXT_TABLE_OF_CONTENT)) {
                        handleTOC(child,box);
                    }

                    else if (nodeName.equals(XMLString.TEXT_ILLUSTRATION_INDEX)) {
                        // ignore
                    }

                    else if (nodeName.equals(XMLString.TEXT_TABLE_INDEX)) {
                        // ignore
                    }

                    else if (nodeName.equals(XMLString.TEXT_OBJECT_INDEX)) {
                        // ignore
                    }

                    else if (nodeName.equals(XMLString.TEXT_USER_INDEX)) {
                        // ignore
                    }

                    else if (nodeName.equals(XMLString.TEXT_ALPHABETICAL_INDEX)) {
                        // ignore
                    }

                    else if (nodeName.equals(XMLString.TEXT_BIBLIOGRAPHY)) {
                        handleBibliography(child,box);
                    }

                    else if (nodeName.equals(XMLString.TABLE_SUB_TABLE)) {
                        handleTable(child,box);
                    }

                    else if (nodeName.equals(XMLString.TEXT_SEQUENCE_DECLS)) {
                        // this should not happen inside sections, but we don't care
                        //handleSeqeuenceDecls(child);
                    }
                    // other tags are ignored
                }
            }
        }
    }
    
    /*
     * Process a table:table tag
     */
    private void handleTable(Node onode, VBox box) {
	/* postponed until TBox is in place!
        TableGridModel tgm = new TableGridModel(onode);
        int nRowCount = tgm.getRowCount();
        int nColCount = tgm.getColCount();
        // Create table
        TBox table = new TBox(nRowCount,nColCount);
        // TODO: Apply table style??
        // TODO: Apply column styles??
        // Create rows
        for (int nRow=0; nRow<nRowCount; nRow++) {
            for (int nCol=0; nCol<nColCount; nCol++) {
                Node cell = tgm.getCell(nRow,nCol);
                if (XMLString.TABLE_TABLE_CELL.equals(cell.getNodeName())) {
                    String sStyleName = Misc.getAttribute(cell,XMLString.TABLE_STYLE_NAME);
                    // TODO: Borders
                    int nRowSpan = Misc.getPosInteger(
                        Misc.getAttribute(cell,XMLString.TABLE_NUMBER_ROWS_SPANNED),1);
                    int nColSpan = Misc.getPosInteger(
                        Misc.getAttribute(cell,XMLString.TABLE_NUMBER_COLUMNS_SPANNED),1);
                    table.formatCell(nRow,nCol,nRowSpan,nColSpan,false,false,false,false);                    
                    traverseBlockText(cell,table.getCell(nRow,nCol));
                }
                else if (XMLString.TABLE_COVERED_TABLE_CELL.equals(cell.getNodeName())) {
                    // covered table cells are ignored
                }
            }
        }*/
    }

    /*
     * Process a text:h tag
     */
    private void handleHeading(Node onode, VBox box) {
        int nLevel = Misc.getPosInteger(Misc.getAttribute(onode, XMLString.TEXT_LEVEL),1);
        String sStyleName = Misc.getAttribute(onode,XMLString.TEXT_STYLE_NAME);
        PBox heading = getStyledPBox(sStyleName);
        box.addBox(heading);
        // Prepend sAsap
        heading.addText(sAsap); sAsap = "";
			
        // Prepend numbering
        heading.addText( outlineNumbering.step(nLevel).getLabel() );
			
        bInHeading = true;
        traverseInlineText(onode,heading);
        bInHeading = false;
    }

    /*
     * Process a text:p tag
     */
    private void handleParagraph(Node onode, VBox box) {
        String sStyleName = Misc.getAttribute(onode,XMLString.TEXT_STYLE_NAME);
        PBox par = getStyledPBox(sStyleName);
        box.addBox(par);
        // Prepend sAsap
        par.addText(sAsap); sAsap = "";
			
        traverseInlineText(onode,par);
    }
    
    /*
     * Process a text:ordered-list tag.  
     */
    private void handleOL (Node onode, int nLevel, String sStyleName, VBox box) {
        String sStyleName1 = Misc.getAttribute(onode,XMLString.TEXT_STYLE_NAME);
        if (sStyleName1!=null) { sStyleName = sStyleName1; }
        traverseList(onode,nLevel,sStyleName,box);
    }

    /*
     * Process a text:unordered-list tag.  
     */
    private void handleUL (Node onode, int nLevel, String sStyleName, VBox box) {
        String sStyleName1 = Misc.getAttribute(onode,XMLString.TEXT_STYLE_NAME);
        if (sStyleName1!=null) { sStyleName = sStyleName1; }
        traverseList(onode,nLevel,sStyleName,box);
    }

    /*
     * Process the contents of a list
     */
    private void traverseList (Node onode, int nLevel, String sStyleName, VBox box) {
        if (onode.hasChildNodes()) {
            // Render list in a compact box:
            VBox list = new VBox(true);
            box.addBox(list);
            // Get the numbering, if any
            ListCounter listCounter = null;
            if (numberings.containsKey(sStyleName)) {
                listCounter = (ListCounter) numberings.get(sStyleName);
            }
            else {
                ListStyle style = wsc.getListStyle(sStyleName);
                if (style!=null) {
                    listCounter = new ListCounter(style);
                    numberings.put(sStyleName,listCounter);
                }
            }
            // Traverse the list
            NodeList nList = onode.getChildNodes();
            int len = nList.getLength();
            
            for (int i = 0; i < len; i++) {
                Node child = nList.item(i);
                
                if (child.getNodeType() == Node.ELEMENT_NODE) {
                    String nodeName = child.getNodeName();
                    
                    if (nodeName.equals(XMLString.TEXT_LIST_ITEM)) {
                        boolean bRestart = "true".equals(Misc.getAttribute(child,
                            XMLString.TEXT_RESTART_NUMBERING));
                        int nStartValue = Misc.getPosInteger(Misc.getAttribute(child,
                            XMLString.TEXT_START_VALUE),1);
                        sAsap += listCounter.step(nLevel).getLabel() + " ";
                        traverseListItem(child,nLevel,sStyleName,list);
                    }
                    if (nodeName.equals(XMLString.TEXT_LIST_HEADER)) {
                        // TODO: What to do??
                        traverseListItem(child,nLevel,sStyleName,list);
                    }
                }
            }
        }
    }
    
    
    /*
     * Process the contents of a list item
     * (a list header should only contain paragraphs, but we don't care)
     */
    private void traverseListItem (Node onode, int level, String styleName, VBox box) {
        if (onode.hasChildNodes()) {
            NodeList nList = onode.getChildNodes();
            int len = nList.getLength();
            
            for (int i = 0; i < len; i++) {
                Node child = nList.item(i);
                
                if (child.getNodeType() == Node.ELEMENT_NODE) {
                    String nodeName = child.getNodeName();
                    
                    if (nodeName.equals(XMLString.TEXT_P)) {
                        handleParagraph(child,box);
                    }
                    
                    if (nodeName.equals(XMLString.TEXT_ORDERED_LIST)) {
                        handleOL(child,level+1,styleName,box);
                    }
                    
                    if (nodeName.equals(XMLString.TEXT_UNORDERED_LIST)) {
                        handleUL(child,level+1,styleName,box);
                    }
                }
            }
        }
    }
    
    /* Process table of contents
     */
    private void handleTOC (Node onode, VBox box) {
        if (headings==null) { // collect all headings
            headings = contentDOM.getElementsByTagName(XMLString.TEXT_H);
        }
        Node source = Misc.getChildByTagName(onode,XMLString.TEXT_TABLE_OF_CONTENT_SOURCE);
        if (source!=null) {
            if ("chapter".equals(Misc.getAttribute(source,XMLString.TEXT_INDEX_SOURCE))) {
                // TODO
            }
            else { // generate toc
                // Generate title
                Node title = Misc.getChildByTagName(source,XMLString.TEXT_INDEX_TITLE_TEMPLATE);
                if (title!=null) {
                    String sStyleName = Misc.getAttribute(title,XMLString.TEXT_STYLE_NAME);
                    PBox par = getStyledPBox(sStyleName);
                    box.addBox(par);
                    par.addText(traversePCDATA(title).toUpperCase());
                }
                // Collect style names for entries
                // TODO: Should read the entire template
                String sEntryStyleName[] = new String[11];
                if (source.hasChildNodes()) {
                    NodeList nl = source.getChildNodes();
                    int nLen = nl.getLength();
                    for (int i = 0; i < nLen; i++) {
                        Node child = nl.item(i);
                        if (child.getNodeType() == Node.ELEMENT_NODE
                            && child.getNodeName().equals(XMLString.TEXT_TABLE_OF_CONTENT_ENTRY_TEMPLATE)) {
	                        int nLevel = Misc.getPosInteger(Misc.getAttribute(child,XMLString.TEXT_OUTLINE_LEVEL),1);
                            if (nLevel<=10) {
                                sEntryStyleName[nLevel] = Misc.getAttribute(child,XMLString.TEXT_STYLE_NAME);
	                        }
                        }
                    }	                        
                }
				
                // Generate entries (in a compact VBox)
                VBox entries = new VBox(true);
                box.addBox(entries);
                int nMaxLevel = Misc.getPosInteger(Misc.getAttribute(source,XMLString.TEXT_OUTLINE_LEVEL),1);
                ListCounter counter = new ListCounter(wsc.getOutlineStyle());
                int nLen = headings.getLength();
                for (int i=0; i<nLen; i++) {
                    Node heading = headings.item(i);
                    int nLevel = Misc.getPosInteger(Misc.getAttribute(heading,XMLString.TEXT_LEVEL),1);
                    counter.step(nLevel);
                    if (nLevel<=nMaxLevel) {
                        PBox par = getStyledPBox(sEntryStyleName[nLevel]);
                        entries.addBox(par);
                        par.addText(counter.getLabel());
                        traverseInlineText(heading,par);
                    }
                }
            }
        }
    }

    /*
     * Process Bibliography
     */
    private void handleBibliography (Node onode, VBox box) {
        // Use the content, not the template
        // This is a temp. solution. Later we want to be able to create
        // hyperlinks from the bib-item to the bibliography, so we
        // have to recreate the bibliography from the template.
        Node body = Misc.getChildByTagName(onode,XMLString.TEXT_INDEX_BODY);
        if (body!=null) {
            Node title = Misc.getChildByTagName(body,XMLString.TEXT_INDEX_TITLE);
            bInHeading = true;
            if (title!=null) { traverseBlockText(title,box); }
            bInHeading = false;
            // Typeset entries in compact VBox:
            VBox entries = new VBox(true);
            box.addBox(entries);
            traverseBlockText(body,entries);
        }     
    }

    /*
     * Process inline text
     */
    private void traverseInlineText (Node onode,PBox box) {        
                              
        if (onode.hasChildNodes()) {
            NodeList nList = onode.getChildNodes();
            int nLen = nList.getLength();
                       
            for (int i = 0; i < nLen; i++) {
                
                Node child = nList.item(i);
                short nodeType = child.getNodeType();
               
                switch (nodeType) {
                    case Node.TEXT_NODE:
                        String s = child.getNodeValue();
                        if (bInHeading) { s = s.toUpperCase(); }
                        if (s.length() > 0) { box.addText(s); }
                        break;
                        
                    case Node.ELEMENT_NODE:
                        String sName = child.getNodeName();
                        if (child.getNodeName().equals(XMLString.TEXT_S)) {
                            int count= Misc.getPosInteger(Misc.getAttribute(child,XMLString.TEXT_C),1);
                            for ( ; count > 0; count--) { box.addText(" "); }
                        }
                        else if (sName.equals(XMLString.TEXT_TAB_STOP)) {
                            // Tab stops are not supported (yet?)
                            box.addText(" ");
                        }
                        else if (sName.equals(XMLString.TEXT_LINE_BREAK)) {
                            // Line breaks are not supported (yet?)
                            box.addText(" ");
                        }
                        else if (sName.equals(XMLString.TEXT_SPAN)) {
                            handleSpan(child,box);
                        }
                        else if (sName.equals(XMLString.TEXT_FOOTNOTE)) {
                            handleFootnote(child,box);
                        }
                        else if (sName.equals(XMLString.TEXT_ENDNOTE)) {
                            handleEndnote(child,box);
                        }
                        else if (sName.equals(XMLString.DRAW_OBJECT)) {
                            handleDrawObject(child,box);
                        }
                        else if (sName.equals(XMLString.DRAW_TEXT_BOX)) {
                            handleDrawTextBox(child,box);
                        }
						else if (sName.startsWith("text:")) {
							 traverseInlineText(child,box);
						}
				// lots of stuff missing.....  

                        else {
                            // Should maybe have a default in here.
                        }
                        break;
                    default:
                        // Do nothing
                }
            }
        }
    }
	
    private void handleSpan(Node onode,PBox box) {
        String sStyleName = Misc.getAttribute(onode,XMLString.TEXT_STYLE_NAME);
        StyleWithProperties style = wsc.getTextStyle(sStyleName);
 
        boolean bEmphasize = false;
        boolean bStrong = false;
        boolean bUnderline = false;
        boolean bSub = false;
        boolean bSuper = false;		

        if (style!=null) {
            String s;
			
            s = style.getProperty(XMLString.FO_FONT_STYLE);
            if ("italic".equals(s) || "oblique".equals(s)) { bEmphasize = true; }
	  
            s = style.getProperty(XMLString.FO_FONT_WEIGHT);
            if ("bold".equals(s)) { bStrong = true; }

            s = style.getProperty(XMLString.STYLE_TEXT_UNDERLINE);
            if (s!=null && !"none".equals(s)) { bUnderline = true; }
        }
		
        StringBuffer buf = new StringBuffer();
		
        if (bSub) { box.addText("_"); }
        if (bSuper) { box.addText("^"); }
        if (bUnderline) { box.addText("_"); }
        if (bStrong) { box.addText("*"); }
        if (bEmphasize) { box.addText("/"); }

        traverseInlineText (onode,box);

        if (bEmphasize) { box.addText("/"); }
        if (bStrong) { box.addText("*"); }
        if (bUnderline) { box.addText("_"); }
    }
   

	
    private String traversePCDATA(Node onode) {
        StringBuffer buf = new StringBuffer();
        if (onode.hasChildNodes()) {
            NodeList nl = onode.getChildNodes();
            int nLen = nl.getLength();
            for (int i=0; i<nLen; i++) {
                if (nl.item(i).getNodeType()==Node.TEXT_NODE) {
                    buf.append(nl.item(i).getNodeValue());
                }
            }
        }
        return buf.toString();
    }
    
    /* Process a footnote */
    private void handleFootnote(Node onode, PBox box) {
        // The citation
        Node citation = Misc.getChildByTagName(onode,XMLString.TEXT_FOOTNOTE_CITATION);
        box.addText("(");
        box.addText(traversePCDATA(citation));
        box.addText(")");
        // The footnote
        Node body = Misc.getChildByTagName(onode,XMLString.TEXT_FOOTNOTE_BODY);
        traverseNoteBody(citation,body,footnotes);
    }

    /* Process an endnote */
    private void handleEndnote(Node onode, PBox box) {
        // The citation
        Node citation = Misc.getChildByTagName(onode,XMLString.TEXT_ENDNOTE_CITATION);
        box.addText("(");
        box.addText(traversePCDATA(citation));
        box.addText(")");
        // The endnote
        Node body = Misc.getChildByTagName(onode,XMLString.TEXT_ENDNOTE_BODY);
        traverseNoteBody(citation,body,endnotes);
    }

	/*
     * Process the contents of a footnote or endnote
     */
    private void traverseNoteBody (Node citation, Node onode, VBox box) {
        // Create the end-/footnote symbol
        // Save citation for later insertion:
        sAsap += "(" + traversePCDATA(citation) + ") ";

        if (onode.hasChildNodes()) {
            NodeList nList = onode.getChildNodes();
            int len = nList.getLength();
            
            for (int i = 0; i < len; i++) {
                Node child = nList.item(i);
                
                if (child.getNodeType() == Node.ELEMENT_NODE) {
                    String nodeName = child.getNodeName();
                    
                    if (nodeName.equals(XMLString.TEXT_H)) {
                        handleHeading(child,box);
                    }

                    if (nodeName.equals(XMLString.TEXT_P)) {
                        handleParagraph(child,box);
                    }
                    
                    if (nodeName.equals(XMLString.TEXT_ORDERED_LIST)) {
                        handleOL(child,0,null,box);
                    }
                    
                    if (nodeName.equals(XMLString.TEXT_UNORDERED_LIST)) {
                        handleUL(child,0,null,box);
                    }
                }
            }
        }
    }

    // elements in the draw namespace

    private String getAnnotation(Node onode) {
        return "(formula)"; //TODO
    }

    private void handleDrawObject(Node onode, PBox box) {
        String sHref = Misc.getAttribute(onode, XMLString.XLINK_HREF);
        if (sHref!=null) { // Embedded object in package or linked object
            if (sHref.startsWith("#")) { // Embedded object in package
                sHref = sHref.substring(3); // Strip leading path chars ("#/.")
                EmbeddedObject object = sxwDoc.getEmbeddedObject(sHref); 
                if (MIMETypes.MATH.equals(object.getType())) { // Formula!
                    EmbeddedXMLObject xmlObject = (EmbeddedXMLObject) object;
                    // Document settings = object.getSettingsDOM();
                    try {
                        box.addText(getAnnotation(xmlObject.getContentDOM().getDocumentElement()));
                    } catch (Exception e) {
                        box.addText("(formula???)");
                    }
                }
                else { // unsupported object
                }
            }
            else { // TODO: Linked object
            }
        }
        else { // flat xml format
            Node formula = Misc.getChildByTagName(onode,XMLString.MATH_MATH);
            if (formula != null) {
                box.addText(getAnnotation(formula));
            }
        }
    }
	
    private void handleDrawTextBox(Node onode, PBox box) {
        String sWidth = Misc.getAttribute(onode,XMLString.SVG_WIDTH);
        VBox textBox = new VBox(false);
        String sStyleName = Misc.getAttribute(onode, XMLString.DRAW_STYLE_NAME);
        // TODO apply style (border)
        traverseBlockText(onode,textBox);
        // TODO: Collect for insertion after this paragraph
    }
}