/************************************************************************
 *
 *  FormattedText.java
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2.1, as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA  02111-1307  USA
 *
 *  Copyright: 2002-2003 by Henrik Just
 *
 *  All Rights Reserved.
 * 
 *  Version 0.3 (2003-04-04) 
 *
 */

package writer2latex.text;

import java.util.Enumeration;
import java.util.LinkedList;

/**
 * <p>Classes to do "discrete typesetting".</p>
 */
 
abstract class Box {

    protected int nWidth = 0;
    protected int nHeight = 0;
	
    // return the minimal width of this box
    abstract int getMinWidth();
	
    // (Re)format the content of the box to the required width
    // (if possible, cf. getMinWidth()).
    abstract void format(int nWidth);

    // Return the resulting width from the previous call of format.
    // This may be larger than requested, if the formatter failed to format
    // the box with the desired with. 
    int getWidth() { return nWidth; };
	
    // Return the resulting height from the previous call of format.
    int getHeight() { return nHeight; };
	
    // Return the character at row nRow, column nCol as of the previous
    // call of format
    abstract char getChar(int nRow, int nCol);
} 

// Struct to hold info about a single typeset line
class LineDescriptor {
    int nOff;
    int nStart;
    int nLength;
}

// This box is the basic building block.
// It contains a single, formatted paragraph.
class PBox extends Box {
    static final int MIN_TEXT_WIDTH = 5;
	
    static final int LEFT = 0;
    static final int RIGHT = 1;
    static final int CENTER = 2;

    private int nAlign; // flushleft, flushright or centered
    private int nLeftMargin; // left indent - number of characters
    private int nRightMargin; // right indent - number of characters
    private int nIndent; // first line indent - number of characters

    private StringBuffer content = new StringBuffer();
    private String sContent = "";
    private LinkedList lines = new LinkedList();

    PBox(int nAlign, int nLeftMargin, int nRightMargin, int nIndent) {
        this.nAlign = nAlign;
        // Margins must be non negative:
        this.nLeftMargin = nLeftMargin<0 ? 0 : nLeftMargin;
        this.nRightMargin = nRightMargin<0 ? 0 : nRightMargin;
        this.nIndent = this.nLeftMargin+nIndent<0 ? -nLeftMargin : nIndent;
    }
	
    void addText(String s) { content.append(s); }
	
    int getMinWidth() { return nLeftMargin+MIN_TEXT_WIDTH+nRightMargin; }

    void format (int nWidth) {
        // Adjust width to minimal text width
        this.nWidth = nWidth<getMinWidth() ? getMinWidth() : nWidth;
        // This method may be invoked several times; clear the previos result
        lines.clear();

        sContent = content.toString();
        int nLen = sContent.length();
        int nStart = 0;
        while (nStart<nLen) {
            LineDescriptor ld = new LineDescriptor();
            lines.add(ld);
            ld.nStart = nStart;
            ld.nOff = nLeftMargin;
            int nTextWidth = this.nWidth-nLeftMargin-nRightMargin;
            if (nStart==0 && nTextWidth>=nIndent+MIN_TEXT_WIDTH) {
                // indent first line
                nTextWidth = nTextWidth-nIndent;
                ld.nOff = nLeftMargin+nIndent;
            }
            // Search for a suitable line break
            int nEnd = nStart + nTextWidth;
            if (nEnd>=nLen) { // finished!
                ld.nLength = nLen-nStart;
                nStart = nLen;
            }
            else {
                while (nEnd>nStart && sContent.charAt(nEnd)!=' ') { nEnd--; }
                if (nEnd==nStart) { // emergency break!!
                    ld.nLength = nTextWidth;
                    nStart = nStart+nTextWidth;
                }
                else { // found a breakpoint
                    ld.nLength = nEnd - nStart;
                    nStart = nEnd+1;
                }
            }
            // Justify
            switch (nAlign) {
                case LEFT : break;
                case RIGHT : ld.nOff += nTextWidth - ld.nLength; break;
                case CENTER: ld.nOff += (nTextWidth - ld.nLength)/2;
            }
        }
        this.nHeight = lines.size();
    }
	
    char getChar(int nRow, int nCol) {
        if (nRow<0 || nRow>=nHeight || nCol<0 || nCol>=nWidth) { return ' '; }
        LineDescriptor ld = (LineDescriptor) lines.get(nRow);
        if (nCol<ld.nOff) { return ' '; }
        if (nCol>=ld.nOff+ld.nLength) { return ' '; }
        return sContent.charAt(ld.nStart+nCol-ld.nOff);
    }

}

// Struct to hold info about a Box and it's position
class BoxDescriptor {
    Box box;
    int nRow;
    int nCol;
	
    BoxDescriptor(Box box) { this.box = box; }
}

// This box contains a list of boxes to be stacked vertically
class VBox extends Box {
    private LinkedList boxes = new LinkedList();
    private boolean bCompact;
	
    // remember the last box to speed up searching
    private int nLastRow = -1;
    private BoxDescriptor lastBox = null;
	
    VBox(boolean bCompact) {
        super();
        this.bCompact = bCompact;
    }
	
    void addBox(Box box) {
        boxes.add(new BoxDescriptor(box));
    }
	
    int getMinWidth() {
        int nMinWidth = 0;
        int nBoxCount = boxes.size();
        for (int i=0; i<nBoxCount; i++) {
        
        }
        return 10;
    }

    void format(int nWidth) {
        this.nWidth = nWidth;
        int nBoxCount = boxes.size();
        int nRow = 0;
        for (int i=0; i<nBoxCount; i++) {
            if (i>0 && !bCompact) { nRow++; } // empty line between boxes
            BoxDescriptor bd = (BoxDescriptor) boxes.get(i);
            bd.box.format(nWidth);
            if (bd.box.getWidth()>this.nWidth) {
                this.nWidth = bd.box.getWidth();
            }
            bd.nRow = nRow;
            bd.nCol = 0; // flushleft!
            nRow += bd.box.getHeight();
        }
        nHeight = nRow;
    }
	
    char getChar(int nRow, int nCol) {
        BoxDescriptor bd;
        if (nRow==nLastRow) {
            bd = lastBox;
        }
        else { // search for the box
            int i = boxes.size();
            while (--i>=0 && ((BoxDescriptor) boxes.get(i)).nRow>nRow);
            if (i<0) {
                return ' ';
            }
            bd = (BoxDescriptor) boxes.get(i);
            nLastRow = nRow;
            lastBox = bd;
        }
        return bd.box.getChar(nRow-bd.nRow,nCol);
    }
		
}


// This box is a framed VBox
class FBox extends VBox {
    private boolean bTop;
    private boolean bBottom;
    private boolean bLeft;
    private boolean bRight;

    FBox(boolean bCompact, boolean bTop, boolean bBottom, boolean bLeft, boolean bRight) {
        super(bCompact);
        this.bTop = bTop;
        this.bBottom = bBottom;
        this.bLeft = bLeft;
        this.bRight = bRight;
    }

    void format(int nWidth) {
        // Leave space for borders when typesetting content:
        if (bLeft) { nWidth--; }
        if (bRight) { nWidth--; }
        super.format(nWidth);
        // Adjust width and height to make room for borders
        if (bLeft) { this.nWidth++; }
        if (bRight) { this.nWidth++; }
        if (bTop) { this.nHeight++; }
        if (bBottom) { this.nHeight++; }
    }

    char getChar(int nRow, int nCol) {
        // Outside the FBox?
        if (nRow<0 || nRow>=nHeight || nCol<0 || nCol>=nWidth) { 
            return ' ';
        }
        // Top or bottom border (including corners)?
        else if ((nRow==0 && bTop) || (nRow==nHeight-1 && bBottom)) {
            if (nCol==0 && bLeft) { return '+'; }
            else if (nCol==nWidth-1 && bRight) { return '+'; }
            else { return '-'; }
        }
        // Left or right border?
        else if ((nCol==0 && bLeft) || (nCol==nWidth-1 && bRight)) {
            return '|';
        }
        // Interior of the FBox!
        else {
            return super.getChar(nRow-(bTop ? 1 : 0), nCol-(bLeft ? 1 : 0));
        }
    }
	
} 

// This box contains a table of Box'es
/* TODO
class TBox extends Box {
    private BoxDescriptor[] bd;
    private int[] nColWidth;
    private int nRows;
    private int nCols;
	
    TBox(int nRows, int nCols) {
        this.nRows = nRows;
        this.nCols = nCols;
        nColWidth = new int[nCols];
    }
	
    void addBox(int nRow, int nCol, Box box) {
        bd[nRow,nCol].box = box;
    }
	
    void format(int nWidth) {
        // Calculate "ideal" column widths
        int nStdColWidth = (nWidth-nCols-1)/nCols;
        int nWiderCols = nWidth-nCols-1-nCols*nStdColWidth;
        for (int nCol=0; nCol<nCols; nCol++) {
            nColWidth = nCol<nWiderCols ? nStdColWidth : nStdColWidth+1; 
        }
        // Format all cells
        for (int nRow=0; nRow<nRows; nRow++) {
            for (int nCol=0; nCol<nCols; nCol++) {
                if (bd[nRow,nCol]!=null) bd[nRow,nCol].box.format(nColWidth[nCol]);
            }
        }
        // Collect actual column widths
        for (int nCol=0; nCol<nCols; nCol++) {
            for (int nRow=0; i<nRows; i++) {
                if (bd[nRow,nCol].box.getWidth()>nColWidth[nCol]) {
                    nColWidth[nCol] = bd[nRow,nCol].box.getWidth();
                }
            }
        }
        // TODO: Try harder! - Maybe some columns can be narrower...
    }
	
    char getChar(int nRow, int nCol) {
        return '*';
        // TODO (like FBox, only more complicated!)
    }
 
}
*/

/* Future ideas:
* CBox to typeset columns (easy to implement, but ugly result!?)
* Text wrap, ie.
     xxxxxx +-----+
     xxxxxx |yyyyy|
     xxxxxx |yyyyy|
     xxxxxx +-----+
     xxxxxxxxxxxxxx
     xxxxxxxxxxxxxx
This requires that PBox supports "shapepar" ala TeX
* Tables with unevenly distributed columns or:
* Tables with optimized column widths
* Real sub- and superscripts, ie.
                              2
Water is H 0 and a square is a .
          2
Hm... looks like a bad idea for running text, but maybe formulas could be
rendered this way (displayed, not inline).
ie. A= left ( matrix { ... } right ) cdot left ( matrix { ... } right ) would become
    / 1 2 3 \   / 10 \
A = | 4 5 6 | * | 11 |
    \ 7 8 9 /   \ 12 /
	
and int x^3+2x^2+1 dx = 5 would become
/  3     2
| x  + 2x  + 1 dx = 5
/ 
and A_1 rightarrow csup f A_2 would become
    f
A  --> A
 1      2
and sqrt {{2x+5} over {3y-8}} +1 would become
     ____
    /2x+5
_  / ---- + 1
 \/  3y-8
and finally a simple one
(x+1)(x-1) = 0 <=> x=-1 or x=1
Of course not all symbols can be represnted in ascii (alpha, nabla, ...)
that may be worth considering....
*/
