/************************************************************************
 *
 *  DocumentSerializerImpl.java
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2.1, as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA  02111-1307  USA
 *
 *  Copyright: 2002-2004 by Henrik Just
 *
 *  All Rights Reserved.
 * 
 *  Version 0.3.1 (2004-01-04)
 *
 */

package writer2latex.xhtml;

import java.text.Collator;

import java.util.ListIterator;
import java.util.LinkedList;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Locale;

import java.io.File;
import java.io.IOException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.xml.sax.SAXException;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Element;
import org.w3c.dom.*;

import org.openoffice.xmerge.converter.dom.DOMDocument;
import org.openoffice.xmerge.converter.xml.sxw.SxwDocument;
import org.openoffice.xmerge.converter.xml.EmbeddedObject;
import org.openoffice.xmerge.converter.xml.EmbeddedBinaryObject;
import org.openoffice.xmerge.converter.xml.EmbeddedXMLObject;
import org.openoffice.xmerge.ConvertData;
import org.openoffice.xmerge.ConvertException;
import org.openoffice.xmerge.DocumentSerializer;

import writer2latex.Config;
import writer2latex.util.*;
import writer2latex.xmerge.*;
import writer2latex.office.*;

/**
 * <p>Helper class (a struct) to contain information about a Link.</p>
 */
final class LinkDescriptor {
    Element element; // the a-element
    String sId; // the id to link to
    int nIndex; // the index of *this* file
}

/**
 * <p>Helper class (a struct) to contain information about an alphabetical index entry.</p>
 */
final class AlphabeticalEntry {
    String sWord; // the word for the index
    int nIndex; // the original index of this entry
}

/**
 * <p>XHTML implementation of <code>DocumentDeserializer</code> 
 * for use by {@link
 * writer2latex.xhtml.PluginFactoryImpl
 * PluginFactoryImpl}.</p>
 *
 * <p>This converts an OpenOffice Writer XML file to an XHTML+MathML<.</p>
 *
 */
public class DocumentSerializerImpl implements DocumentSerializer {

    private Config config;
	
    private L10n l10n;
                        
    private SxwDocument sxwDoc;
	
    private MathMLConverter mathMLConverter;
    private ImageLoader imageLoader;
    private StyleConverter sconv;

    // At which outline level should we split files (0=no split)?
    int nSplit = 0;

    // List of xhtml output files
    Vector outFiles = new Vector();

    // Current xhtml output file
    protected int nType = XhtmlDocument.XHTML10; // the doctype
    private int nDontSplitLevel = 0; // if > 0 splitting is forbidden
    private int nTocIndex = -1;
    private int nAlphabeticalIndex = -1;
    private int nOutFileIndex = -1;
    private XhtmlDocument htmlDoc;
    private Document htmlDOM;
    private Element htmlHead;
    private Element htmlBody;
    private Element htmlBodyContent;

    // The source document
    private Document styleDOM;
    private Document contentDOM;
    private WriterStyleCollection sc;
    private MetaData metaData;

    // additional data used in the conversion:
    private ListCounter outlineNumbering;
    private int nHeadingCount=0; // use to create hypertargets in heading (used by toc)
    private ConvertData cd;
    private String sScale;
    private boolean bInToc = false;
	
    // Hyperlinks
    Hashtable targets = new Hashtable();
    LinkedList links = new LinkedList();
	
    // Strip illegal characters from internal hyperlink targets
    private ExportNameCollection targetNames = new ExportNameCollection(true);
	
    // Alphabetical index
    Vector index = new Vector();
    int nIndexIndex = -1;
	
    private String sFntCitBodyStyle = null;
    private String sFntCitStyle = null;
    private String sEntCitBodyStyle = null;
    private String sEntCitStyle = null;

    // The collection of all headings (used for toc)
    NodeList headings = null;
	
    // Gather the footnotes and endnotes
    private LinkedList footnotes = new LinkedList();
    private LinkedList endnotes = new LinkedList();
    // Sometimes we have to create an inlinenode in a block context
    // We put it here and insert it in the first paragraph/heading to come:
    private Node asapNode = null;
	
    /**
     *  <p>Initialises a new <code>DocumentSerializerImpl</code> using the.<br>
     *     supplied <code>Document</code></p>
     *
     * <p>The supplied document should be an {@link
     *    org.openoffice.xmerge.converter.xml.sxw.SxwDocument SxwDocument}
     *    object.</p>
     *
     *  @param  document  The <code>Document</code> to convert.    
     */
    public DocumentSerializerImpl(org.openoffice.xmerge.Document doc, int nType) {
        sxwDoc = (SxwDocument) doc;
        this.nType = nType;
    }
    
    /**
     *  <p>Convert the data passed into the <code>DocumentSerializerImpl</code>
     *  constructor into XHTML format.</p>
     *
     *  <p>This method may or may not be thread-safe.  It is expected
     *  that the user code does not call this method in more than one
     *  thread.  And for most cases, this method is only done once.</p>
     *
     *  @return  <code>ConvertData</code> object to pass back the
     *           converted data.
     *
     *  @throws  ConvertException  If any conversion error occurs.
     *  @throws  IOException       If any I/O error occurs.
     */
    public ConvertData serialize() throws IOException, ConvertException {      
        // Get configuration
        String sConfigFileName = sxwDoc.getName()+"-config.xml";
        config = new Config();
        File f = new File(sConfigFileName);
        if (f.exists()) {
            config.read(sConfigFileName);
        }
        else if (config.createDocumentConfig()) {
            config.write(sConfigFileName);
        }
        sScale = config.getXhtmlScaling();
        nSplit = config.getXhtmlSplitLevel();

        l10n = new L10n();
		
        // load root xml streams. Note: When used as an export filter,
        // contentDOM contains *everything* (flat XML).
        styleDOM = sxwDoc.getStyleDOM();
        contentDOM = sxwDoc.getContentDOM();
        sc = new WriterStyleCollection();
        sconv = new StyleConverter(sc,config,nType);
        metaData = new MetaData(sxwDoc);
        mathMLConverter = new MathMLConverter(nType!=XhtmlDocument.XHTML10);
        imageLoader = new ImageLoader(sxwDoc,true);

        cd = new ConvertData();
		

        // Load style info first
        sc.loadStylesFromDOM(styleDOM,contentDOM);
        outlineNumbering = new ListCounter(sc.getOutlineStyle());
		
        // Set locale to document language
        ParStyle style = sc.getDefaultParStyle();
        if (style!=null) {
            String sLang = style.getProperty(XMLString.FO_LANGUAGE);
            String sCountry = style.getProperty(XMLString.FO_COUNTRY);
            if (sLang!=null) {
                if (sCountry==null) { l10n.setLocale(sLang); }
                else { l10n.setLocale(sLang+"-"+sCountry); }
            }
        }

        // Styles for footnotes and endnotes
        PropertySet notes = sc.getFootnotesConfiguration();
        sFntCitBodyStyle = notes.getProperty(XMLString.TEXT_CITATION_BODY_STYLE_NAME);
        sFntCitStyle = notes.getProperty(XMLString.TEXT_CITATION_STYLE_NAME);
        notes = sc.getEndnotesConfiguration();
        sEntCitBodyStyle = notes.getProperty(XMLString.TEXT_CITATION_BODY_STYLE_NAME);
        sEntCitStyle = notes.getProperty(XMLString.TEXT_CITATION_STYLE_NAME);
		
		NodeList list;
        // Traverse the body
        nextOutFile();
        list = contentDOM.getElementsByTagName(XMLString.OFFICE_BODY);
        int nLen = list.getLength();
        if (nLen > 0) {
            Node body = list.item(0);
            traverseBlockText(body, htmlBodyContent);
        }
		
        // Add footnotes and endnotes
        insertFootnotes(htmlBody);
        insertEndnotes(htmlBody);

        // Resolve links
        ListIterator iter = links.listIterator();
        while (iter.hasNext()) {
            LinkDescriptor ld = (LinkDescriptor) iter.next();
            Integer targetIndex = (Integer) targets.get(ld.sId);
            if (targetIndex!=null) {
                int nTargetIndex = targetIndex.intValue();
                if (nTargetIndex == ld.nIndex) { // same file
                    ld.element.setAttribute("href","#"+targetNames.getExportName(ld.sId));
                }
                else {
                    ld.element.setAttribute("href",getOutFileName(nTargetIndex,true)
                                            +"#"+targetNames.getExportName(ld.sId));
                }
            }
        }

        // Export styles (temp.)
        if (!config.xhtmlIgnoreStyles()) {
            for (int i=0; i<=nOutFileIndex; i++) {
                Document dom = ((XhtmlDocument) outFiles.get(i)).getContentDOM();
                NodeList hlist = dom.getElementsByTagName("head");
                hlist.item(0).appendChild(sconv.exportStyles(dom));
            }
        }
        
        // Create headers & footers
        if (nSplit>0) {
            for (int i=0; i<=nOutFileIndex; i++) {
                Document dom = ((XhtmlDocument) outFiles.get(i)).getContentDOM();
                NodeList hlist = dom.getElementsByTagName("div");
                Node header = hlist.item(0);
                Node footer = hlist.item(hlist.getLength()-1);
                addNavigationLink(dom,header,l10n.get(L10n.HOME),0);
                addNavigationLink(dom,header,l10n.get(L10n.PREVIOUS),i-1);
                addNavigationLink(dom,header,l10n.get(L10n.NEXT),i+1);
                if (nTocIndex>=0) {
                    addNavigationLink(dom,header,l10n.get(L10n.CONTENTS),nTocIndex);
                }
                if (nAlphabeticalIndex>=0) {
                    addNavigationLink(dom,header,l10n.get(L10n.INDEX),nAlphabeticalIndex);
                }
                header.appendChild(dom.createElement("hr"));
                footer.appendChild(dom.createElement("hr"));
                addNavigationLink(dom,footer,l10n.get(L10n.HOME),0);
                addNavigationLink(dom,footer,l10n.get(L10n.PREVIOUS),i-1);
                addNavigationLink(dom,footer,l10n.get(L10n.NEXT),i+1);
                if (nTocIndex>=0) {
                    addNavigationLink(dom,footer,l10n.get(L10n.CONTENTS),nTocIndex);
                }
                if (nAlphabeticalIndex>=0) {
                    addNavigationLink(dom,footer,l10n.get(L10n.INDEX),nAlphabeticalIndex);
                }
            }
        }

        return cd;
    }
	
    private void addNavigationLink(Document dom, Node node, String s, int nIndex) {
        if (nIndex>=0 && nIndex<=nOutFileIndex) {
            Element a = dom.createElement("a");
            a.setAttribute("href",getOutFileName(nIndex,true));
            a.appendChild(dom.createTextNode(s));
            node.appendChild(dom.createTextNode("["));
            node.appendChild(a);
            node.appendChild(dom.createTextNode("] "));
        }
        else {
            node.appendChild(dom.createTextNode("["+s+"] "));
        }
    }

    ////////////////////////////////////////////////////////////////////////
    // BLOCK TEXT

    private void traverseBlockText(Node onode, Node hnode) {
        traverseBlockText(onode,0,null,hnode);
    } 
	
    private void traverseBlockText(Node onode, int nLevel, String styleName, Node hnode) {
        if (!onode.hasChildNodes()) { return; }
        boolean bAfterHeading = false;
        NodeList nList = onode.getChildNodes();
        int nLen = nList.getLength();
        int i = 0;
        while (i < nLen) {
            Node child = nList.item(i);
            
            if (child.getNodeType() == Node.ELEMENT_NODE) {
                String nodeName = child.getNodeName();
                // Anything but a section bloks splitting
                if (!nodeName.equals(XMLString.TEXT_SECTION)) { nDontSplitLevel++; }
                
                if (nodeName.startsWith("draw:")) {
                    handleDrawElement(child,hnode);
                }
                else if (nodeName.equals(XMLString.TEXT_P)) {
                    // is there a block element, we should use?
                    String sStyleName = Misc.getAttribute(child,XMLString.TEXT_STYLE_NAME);
                    XhtmlStyleMap xpar = config.getXParStyleMap();
                    if (sStyleName!=null && xpar.contains(sStyleName)) {
                        Node curHnode = hnode;
                        String sBlockElement = xpar.getBlockElement(sStyleName);
                        String sBlockCss = xpar.getBlockCss(sStyleName);
                        if (xpar.getBlockElement(sStyleName).length()>0) {
                            Element block = htmlDOM.createElement(xpar.getBlockElement(sStyleName));
                            if (!"(none)".equals(xpar.getBlockCss(sStyleName))) {
                                block.setAttribute("class",xpar.getBlockCss(sStyleName));
                            }
                            hnode.appendChild(block);
                            curHnode = block;
                        }
                        boolean bMoreParagraphs = true;
                        do {
                            handleParagraph(child,curHnode);
                            bMoreParagraphs = false;
                            if (++i<nLen) {
                                child = nList.item(i);
                                nodeName = child.getNodeName();
                                if (nodeName.equals(XMLString.TEXT_P)) {
                                    String sCurStyleName = Misc.getAttribute(child,XMLString.TEXT_STYLE_NAME);
                                    if (sCurStyleName!=null && xpar.contains(sCurStyleName)) {
                                        if (sBlockElement.equals(xpar.getBlockElement(sCurStyleName)) &&
	                                        sBlockCss.equals(xpar.getBlockCss(sCurStyleName))) {
                                            bMoreParagraphs = true;
                                         }
                                    }
                                }
                            }
                        } while (bMoreParagraphs);
                        i--;
                    }
                    else {
                        handleParagraph(child,hnode);
                    }
                }
                else if(nodeName.equals(XMLString.TEXT_H)) {
                    int nOutlineLevel = Misc.getPosInteger(Misc.getAttribute(child, XMLString.TEXT_LEVEL),1);
                    hnode = maybeSplit(hnode,nOutlineLevel,bAfterHeading);
                    handleHeading(child,hnode);
                }
                else if (nodeName.equals(XMLString.TEXT_UNORDERED_LIST)) {
                    handleUL(child,nLevel+1,null,hnode);
                }
                else if (nodeName.equals(XMLString.TEXT_ORDERED_LIST)) {
                    handleOL(child,nLevel+1,null,hnode);
                }
                else if (nodeName.equals(XMLString.TABLE_TABLE)) {
                    handleTable(child,hnode);
                }
                else if (nodeName.equals(XMLString.TABLE_SUB_TABLE)) {
                    handleTable(child,hnode);
                }
                else if (nodeName.equals(XMLString.TEXT_SECTION)) {
                    handleSection(child,hnode);
                }
                else if (nodeName.equals(XMLString.TEXT_TABLE_OF_CONTENT)) {
                    hnode = maybeSplit(hnode,1,bAfterHeading);
                    handleTOC(child,hnode);
                }
                else if (nodeName.equals(XMLString.TEXT_ILLUSTRATION_INDEX)) {
                    handleLOF(child,hnode);
                }
                else if (nodeName.equals(XMLString.TEXT_TABLE_INDEX)) {
                    handleLOT(child,hnode);
                }
                else if (nodeName.equals(XMLString.TEXT_OBJECT_INDEX)) {
                    handleObjectIndex(child,hnode);
                }
                else if (nodeName.equals(XMLString.TEXT_USER_INDEX)) {
                    handleUserIndex(child,hnode);
                }
                else if (nodeName.equals(XMLString.TEXT_ALPHABETICAL_INDEX)) {
                    hnode = maybeSplit(hnode,1,bAfterHeading);
                    handleAlphabeticalIndex(child,hnode);
                }
                else if (nodeName.equals(XMLString.TEXT_BIBLIOGRAPHY)) {
                    hnode = maybeSplit(hnode,1,bAfterHeading);
                    handleBibliography(child,hnode);
                }
                else if (nodeName.equals(XMLString.TEXT_SEQUENCE_DECLS)) {
                    //handleSeqeuenceDecls(child);
                }
                // Reenable splitting
                if (!nodeName.equals(XMLString.TEXT_SECTION)) { nDontSplitLevel--; }
                // Remember if this was a heading
                bAfterHeading = nodeName.equals(XMLString.TEXT_H);
            }
            i++;
        }
    }
    
    private Node maybeSplit(Node node, int nLevel, boolean bAfterHeading) {
        if (bAfterHeading || nDontSplitLevel>1) { return node; }
        if (nSplit>=nLevel) { nextOutFile(); return htmlBodyContent; }
        return node;
    }

    /* Process a text:section tag */
    private void handleSection(Node onode, Node hnode) {
	    // sections are ignored - columns wouldn't be interesting
        String sName = Misc.getAttribute(onode,XMLString.TEXT_NAME);
        asapNode = createTarget(sName+"%7Cregion");
        traverseBlockText(onode, hnode);
    }

    /*
     * Process a table:table tag
     */
    private void handleTable(Node onode, Node hnode) {
        String sName = Misc.getAttribute(onode,XMLString.TABLE_NAME);
        //asapNode = createTarget(sName+"%7Ctable");

        TableGridModel tgm = new TableGridModel(onode);
        // Create table
        Element table = htmlDOM.createElement("table");
        addTarget(table,sName+"%7Ctable");
        hnode.appendChild(table);
        boolean bIsSubtable = XMLString.TABLE_SUB_TABLE.equals(onode.getNodeName());
        if (bIsSubtable) { // a subtable should have no margin!
            table.setAttribute("style","margin:0pt");
        }
        else {
            String sStyleName = Misc.getAttribute(onode,XMLString.TABLE_STYLE_NAME);
            if (!config.xhtmlIgnoreStyles()) {
                sconv.applyTableStyle(sStyleName,table);
            }
        }
        // Create columns
        int nColCount = tgm.getColCount();
        for (int nCol=0; nCol<nColCount; nCol++) {
            Element col = htmlDOM.createElement("col");
            table.appendChild(col);
            // Note: Always apply column style!
            sconv.applyColumnStyle(tgm.getCol(nCol).getStyleName(),col);
        }
        // Create groups
        Element thead = htmlDOM.createElement("thead");
        Element tbody = htmlDOM.createElement("tbody");
        // Create rows
        int nRowCount = tgm.getRowCount();
        Element tgroup = thead;
        if (tgm.getRow(nRowCount-1).isHeader()) { // no body? xhtml will be unhappy
            tgroup = tbody;
        }
        for (int nRow=0; nRow<nRowCount; nRow++) {
            if (!tgm.getRow(nRow).isHeader()) {
                tgroup = tbody;
            }
            Element tr = htmlDOM.createElement("tr");
            tgroup.appendChild(tr);
            for (int nCol=0; nCol<nColCount; nCol++) {
                Node cell = tgm.getCell(nRow,nCol);
                if (XMLString.TABLE_TABLE_CELL.equals(cell.getNodeName())) {
                    Element td = htmlDOM.createElement("td");
                    tr.appendChild(td);
                    String sRowSpan = Misc.getAttribute(cell,XMLString.TABLE_NUMBER_ROWS_SPANNED);
                    if (Misc.getPosInteger(sRowSpan,1)>1) {
                        td.setAttribute("rowspan",sRowSpan);
                    }
                    String sColSpan = Misc.getAttribute(cell,XMLString.TABLE_NUMBER_COLUMNS_SPANNED);
                    if (Misc.getPosInteger(sColSpan,1)>1) {
                        td.setAttribute("colspan",sColSpan);
                    }
                    // Is this a subtable?
                    Node subTable = Misc.getChildByTagName(cell,XMLString.TABLE_SUB_TABLE);
                    if (subTable!=null) { // if it is, there should be no padding in this cell
                        td.setAttribute("style","padding:0mm");
                    }
                    else {
                        String sStyleName = Misc.getAttribute(cell,XMLString.TABLE_STYLE_NAME);
                        if (!config.xhtmlIgnoreStyles()) {
                            sconv.applyCellStyle(sStyleName,td);
                        }
                    }
                    traverseBlockText(cell,td);
                }
                else if (XMLString.TABLE_COVERED_TABLE_CELL.equals(cell.getNodeName())) {
                    // covered table cells are not part of xhtml table model
                }
            }
        }
        // Put the content into the table
        if (thead.hasChildNodes()) { table.appendChild(thead); }
        table.appendChild(tbody);
    }

    /*
     * Process a text:h tag
     */
    private void handleHeading(Node onode, Node hnode) {
        // Numbering: It is possible to define outline numbering in CSS2
        // using counters; but this is not supported by Mozilla 1.0.
        // TODO: Offer CSS2 solution as an alternative later.

        // Note: Conditional styles are not supported
        int nLevel = Misc.getPosInteger(Misc.getAttribute(onode, XMLString.TEXT_LEVEL),1);
        if (nLevel<=6) {
			// add Hx element
            Element heading = htmlDOM.createElement("h"+nLevel);
            hnode.appendChild(heading);
            sconv.setHeadingStyle(nLevel,Misc.getAttribute(onode,XMLString.TEXT_STYLE_NAME));
			
            // Prepend hyperlink target (from toc):
            if (!bInToc) {
                //heading.appendChild(createTarget("heading"+(++nHeadingCount)));
                addTarget(heading,"heading"+(++nHeadingCount));
            }

            // Prepend asapNode
            prependAsapNode(heading);
			
            // Prepend numbering
            heading.appendChild( htmlDOM.createTextNode(outlineNumbering.step(nLevel).getLabel()) );
			
            traverseInlineText(onode,heading);
        }
        else { // beyond h6 - export as ordinary paragraph
            handleParagraph(onode,hnode);
        }
    }

    /*
     * Process a text:p tag
     */
    private void handleParagraph(Node onode, Node hnode) {
        // Note: Conditional styles are not supported
        String sStyleName = Misc.getAttribute(onode,XMLString.TEXT_STYLE_NAME);
        Element par = createParagraph((Element) hnode, sStyleName);
        prependAsapNode(par);
        traverseInlineText(onode,par);        
    }
    
    private void prependAsapNode(Node node) {
        if (asapNode!=null) {
            // May float past a split; check this first
            if (asapNode.getOwnerDocument()!=htmlDOM) {
                asapNode = htmlDOM.importNode(asapNode,true);
            }
            node.appendChild(asapNode); asapNode = null;
        }
    }

    /*
     * Process a text:ordered-list tag.  
     */
    private void handleOL (Node onode, int nLevel, String sStyleName, Node hnode) {
	    // add an OL element
        String sStyleName1 = Misc.getAttribute(onode,XMLString.TEXT_STYLE_NAME);
        if (sStyleName1!=null) { sStyleName = sStyleName1; }
        Element list = htmlDOM.createElement("ol");
        applyListStyle(nLevel,sStyleName,list);
        hnode.appendChild(list);
        traverseList(onode,nLevel,sStyleName,list);
    }

    /*
     * Process a text:unordered-list tag.  
     */
    private void handleUL (Node onode, int nLevel, String sStyleName, Node hnode) {
        // add an UL element
        String sStyleName1 = Misc.getAttribute(onode,XMLString.TEXT_STYLE_NAME);
        if (sStyleName1!=null) { sStyleName = sStyleName1; }
        Element list = htmlDOM.createElement("ul");
        applyListStyle(nLevel,sStyleName,list);
        hnode.appendChild(list);
        traverseList(onode,nLevel,sStyleName,list);
    }

    /*
     * Process the contents of a list
     */
    private void traverseList (Node onode, int nLevel, String styleName, Node hnode) {
        if (onode.hasChildNodes()) {
            NodeList nList = onode.getChildNodes();
            int len = nList.getLength();
            
            for (int i = 0; i < len; i++) {
                Node child = nList.item(i);
                
                if (child.getNodeType() == Node.ELEMENT_NODE) {
                    String nodeName = child.getNodeName();
                    
                    if (nodeName.equals(XMLString.TEXT_LIST_ITEM)) {
					    // add an li element
                        Element item = htmlDOM.createElement("li");
                        hnode.appendChild(item);
                        boolean bRestart = "true".equals(Misc.getAttribute(child,
                            XMLString.TEXT_RESTART_NUMBERING));
                        int nStartValue = Misc.getPosInteger(Misc.getAttribute(child,
                            XMLString.TEXT_START_VALUE),1);
                        if (bRestart) {
                           item.setAttribute("value",Integer.toString(nStartValue));
                        }
                        traverseListItem(child,nLevel,styleName,item);
                    }
                    if (nodeName.equals(XMLString.TEXT_LIST_HEADER)) {
                        // add an li element
                        Element item = htmlDOM.createElement("li");
                        hnode.appendChild(item);
                        item.setAttribute("style","list-style-type:none");
                        traverseListItem(child,nLevel,styleName,item);
                    }
                }
            }
        }
    }
    
    
    /*
     * Process the contents of a list item
     * (a list header should only contain paragraphs, but we don't care)
     */
    private void traverseListItem (Node onode, int nLevel, String styleName, Node hnode) { 
        // First check if we have a single paragraph to be omitted
        // This should happen if we ignore styles and have no style-map
        // for the paragraph style used        
        if (config.xhtmlIgnoreStyles() && onode.hasChildNodes()) {
            NodeList list = onode.getChildNodes();
            int nLen = list.getLength();
            int nParCount = 0;
            boolean bNoPTag = true;
            for (int i=0; i<nLen; i++) {
                if (list.item(i).getNodeType()==Node.ELEMENT_NODE) {
                    if (list.item(i).getNodeName().equals(XMLString.TEXT_P) && bNoPTag) {
                        String sStyleName = Misc.getAttribute(list.item(0),XMLString.TEXT_STYLE_NAME);
                        if (config.getXParStyleMap().contains(sStyleName)) {
                            bNoPTag = false;
                        }
                    }
                    else { // found non-text:p element
                        bNoPTag=false;
                    }
                }
            }
            if (bNoPTag) {
                // traverse the list
                for (int i = 0; i < nLen; i++) {
                    Node child = list.item(i);
          
                    if (child.getNodeType() == Node.ELEMENT_NODE) {
                        String nodeName = child.getNodeName();
                    
                        if (nodeName.equals(XMLString.TEXT_P)) {
                            traverseInlineText(child,hnode);
                        }
                        if (nodeName.equals(XMLString.TEXT_ORDERED_LIST)) {
                            handleOL(child,nLevel+1,styleName,hnode);
                        }
                        if (nodeName.equals(XMLString.TEXT_UNORDERED_LIST)) {
                            handleUL(child,nLevel+1,styleName,hnode);
                        }
                    }
                }
                return;
            }
        }
        // Still here? - traverse block text as usual!
        traverseBlockText(onode,nLevel,styleName,hnode);
    }
		    
    /* Process table of contents
     */
    private void handleTOC (Node onode, Node hnode) {
        nTocIndex = nOutFileIndex;
        bInToc = true;
        if (headings==null) { // collect all headings
            headings = contentDOM.getElementsByTagName(XMLString.TEXT_H);
        }
        Node source = Misc.getChildByTagName(onode,XMLString.TEXT_TABLE_OF_CONTENT_SOURCE);
        if (source!=null) {
            if ("chapter".equals(Misc.getAttribute(source,XMLString.TEXT_INDEX_SOURCE))) {
                // TODO
            }
            else { // generate toc
                Element div = htmlDOM.createElement("div");
                addTarget(div,"tableofcontents");
                hnode.appendChild(div);
                // Generate title
                Node title = Misc.getChildByTagName(source,XMLString.TEXT_INDEX_TITLE_TEMPLATE);
                if (title!=null) {
                    String sStyleName = Misc.getAttribute(title,XMLString.TEXT_STYLE_NAME);
                    Element p = createParagraph(div,sStyleName);
                    traversePCDATA(title,p);
                }
                // Collect style names for entries
                // TODO: Should read the entire template
                String sEntryStyleName[] = new String[11];
                if (source.hasChildNodes()) {
                    NodeList nl = source.getChildNodes();
                    int nLen = nl.getLength();
                    for (int i = 0; i < nLen; i++) {
                        Node child = nl.item(i);
                        if (child.getNodeType() == Node.ELEMENT_NODE
                            && child.getNodeName().equals(XMLString.TEXT_TABLE_OF_CONTENT_ENTRY_TEMPLATE)) {
	                        int nLevel = Misc.getPosInteger(Misc.getAttribute(child,XMLString.TEXT_OUTLINE_LEVEL),1);
                            if (nLevel<=10) {
                                sEntryStyleName[nLevel] = Misc.getAttribute(child,XMLString.TEXT_STYLE_NAME);
	                        }
                        }
                    }	                        
                }
				
                // Generate entries
                int nMaxLevel = Misc.getPosInteger(Misc.getAttribute(source,XMLString.TEXT_OUTLINE_LEVEL),1);
                ListCounter counter = new ListCounter(sc.getOutlineStyle());
                int nHeadingCount = 0;
                int nLen = headings.getLength();
                for (int i=0; i<nLen; i++) {
                    Node heading = headings.item(i);
                    int nLevel = Misc.getPosInteger(Misc.getAttribute(heading,XMLString.TEXT_LEVEL),1);
                    nHeadingCount++;
                    counter.step(nLevel);
                    if (nLevel<=nMaxLevel) {
                        Element p = createParagraph(div,sEntryStyleName[nLevel]);
                        p.appendChild(htmlDOM.createTextNode(counter.getLabel()));
                        Element a = createLink("heading"+nHeadingCount);
                        p.appendChild(a);
                        traverseInlineText(heading,a);
                    }
                }
            }
        }
        bInToc = false;
    }

    /*
     * Process list of illustrations
     */
    private void handleLOF (Node onode, Node hnode) {
        // later
    }

    /*
     * Process list of tables
     */
    private void handleLOT (Node onode, Node hnode) {
        // later
    }

    /*
     * Process Object index
     */
    private void handleObjectIndex (Node onode, Node hnode) {
        // later
    }

    /*
     * Process User index
     */
    private void handleUserIndex (Node onode, Node hnode) {
        // later
    }

    /*
     * Process Alphabetical index
     */
    private void handleAlphabeticalIndex (Node onode, Node hnode) {
        nAlphabeticalIndex = nOutFileIndex;
        Node source = Misc.getChildByTagName(onode,XMLString.TEXT_ALPHABETICAL_INDEX_SOURCE);
        if (source!=null) {
            Element div = htmlDOM.createElement("div");
            addTarget(div,"alphabeticalindex");
            hnode.appendChild(div);
            // Generate title
            Node title = Misc.getChildByTagName(source,XMLString.TEXT_INDEX_TITLE_TEMPLATE);
            if (title!=null) {
                String sStyleName = Misc.getAttribute(title,XMLString.TEXT_STYLE_NAME);
                Element p = createParagraph(div,sStyleName);
                traversePCDATA(title,p);
            }
            // Collect style name for entries
            // TODO: Should read the entire template
            String sEntryStyleName = null;
            if (source.hasChildNodes()) {
                NodeList nl = source.getChildNodes();
                int nLen = nl.getLength();
                for (int i = 0; i < nLen; i++) {
                    Node child = nl.item(i);
                    if (child.getNodeType() == Node.ELEMENT_NODE
                        && child.getNodeName().equals(XMLString.TEXT_ALPHABETICAL_INDEX_ENTRY_TEMPLATE)) {
		                // Note: There are actually three outline-levels: separator, 1, 2 and 3
                        int nLevel = Misc.getPosInteger(Misc.getAttribute(child,XMLString.TEXT_OUTLINE_LEVEL),1);
                        if (nLevel==1) {
                            sEntryStyleName = Misc.getAttribute(child,XMLString.TEXT_STYLE_NAME);
                        }
                    }	                        
                }
            }
            // Sort the index entries
            Collator collator;
            String sLanguage = Misc.getAttribute(source,XMLString.FO_LANGUAGE);
            if (sLanguage==null) { // use default locale
                collator = Collator.getInstance();
            }
            else {
                String sCountry = Misc.getAttribute(source,XMLString.FO_COUNTRY);
                if (sCountry==null) { sCountry=""; }
                collator = Collator.getInstance(new Locale(sLanguage,sCountry));
            }
            for (int i = 0; i<=nIndexIndex; i++) {
                for (int j = i+1; j<=nIndexIndex ; j++) {
                    AlphabeticalEntry entryi = (AlphabeticalEntry) index.get(i);
                    AlphabeticalEntry entryj = (AlphabeticalEntry) index.get(j);
                    if (collator.compare(entryi.sWord, entryj.sWord) > 0) {
                        index.set(i,entryj);
                        index.set(j,entryi);
                    }
                }
            }
            // Generate the index
            Element table = htmlDOM.createElement("table");
            table.setAttribute("style","width:100%");
            div.appendChild(table);
            Element tr = htmlDOM.createElement("tr");
            table.appendChild(tr);
            Element[] td = new Element[4];
            for (int i=0; i<4; i++) {
                td[i] = htmlDOM.createElement("td");
                td[i].setAttribute("style","vertical-align:top");
                tr.appendChild(td[i]);
            }
            int nColEntries = nIndexIndex/4+1;
            int nColIndex = -1;
            for (int i=0; i<=nIndexIndex; i++) {
                if (i%nColEntries==0) { nColIndex++; } 
                AlphabeticalEntry entry = (AlphabeticalEntry) index.get(i);
                Element p = createParagraph(td[nColIndex],sEntryStyleName);
                Element a = createLink("idx"+entry.nIndex);
                p.appendChild(a);
                a.appendChild(htmlDOM.createTextNode(entry.sWord));
            }
        }
        
    }

    /*
     * Process Bibliography
     */
    private void handleBibliography (Node onode, Node hnode) {
        // Use the content, not the template
        // This is a temp. solution. Later we want to be able to create
        // hyperlinks from the bib-item to the actual entry in the bibliography,
        // so we have to recreate the bibliography from the template.
        Node body = Misc.getChildByTagName(onode,XMLString.TEXT_INDEX_BODY);
        if (body!=null) {
            Element div = htmlDOM.createElement("div");
            addTarget(div,"bibliography");
            hnode.appendChild(div);
            //asapNode = createTarget("bibliography");
            Node title = Misc.getChildByTagName(body,XMLString.TEXT_INDEX_TITLE);
            if (title!=null) { traverseBlockText(title,div); }
            traverseBlockText(body,div);
        }     
    }

    ////////////////////////////////////////////////////////////////////////
    // INLINE TEXT

    /*
     * Process inline text
     */
    private void traverseInlineText (Node onode,Node hnode) {        
        String styleName = Misc.getAttribute(onode, XMLString.TEXT_STYLE_NAME);
                              
        if (onode.hasChildNodes()) {
            NodeList nList = onode.getChildNodes();
            int nLen = nList.getLength();
                       
            for (int i = 0; i < nLen; i++) {
                
                Node child = nList.item(i);
                short nodeType = child.getNodeType();
               
                switch (nodeType) {
                    case Node.TEXT_NODE:
                        String s = child.getNodeValue();
                        if (s.length() > 0) {
                            hnode.appendChild( htmlDOM.createTextNode(s) );
                        }
                        break;
                        
                    case Node.ELEMENT_NODE:
                        String sName = child.getNodeName();
                        if (child.getNodeName().startsWith("draw:")) {
                            handleDrawElement(child,hnode);
                        }
                        else if (child.getNodeName().equals(XMLString.TEXT_S)) {
                            int count= Misc.getPosInteger(Misc.getAttribute(child,XMLString.TEXT_C),1);
                            for ( ; count > 0; count--) {
                            hnode.appendChild( htmlDOM.createTextNode("\u00A0") );
                            }
                        }
                        else if (sName.equals(XMLString.TEXT_TAB_STOP)) {
                            // xhtml does not have tab stops
                            hnode.appendChild( htmlDOM.createTextNode(" ") );
                        }
                        else if (sName.equals(XMLString.TEXT_LINE_BREAK)) {
                            hnode.appendChild( htmlDOM.createElement("br") );
                        }
                        else if (sName.equals(XMLString.TEXT_SPAN)) {
                            handleSpan(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_A)) {
                            handleAnchor(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_FOOTNOTE)) {
                            handleFootnote(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_ENDNOTE)) {
                            handleEndnote(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_SEQUENCE)) {
	                        handleSequence(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_PAGE_NUMBER)) {
	                        handlePageNumber(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_PAGE_COUNT)) {
	                        handlePageCount(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_SEQUENCE_REF)) {
	                        handleSequenceRef(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_FOOTNOTE_REF)) {
	                        handleFootnoteRef(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_ENDNOTE_REF)) {
	                        handleEndnoteRef(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_REFERENCE_MARK)) {
	                        handleReferenceMark(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_REFERENCE_MARK_START)) {
	                        handleReferenceMark(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_REFERENCE_REF)) {
	                        handleReferenceRef(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_BOOKMARK)) {
	                        handleBookmark(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_BOOKMARK_START)) {
	                        handleBookmark(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_BOOKMARK_REF)) {
	                        handleBookmarkRef(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_ALPHABETICAL_INDEX_MARK)) {
	                        handleAlphabeticalIndexMark(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_ALPHABETICAL_INDEX_MARK_START)) {
	                        handleAlphabeticalIndexMarkStart(child,hnode);
                        }
                        else if (sName.equals(XMLString.TEXT_BIBLIOGRAPHY_MARK)) {
	                        handleBibliographyMark(child,hnode);
                        }
                        else if (sName.equals(XMLString.OFFICE_ANNOTATION)) {
                            handleOfficeAnnotation(child,hnode);
                        }
						else if (sName.startsWith("text:")) {
							 traverseInlineText(child,hnode);
						}
                        // other tags are ignores;
                        break;
                    default:
                        // Do nothing
                }
            }
        }
    }
	
    private void handleSpan(Node onode, Node hnode) {
        String sStyleName = Misc.getAttribute(onode,XMLString.TEXT_STYLE_NAME);
        Element span = createInline((Element) hnode,sStyleName);
        traverseInlineText (onode,span);
    }

    private void traversePCDATA(Node onode, Node hnode) {
        if (onode.hasChildNodes()) {
            NodeList nl = onode.getChildNodes();
            int nLen = nl.getLength();
            for (int i=0; i<nLen; i++) {
                if (nl.item(i).getNodeType()==Node.TEXT_NODE) {
                    hnode.appendChild( htmlDOM.createTextNode(nl.item(i).getNodeValue()) );
                }
            }
        }
    }
    
    private void handleAnchor(Node onode, Node hnode) {
        // We use the same style to all anchors, even though they may have
        // individual styles assigned. Pass on the names to the style converter:
        // Note: OOo does not seem to use the style attributes?
        String sStyleName = Misc.getAttribute(onode,XMLString.TEXT_STYLE_NAME);
        String sVisitedStyleName = Misc.getAttribute(onode,XMLString.TEXT_VISITED_STYLE_NAME);
        sconv.setAnchorStyle(sStyleName,sVisitedStyleName);
        // Now create the anchor:
        String sHref = Misc.getAttribute(onode,XMLString.XLINK_HREF);
        if (sHref!=null) {
            Element anchor;
            if (sHref.startsWith("#")) { // internal link
                anchor = createLink(sHref.substring(1));
            }
            else { // external link
                anchor = htmlDOM.createElement("a");
                anchor.setAttribute("href",sHref);
            }
            hnode.appendChild(anchor);
            traverseInlineText(onode,anchor);
        }
        else {
            traverseInlineText(onode,hnode);
        }
    }

    /* Process a footnote */
    private void handleFootnote(Node onode, Node hnode) {
        String sId = Misc.getAttribute(onode,XMLString.TEXT_ID);
		Element span = createInline((Element) hnode,sFntCitBodyStyle);
        // Create target and back-link
        Element link = createLink(sId);
        addTarget(link,"body"+sId);
		span.appendChild(link);
        Node citation = Misc.getChildByTagName(onode,XMLString.TEXT_FOOTNOTE_CITATION);
        traversePCDATA(citation,link);
        footnotes.add(onode);
	} 
	
    private void insertFootnotes(Node hnode) {
        int n = footnotes.size();
        for (int i=0; i<n; i++) {
            Node footnote = (Node) footnotes.get(i);
            String sId = Misc.getAttribute(footnote,XMLString.TEXT_ID); 
            Node citation = Misc.getChildByTagName(footnote,XMLString.TEXT_FOOTNOTE_CITATION);
            Node body = Misc.getChildByTagName(footnote,XMLString.TEXT_FOOTNOTE_BODY);
            traverseNoteBody(sId,sFntCitStyle,citation,body,hnode);
        }
        footnotes.clear();
    }

    /* Process an endnote */
    private void handleEndnote(Node onode, Node hnode) {
        String sId = Misc.getAttribute(onode,XMLString.TEXT_ID);
		Element span = createInline((Element) hnode,sEntCitBodyStyle);
        // Create target and back-link
        Element link = createLink(sId);
        addTarget(link,"body"+sId);
		span.appendChild(link);
        Node citation = Misc.getChildByTagName(onode,XMLString.TEXT_ENDNOTE_CITATION);
        traversePCDATA(citation,link);
        endnotes.add(onode);
	} 

    private void insertEndnotes(Node hnode) {
        int n = endnotes.size();
        if (nSplit>0 && n>0) { nextOutFile(); hnode = htmlBodyContent; }
        for (int i=0; i<n; i++) {
            Node endnote = (Node) endnotes.get(i);
            String sId = Misc.getAttribute(endnote,XMLString.TEXT_ID); 
            Node citation = Misc.getChildByTagName(endnote,XMLString.TEXT_ENDNOTE_CITATION);
            Node body = Misc.getChildByTagName(endnote,XMLString.TEXT_ENDNOTE_BODY);
            traverseNoteBody(sId,sEntCitStyle,citation,body,hnode);
        }
    }

	/*
     * Process the contents of a footnote or endnote
     */
    private void traverseNoteBody (String sId, String sCitStyle, Node citation,Node onode, Node hnode) {
        // Create the anchor/footnote symbol:
        // Create target and link
        Element link = createLink("body"+sId);
        addTarget(link,sId);
        if (!config.xhtmlIgnoreStyles()) {
            sconv.applyTextStyle(sCitStyle,link);
        }
        traversePCDATA(citation,link);
        // Add a space and save it for later insertion:
        Element span = htmlDOM.createElement("span");
        span.appendChild(link);
        span.appendChild(htmlDOM.createTextNode(" "));
        asapNode = span;
		
        traverseBlockText(onode,hnode);

        /*if (onode.hasChildNodes()) {
            NodeList nList = onode.getChildNodes();
            int len = nList.getLength();
            
            for (int i = 0; i < len; i++) {
                Node child = nList.item(i);
                
                if (child.getNodeType() == Node.ELEMENT_NODE) {
                    String nodeName = child.getNodeName();
                    
                    if (nodeName.equals(XMLString.TEXT_H)) {
                        handleHeading(child,hnode);
                    }

                    if (nodeName.equals(XMLString.TEXT_P)) {
                        handleParagraph(child,hnode);
                    }
                    
                    if (nodeName.equals(XMLString.TEXT_ORDERED_LIST)) {
                        handleOL(child,0,null,hnode);
                    }
                    
                    if (nodeName.equals(XMLString.TEXT_UNORDERED_LIST)) {
                        handleUL(child,0,null,hnode);
                    }
                }
            }
        }*/
    }

    private void handlePageNumber(Node onode, Node hnode) {
        // doesn't make any sense...
        hnode.appendChild( htmlDOM.createTextNode("(Page number)") );
    }
    
    private void handlePageCount(Node onode, Node hnode) {
       // also no sense
        hnode.appendChild( htmlDOM.createTextNode("(Page count)") );
    }

    private void handleSequence(Node onode, Node hnode) {
        // Use current value, but turn references into hyperlinks
        String sName = Misc.getAttribute(onode,XMLString.TEXT_REF_NAME);
        if (sName!=null && !bInToc) {
            Element anchor = createTarget("seq"+sName);
            hnode.appendChild(anchor);
            traversePCDATA(onode,anchor);
        }
        else {
            traversePCDATA(onode,hnode);
        }        
    }
	
    private void createReference(Node onode, Node hnode, String sPrefix) {
        // Turn reference into hyperlink
        String sFormat = Misc.getAttribute(onode,XMLString.TEXT_REFERENCE_FORMAT);
        String sName = Misc.getAttribute(onode,XMLString.TEXT_REF_NAME);
        Element anchor = createLink(sPrefix+sName);
        hnode.appendChild(anchor);
        if ("page".equals(sFormat)) { // all page numbers are 1 :-)
            anchor.appendChild( htmlDOM.createTextNode("1") );
        }
        else { // in other cases use current value
            traversePCDATA(onode,anchor);
        }
    }
		
    private void handleSequenceRef(Node onode, Node hnode) {
        createReference(onode,hnode,"seq");
    } 

    private void handleFootnoteRef(Node onode, Node hnode) {
        createReference(onode,hnode,"");
    } 
        
    private void handleEndnoteRef(Node onode, Node hnode) {
        createReference(onode,hnode,"");
    } 

    private void handleReferenceMark(Node onode, Node hnode) {
        String sName = Misc.getAttribute(onode,XMLString.TEXT_NAME);
        if (sName!=null && !bInToc) {
            hnode.appendChild(createTarget("ref"+sName));
        }
    }
	
    private void handleReferenceRef(Node onode, Node hnode) {
        createReference(onode,hnode,"ref");
    } 

    private void handleBookmark(Node onode, Node hnode) {
        // Note: Two targets (may be the target of a hyperlink or a reference)
        String sName = Misc.getAttribute(onode,XMLString.TEXT_NAME);
        if (sName!=null && !bInToc) {
            hnode.appendChild(createTarget(sName));
            hnode.appendChild(createTarget("bkm"+sName));
        }
    }
	
    private void handleBookmarkRef(Node onode, Node hnode) {
        createReference(onode,hnode,"bkm");
    } 
	
    private void handleAlphabeticalIndexMark(Node onode, Node hnode) {
        if (bInToc) { return; }
        String sWord = Misc.getAttribute(onode,XMLString.TEXT_STRING_VALUE);
        if (sWord==null) { return; }
        AlphabeticalEntry entry = new AlphabeticalEntry();
        entry.sWord = sWord; entry.nIndex = ++nIndexIndex; 
        index.add(entry);
        hnode.appendChild(createTarget("idx"+nIndexIndex));
    }

    private void handleAlphabeticalIndexMarkStart(Node onode, Node hnode) {
        if (bInToc) { return; }
        String sWord = collectMark(onode);
        if (sWord==null) { return; }
        AlphabeticalEntry entry = new AlphabeticalEntry();
        entry.sWord = sWord; entry.nIndex = ++nIndexIndex; 
        index.add(entry);
        hnode.appendChild(createTarget("idx"+nIndexIndex));
    }
	
    private void handleBibliographyMark(Node onode, Node hnode) {
        if (bInToc) {
            traversePCDATA(onode,hnode);
        }
        else {
            Element anchor = createLink("bibliography");
            hnode.appendChild(anchor);
            traversePCDATA(onode,anchor);
        }
    }
	
    /////////////////////////////////////////////////////////////////////////
    // DRAW ELEMENTS
	
    private void handleDrawElement(Node onode, Node hnode) {
        // onode must be an elment in the draw namespace
        String sName = onode.getNodeName();
        if (sName.equals(XMLString.DRAW_OBJECT)) {
            handleDrawObject(onode,hnode);
        }		
        else if (sName.equals(XMLString.DRAW_IMAGE)) {
            handleDrawImage(onode,hnode);
        }		
        else if (sName.equals(XMLString.DRAW_TEXT_BOX)) {
            handleDrawTextBox(onode,hnode);
        }		
        else if (sName.equals(XMLString.DRAW_A)) {
            // we handle this like text:a
            handleAnchor(onode,hnode);
        }		
    }

    private void killAnnotation(Node hnode) {
        // Remove the contents of the annotation, since mozilla renders it!
        Node semantics = Misc.getChildByTagName(hnode.getLastChild(),XMLString.MATH_SEMANTICS);
        if (semantics!=null) {
            Node annotation = Misc.getChildByTagName(semantics,XMLString.MATH_ANNOTATION);
            if (annotation!=null) {
                if (annotation.hasChildNodes()) {
                    NodeList anl = annotation.getChildNodes();
                    int nLen = anl.getLength();
                    for (int i=0; i<nLen; i++) {
                        annotation.removeChild(anl.item(i));
                    }
                }
            }
        }
    }

    private void handleDrawObject(Node onode, Node hnode) {
        String sHref = Misc.getAttribute(onode, XMLString.XLINK_HREF);
        if (sHref!=null) { // Embedded object in package or linked object
            if (sHref.startsWith("#")) { // Embedded object in package
                sHref = sHref.substring(3); // Strip leading path chars ("#/.")
                EmbeddedObject object = sxwDoc.getEmbeddedObject(sHref); 
                if (MIMETypes.MATH.equals(object.getType())) { // Formula!
                    EmbeddedXMLObject xmlObject = (EmbeddedXMLObject) object;
                    // Document settings = object.getSettingsDOM();
                    try {
                        hnode.appendChild(htmlDOM.createTextNode(" "));
                        mathMLConverter.convert(xmlObject.getContentDOM().getDocumentElement(),hnode);
                        //hnode.appendChild(htmlDOM.importNode(
                        //          xmlObject.getContentDOM().getDocumentElement(),true) );
                        hnode.appendChild(htmlDOM.createTextNode(" "));
                    }
                    catch (SAXException e) {
                        e.printStackTrace();
                    }
                    catch (IOException e) {
                        e.printStackTrace();
                    }
                    killAnnotation(hnode);
                }
                else { // unsupported object
                    hnode.appendChild( htmlDOM.createTextNode("[Warning: object ignored]"));
                }
            }
            else { // TODO: Linked object
                hnode.appendChild( htmlDOM.createTextNode("[Warning: Linked object ignored]"));
            }
        }
        else { // flat xml format
            Node formula = Misc.getChildByTagName(onode,XMLString.MATH_MATH);
            if (formula != null) {
                hnode.appendChild(htmlDOM.createTextNode(" "));
                mathMLConverter.convert(formula,hnode);
                //hnode.appendChild(htmlDOM.importNode(formula,true));
                hnode.appendChild(htmlDOM.createTextNode(" "));
                killAnnotation(hnode);
            }
        }
    }
	
    private void handleDrawImage(Node onode, Node hnode) {
        // TODO: sometimes inlinetext...
        String sName = Misc.getAttribute(onode,XMLString.DRAW_NAME);
        asapNode = createTarget(sName+"%7Cgraphics");

        // get the image from the ImageLoader
        String sFileName = null;
        String sHref = Misc.getAttribute(onode,XMLString.XLINK_HREF);
        if (sHref!=null && !sHref.startsWith("#")) {
            // Linked image is not yet handled by ImageLoader. This is a temp.
            // solution (will go away when ImageLoader is finished)
            sFileName = sHref;
            String sExt = sHref.substring(sHref.lastIndexOf(".")).toLowerCase();
        }
        else { // embedded or base64 encoded image
            BinaryGraphicsDocument bgd = imageLoader.getImage(onode);
            if (bgd!=null) {
                cd.addDocument(bgd);
                sFileName = bgd.getFileName();
            }
        }
		
        if (sFileName==null) { return; }
		
        // Now for the actual inclusion (sFileName contains the file name)
        String sWidth=Misc.getAttribute(onode,XMLString.SVG_WIDTH);
        String sHeight=Misc.getAttribute(onode,XMLString.SVG_HEIGHT);
        String sAnchor = Misc.getAttribute(onode,XMLString.TEXT_ANCHOR_TYPE);
        Element image = htmlDOM.createElement("img");
        image.setAttribute("src",sFileName);
        if (sName == null) { sName = sFileName; }
        image.setAttribute("alt",sName);
        CSVList props = new CSVList(";");
        // Note: Keeps original size (no scale())
        if (sWidth!=null) { props.addValue("width",sWidth); }
        if (sHeight!=null) { props.addValue("height",sHeight); }
        String sStyleName = Misc.getAttribute(onode, XMLString.DRAW_STYLE_NAME);
        applyFrameStyle(sStyleName,image,props,!"as-char".equals(sAnchor));
        hnode.appendChild(image);
    }

    private void handleDrawTextBox(Node onode, Node hnode) {
        String sName = Misc.getAttribute(onode,XMLString.DRAW_NAME);
        asapNode = createTarget(sName+"%7Cframe");
        // The div must be included in something else; here an object
        Element container = htmlDOM.createElement("object");
        hnode.appendChild(container);
        Element textbox = htmlDOM.createElement("div");
        container.appendChild(textbox);
        // Now style it
        String sWidth = Misc.getAttribute(onode,XMLString.SVG_WIDTH);
        String sAnchor = Misc.getAttribute(onode,XMLString.TEXT_ANCHOR_TYPE);
        CSVList props = new CSVList(";");
        if (sWidth!=null) { props.addValue("width",scale(sWidth)); }
        String sStyleName = Misc.getAttribute(onode, XMLString.DRAW_STYLE_NAME);
        applyFrameStyle(sStyleName,textbox,props,!"as-char".equals(sAnchor));
        traverseBlockText(onode,textbox);
    }
 
    private void handleOfficeAnnotation(Node onode, Node hnode) {
        // Get the paragraph, and grab the contents
        Node paragraph = Misc.getChildByTagName(onode,XMLString.TEXT_P);
        if (paragraph!=null) {
            if (paragraph.hasChildNodes()) {
                StringBuffer buf = new StringBuffer();
                NodeList nl = paragraph.getChildNodes();
                int nLen = nl.getLength();
                for (int i=0; i<nLen; i++) {
                    if (nl.item(i).getNodeType()==Node.TEXT_NODE) {
                        buf.append(nl.item(i).getNodeValue());
                    }
                }
                Node commentNode = htmlDOM.createComment(buf.toString());
                hnode.appendChild(commentNode);
            }
        }
    }
	
    /////////////////////////////////////////////////////////////////////////
    // UTILITY METHODS
	
    // Create output file name (docname.html, docname1.html, docname2.html etc.)
    private String getOutFileName(int nIndex, boolean bWithExt) {
        return sxwDoc.getName() + (nIndex>0 ? Integer.toString(nIndex) : "") 
                                + (bWithExt ? htmlDoc.getFileExtension() : "");
    }
	
    // Prepare next output file
    private void nextOutFile() {
        if (nOutFileIndex>=0) { insertFootnotes(htmlBodyContent); }
        htmlDoc = new XhtmlDocument(getOutFileName(++nOutFileIndex,false),nType);
        outFiles.add(nOutFileIndex,htmlDoc);
        cd.addDocument(htmlDoc);
        
        // Create head + body
        htmlDOM = htmlDoc.getContentDOM();
        Element rootElement = htmlDOM.getDocumentElement();
        sconv.applyDefaultLanguage(rootElement);
        rootElement.appendChild(htmlDOM.createComment(
             "This file was converted to XHTML+MathML by Writer2xhtml ver. " +
             Misc.VERSION + 
             ". See http://www.hj-gym.dk/~hj/writer2xhtml for more info."));
        htmlHead = htmlDOM.createElement("head");
        rootElement.appendChild(htmlHead);
        htmlBody = htmlDOM.createElement("body");
        rootElement.appendChild(htmlBody);

        // Add title (required by xhtml)
        Element title = htmlDOM.createElement("title");
        htmlHead.appendChild(title);
        String sTitle = metaData.getTitle();
        if (sTitle==null) { // use filename as fallback
            sTitle = htmlDoc.getFileName();
        }
        title.appendChild( htmlDOM.createTextNode(sTitle) );

        // Declare charset (we need this for xhtml because we have no <?xml ... ?>)
        if (nType==XhtmlDocument.XHTML10) {
            Element meta = htmlDOM.createElement("meta");
            meta.setAttribute("http-equiv","Content-Type");
            meta.setAttribute("content","text/html; charset=utf-8");
            htmlHead.appendChild(meta);
        }
		
        // "Traditional" meta data
        //createMeta("generator","Writer2LaTeX "+Misc.VERSION);
        createMeta("description",metaData.getDescription());
        createMeta("keywords",metaData.getKeywords());

        // Dublin core meta data (optional)
        // Format as recommended on dublincore.org
        // Declare meta data profile
        if (config.xhtmlUseDublinCore()) {
            htmlHead.setAttribute("profile","http://dublincore.org/documents/dcq-html/");
            // Add link to declare namespace
            Element dclink = htmlDOM.createElement("link");
            dclink.setAttribute("rel","schema.DC");
            dclink.setAttribute("href","http://purl.org/dc/elements/1.1/");
            htmlHead.appendChild(dclink);
            // Insert the actual meta data
            createMeta("DC.title",metaData.getTitle());
            createMeta("DC.subject",metaData.getSubject());
            createMeta("DC.description",metaData.getDescription());
            createMeta("DC.creator",metaData.getCreator());
            createMeta("DC.date",metaData.getDate());
            createMeta("DC.language",metaData.getLanguage());
        }
        
        // Add link to stylesheet
        if (config.xhtmlCustomStylesheet().length()>0) {
            Element htmlStyle = htmlDOM.createElement("link");
            htmlStyle.setAttribute("rel","stylesheet");
            htmlStyle.setAttribute("type","text/css");
            htmlStyle.setAttribute("media","all");
            htmlStyle.setAttribute("href",config.xhtmlCustomStylesheet());
            htmlHead.appendChild(htmlStyle);
        }
        /* later....
        if (nSplit>0 && !config.xhtmlIgnoreStyles()) {
            Element htmlStyle = htmlDOM.createElement("link");
            htmlStyle.setAttribute("rel","stylesheet");
            htmlStyle.setAttribute("type","text/css");
            htmlStyle.setAttribute("media","all");
            htmlStyle.setAttribute("href",sxwDoc.getName()+"-styles.css");
            htmlHead.appendChild(htmlStyle);
        }*/
        // Note: For single output file, styles are exported to the doc at the end.

        // Create header+footer in case of multiple output files:
        if (nSplit>=1) {
            Element header = htmlDOM.createElement("div");
            header.setAttribute("id","header");
            htmlBody.appendChild(header);

            htmlBodyContent = htmlDOM.createElement("div");
            htmlBodyContent.setAttribute("id","content");
            htmlBody.appendChild(htmlBodyContent);

            Element footer = htmlDOM.createElement("div");
            footer.setAttribute("id","footer");
            htmlBody.appendChild(footer);
        }
        else {
            htmlBodyContent = htmlBody;
        }
		
    }
	
    /* apply hard formatting attribute style maps */
    private Element applyAttributes(Element node, StyleWithProperties style) {
        if (!config.xhtmlIgnoreStyles()) { return node; }
        node = applyAttribute(node,"bold",sconv.isBold(style));
        node = applyAttribute(node,"italics",sconv.isItalics(style));
        node = applyAttribute(node,"fixed",sconv.isFixed(style));
        node = applyAttribute(node,"superscript",sconv.isSuperscript(style));
        node = applyAttribute(node,"subscript",sconv.isSubscript(style));
        return node;
    }
	
    /* apply hard formatting attribute style maps */
    private Element applyAttribute(Element node, String sAttr, boolean bApply) {
        if (!bApply) { return node; }
        XhtmlStyleMap xattr = config.getXAttrStyleMap();
        if (!xattr.contains(sAttr)) { return node; }
        Element attr = htmlDOM.createElement(xattr.getElement(sAttr));
        if (!"(none)".equals(xattr.getCss(sAttr))) {
            attr.setAttribute("class",xattr.getCss(sAttr));
        }
        node.appendChild(attr);
        return attr;
    }
	
    /** Create a styled paragraph node */
    private Element createParagraph(Element node, String sStyleName) {
        XhtmlStyleMap xpar = config.getXParStyleMap();
        Element par;
        // First check to see if there is a map for this style:
        if (xpar.contains(sStyleName)) {
            par = htmlDOM.createElement(xpar.getElement(sStyleName));
            if (!"(none)".equals(xpar.getCss(sStyleName))) {
                par.setAttribute("class",xpar.getCss(sStyleName));
            }
            node.appendChild(par);
            return par;
        }
        // If this is an automatic style we have to check the parent style as well:
        ParStyle style = sc.getParStyle(sStyleName);
        if (style!=null && style.isAutomatic() && style.getParentName()!=null) {
            String sParentName = style.getParentName();
            if (xpar.contains(sParentName)) {
                par = htmlDOM.createElement(xpar.getElement(sParentName));
                if (!"(none)".equals(xpar.getCss(sParentName))) {
                    par.setAttribute("class",xpar.getCss(sParentName));
                }
                node.appendChild(par);
                return applyAttributes(par,style);
            }
        }
        // Otherwise use the StyleConverter
        par = htmlDOM.createElement("p");
        sconv.applyParStyle(sStyleName,par,!config.xhtmlIgnoreStyles());
        node.appendChild(par);
        return applyAttributes(par,style);
    }
	
    /** Create a styled inline node */
    private Element createInline(Element node, String sStyleName) {
        Element inline;
        // First check to see if there is a map for this style:
        XhtmlStyleMap xtext = config.getXTextStyleMap();
        if (xtext.contains(sStyleName)) {
            inline = htmlDOM.createElement(xtext.getElement(sStyleName));
            if (!"(none)".equals(xtext.getCss(sStyleName))) {
                inline.setAttribute("class",xtext.getCss(sStyleName));
            }
            node.appendChild(inline);
            return inline;
        }
        // Special care if we ignore formatting:
        if (config.xhtmlIgnoreStyles()) {
            StyleWithProperties style = sc.getTextStyle(sStyleName);
            if (style!=null && style.isAutomatic()) {
                node = createInline(node,style.getParentName());
                return applyAttributes(node,style);
            }
            else { // ignore completely!
                return node;
            }
        }
        // Still here? Use styleconverter!
        inline = htmlDOM.createElement("span");
        sconv.applyTextStyle(sStyleName,inline);
        node.appendChild(inline);
        return inline;
    }
	
    /* apply frame style */
    private void applyFrameStyle(String sStyleName,Element element,CSVList props,boolean bFloating) {
        XhtmlStyleMap xframe = config.getXFrameStyleMap();
        if (xframe.contains(sStyleName)) {
            if (!"(none)".equals(xframe.getCss(sStyleName))) {
                element.setAttribute("class",xframe.getCss(sStyleName));
            }
            element.setAttribute("style",props.toString());
        }
        else {
            sconv.applyFrameStyle(sStyleName,element,props,bFloating,!config.xhtmlIgnoreStyles());
        }
    }

    /* apply list style */
    private void applyListStyle(int nLevel, String sStyleName, Element list) {
        XhtmlStyleMap xlist = config.getXListStyleMap();
        if (xlist.contains(sStyleName)) {
            if (!"(none)".equals(xlist.getCss(sStyleName))) {
                list.setAttribute("class",xlist.getCss(sStyleName));
            }
        }
        else {
            sconv.applyListStyle(nLevel,sStyleName,list,!config.xhtmlIgnoreStyles());
        }
    }

    // create a target
    private Element createTarget(String sId) {
        Element a = htmlDOM.createElement("a");
        a.setAttribute("id",targetNames.getExportName(sId));
        targets.put(sId, new Integer(nOutFileIndex));
        return a;
    }
	
    // put a target id on an existing element
    private void addTarget(Element node,String sId) {
        node.setAttribute("id",targetNames.getExportName(sId));
        targets.put(sId, new Integer(nOutFileIndex));
    }

    // create a link
    private Element createLink(String sId) {
        Element a = htmlDOM.createElement("a");
        LinkDescriptor ld = new LinkDescriptor();
        ld.element = a; ld.sId = sId; ld.nIndex = nOutFileIndex;
        links.add(ld);
        return a;
    }
	
    private String scale(String s) {
        return Misc.multiply(sScale,s);
    }
	
    private void createMeta(String sName, String sValue) {
        if (sValue==null) { return; }
        Element meta = htmlDOM.createElement("meta");
        meta.setAttribute("name",sName);
        meta.setAttribute("content",sValue);
        htmlHead.appendChild(meta);
    }
	
    // Collect a mark (this is actually general and should be moved to w2l.office later
    private Node getRightNode(Node node) {
        Node nextNode;
        do {nextNode = node.getNextSibling();
            if (nextNode!=null) { return nextNode; }
            node = node.getParentNode();
        } while (node!=null);
        return null;
    }
	
    private String collectMark(Node node) {
        StringBuffer buf = new StringBuffer();
        String sId = Misc.getAttribute(node,XMLString.TEXT_ID);
        node = getRightNode(node);
        while (node!=null) {
            if (node.getNodeType()==Node.TEXT_NODE) {
                buf.append(node.getNodeValue());
                node = getRightNode(node);
            }
            else if (node.getNodeType()==Node.ELEMENT_NODE) {
               boolean bReady = false;
               String sNodeName = node.getNodeName();
               if (sId.equals(Misc.getAttribute(node,XMLString.TEXT_ID))) {
                   node = null; // found the end mark
                   bReady = true;
               }
               else if (sNodeName.startsWith("text:") 
                        && !sNodeName.equals(XMLString.TEXT_FOOTNOTE)
                        && !sNodeName.equals(XMLString.TEXT_ENDNOTE)) {
                   if (node.hasChildNodes()) {
                       node = node.getFirstChild(); bReady=true;
                   }
               }
               if (!bReady) { node=getRightNode(node); };
            }
        }
        return buf.toString();
    }
		
}